
// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_argdispatcher.cpp,v $
// Copyright(c)'2001 by Apache Team
// Author: O. Lobry
// $Id: acom_argdispatcher.cpp,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ==========================================================================
#include "acom_argdispatcher.h"
#include "acom_ns.h"
#include <algorithm>

bool ACOM::Arg::new_module(const char * module) 
{
    if (std::find(_modlist.begin(), _modlist.end(), module) != _modlist.end())
      return true;
    _modlist.push_back(module);
    _map[module].push_back(_progname);
#ifdef ACOM_DEBUG_ARGS
    std::cout << "module: " << module << " arg[" << _map[module].size()-1 << "]=" << _progname << std::endl;
#endif
    return false;
}

void ACOM::Arg::reset(void) 
{
  std::list<std::string>::iterator icurr;
  std::list<std::string>::iterator iend = _modlist.end();
  for (icurr = _modlist.begin(); icurr != iend; icurr++) {
    _map[*icurr].clear();
    _map[*icurr].push_back(_progname);
  }
}


void ACOM::Arg::parse(int argc, const char ** argv)
{
  char * unparsed = "_unparsed_";
  bool definemodule = false;
  const char *module;

  module = unparsed;
  //  _prog_name = argv[0];

  for (int i = 0; i < argc; i++) {
    
    _map["_all_"].push_back(argv[i]);

    if (!strcmp(argv[i], "--")) {
      definemodule=true;
      continue;
    }
    
    if (definemodule) {
      definemodule=false;
      if (std::find(_modlist.begin(), _modlist.end(), argv[i]) 
          != _modlist.end()) {
        module = argv[i];
        
        continue;
      } else { 
        module = unparsed;
        if (_map[module].size() != 1) {
          _map[module].push_back("--");
#ifdef ACOM_DEBUG_ARGS
          std::cout << "module: " << module << " arg[" << _map[module].size()-1 << "]=" << "--" << std::endl;
#endif
        }
      }
#ifdef ACOM_DEBUG_ARGS
      std::cout << "switch to module: " << module << std::endl;
#endif
    }

      
    _map[module].push_back(argv[i]);
#ifdef ACOM_DEBUG_ARGS
    std::cout << "module: " << module << " arg[" << _map[module].size()-1 << "]=" << argv[i] << std::endl;
#endif
  }
  
}

void ACOM::Arg::parse(int argc, const char ** argv, const char *module)
{
  bool foundmod = false;
  bool skipmodule = false;
  new_module(module);

  for (int i = 0; i < argc; i++) {
    if (!strcmp(argv[i], "--")) {
      foundmod=true;
      continue;
    }
    if (foundmod) {
      if (strcmp(argv[i], module))
        skipmodule=true;
      _map["_all_"].push_back("--");
      _map["_all_"].push_back(module);
      foundmod=false;
    }

    if (skipmodule)
      continue;

    _map[module].push_back(argv[i]);
    _map["_all_"].push_back(argv[i]);
#ifdef ACOM_DEBUG_ARGS
    std::cout << "module: " << module << " arg[" << _map[module].size()-1 << "]=" << argv[i] << std::endl;
#endif
  }
  
}

void ACOM::Arg::parse(const char * module, ACOM::Properties prop) 
{
  ACOM::Properties::Iterator icurr;
  ACOM::Properties::Iterator iend = prop.end();
  
  new_module(module);

  for (icurr = prop.begin(); icurr != iend; icurr++) {
    _map["_all_"].push_back("--");
    _map["_all_"].push_back("--" + icurr->first);
    _map[module].push_back("--" + icurr->first);
#ifdef ACOM_DEBUG_ARGS
    std::cout << "module: " << module << " arg[" << _map[module].size()-1 << "]=" << _map[module][_map[module].size() -1] << std::endl;
#endif
    _map[module].push_back(icurr->second);
    _map["_all_"].push_back(icurr->second);
#ifdef ACOM_DEBUG_ARGS
    std::cout << "module: " << module << " arg[" << _map[module].size()-1 << "]=" << _map[module][_map[module].size() -1] << std::endl;
#endif
  }
}

void ACOM::Arg::parse(const char * module, ACOM::Properties prop, 
                      char * prefix) 
{
  ACOM::Properties::Iterator icurr;
  ACOM::Properties::Iterator iend = prop.end();
  
  new_module(module);

  for (icurr = prop.begin(); icurr != iend; icurr++) {

    // exclude all parameter not begining with prefix
    if (strstr(icurr->first.c_str(), prefix) != icurr->first.c_str())
      continue;
    std::string s(icurr->first, strlen(prefix), 
                  icurr->first.length() - strlen(prefix));
    _map["_all_"].push_back("--");
    _map["_all_"].push_back("--" + s);
    _map[module].push_back("--" + s);
#ifdef ACOM_DEBUG_ARGS
    std::cout << "module: " << module << " arg[" << _map[module].size()-1 << "]=" << _map[module][_map[module].size() -1] << std::endl;
#endif
    _map[module].push_back(icurr->second);
    _map["_all_"].push_back(icurr->second);
#ifdef ACOM_DEBUG_ARGS
    std::cout << "module: " << module << " arg[" << _map[module].size()-1 << "]=" << _map[module][_map[module].size() -1] << std::endl;
#endif
  }
}


const ACOM::Arg::ArgVect&  ACOM::Arg::operator[]( const std::string& key) const
{
  return ((std::map<std::string,ACOM::Arg::ArgVect>&)_map)[key];
}


ACOM::Arg::ArgVect&  ACOM::Arg::operator[]( const std::string& key)
{
  return _map[key];
}


ACOM::Arg::ArgInfo::ArgInfo(Arg& arg, const char * module) 
{
  ArgVect av = arg[module];
  argc = av.size();
  argv = new char*[argc];
  for (int i = 0; i < argc; i++) {
    argv[i] = new char[av[i].size() + 1 ]; // + 1 for '\0' term. char ! 
    strcpy(argv[i], av[i].c_str());
  }
}


ACOM::Arg::ArgInfo::~ArgInfo() 
{
#if 0
  for (int i = 0; i < argc; i++)
    delete [] argv[i];
  //free(argv[i]);
  delete [] argv;
#endif
}

ACOM::Arg::ArgInfo::ArgInfo(std::string &arg) {
  int pos;
  int old_pos;
  int i;
  argc = 0;
  old_pos = -1;
  pos = 0;
  while ((pos = arg.find(" ", pos)) != std::string::npos) {
    if (pos > old_pos + 1) { // skip trailing spaces
      argc++;
    }

    old_pos = pos;
    pos++;
  }

  if (old_pos < (int) arg.size() - 1) argc++;

  argv = new char*[argc];
  // argv = (char **) malloc(argc * sizeof(char*));
  pos = 0;
  old_pos = -1;
  i = 0;
  while ((pos = arg.find(" ", pos)) != std::string::npos) {

    if (pos > old_pos + 1) { // skip trailing spaces
      argv[i] = new char[pos - old_pos +1];
      // argv[i] = (char*) malloc(pos - old_pos +1);
      //      std::cout << "addr=" << &(argv[i][0]) << std::endl;
      // std::cout << "oldpos = " << old_pos << ", pos = " << pos << std::endl;
      // std::cout << "cpy: pos=" << old_pos+1 << ", sz=" << pos -old_pos -1 << std::endl;
      //std::cout << "string=" << &arg[old_pos+1] << std::endl;
      strncpy(argv[i], &(arg.c_str()[old_pos+1]), pos - old_pos - 1);
      argv[i][pos - old_pos - 1] = 0;
      i++;
    }

    old_pos = pos;
    pos++;
  }

  if (old_pos < (int) arg.size() - 1) {
    pos = arg.size();
    argv[i] = new char[pos - old_pos +1];
    // argv[i] = (char *) malloc(pos - old_pos +1);
    strncpy(argv[i], &(arg.c_str()[old_pos+1]), pos - old_pos - 1);
    argv[i][pos - old_pos - 1] = 0;
  }
}

