// =========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_format.cpp,v $
// (c) INRIA, projet APACHE, 2001
// Author: T. Gautier
//
//
//
// =========================================================================
#include <stdlib.h>
#include "acom_format.h"
/* 
*/
namespace ACOM_NET_NAMESPACE {


void* _acom_inuk_malloc( size_t s)
{ return malloc(s); }
void* _acom_inuk_calloc( size_t s1, size_t s2)
{ return calloc(s1,s2); }
void _acom_inuk_free( void* p )
{ return free(p); }

/* -----------------------------------
*/
#define ACOM_MAX_FORMAT 8192

int Format::_counter =0;
Format* Format::_all_fm[ACOM_MAX_FORMAT];


// --------------------------------------------------------------------
Format::Format( )
  : _id(-1), 
    _name("<undefined>"),
    _size(0),
    _fwrite(0), _fread(0),
    _falloc(0), _fdealloc(0),
    _fcstor(0), _fdstor(0),
    _fcreate(0), _fdestroy(0),
    _fcopy(0), _iscompiled(false), _couldbecompiled(true)
{}



// --------------------------------------------------------------------
Format::Format( 
    const std::string& n,
    size_t sz, 
    f_write fwrite, f_read fread,
    f_allocate falloc, f_deallocate fdealloc,
    f_cstor fcstor, f_dstor fdstor, 
    f_create fcreate, f_destroy fdestroy, 
    f_copy fcopy
)
  : _id( ++_counter ), 
    _name(n),
    _size(sz),
    _fwrite(fwrite), _fread(fread),
    _falloc(falloc), _fdealloc(fdealloc),
    _fcstor(fcstor), _fdstor(fdstor),
    _fcreate(fcreate), _fdestroy(fdestroy),
    _fcopy(fcopy), _iscompiled(false), _couldbecompiled(true)
{
  if (_id >= ACOM_MAX_FORMAT) 
  {
    std::cerr << "[Format] to much registered format, recompile ACOM and change ACOM_MAX_FORMAT" << std::endl; 
    throw Runtime("[Format] to much registered format");
  }
  _all_fm[_id] = this;
}


// --------------------------------------------------------------------
Format::Format( 
    size_t sz, 
    f_write fwrite, f_read fread,
    f_allocate falloc, f_deallocate fdealloc,
    f_cstor fcstor, f_dstor fdstor, 
    f_create fcreate, f_destroy fdestroy, 
    f_copy fcopy
)
  : _id( ++_counter ), 
    _name("<undefined>"),
    _size(sz),
    _fwrite(fwrite), _fread(fread),
    _falloc(falloc), _fdealloc(fdealloc),
    _fcstor(fcstor), _fdstor(fdstor),
    _fcreate(fcreate), _fdestroy(fdestroy),
    _fcopy(fcopy), _iscompiled(false), _couldbecompiled(true)
{
  if (_id >= ACOM_MAX_FORMAT) 
  {
    std::cerr << "[Format] to much registered format, recompile ACOM and change ACOM_MAX_FORMAT" << std::endl; 
    throw Runtime("[Format] to much registered format");
  }
  _all_fm[_id] = this;
}



// --------------------------------------------------------------------
Format::Format( const Format& fmt )
 : _id(fmt._id), _name(fmt._name), _size(fmt._size),
   _fwrite(fmt._fwrite), _fread(fmt._fread),
   _falloc(fmt._falloc), _fdealloc(fmt._fdealloc),
   _fcstor(fmt._fcstor), _fdstor(fmt._fdstor),
   _fcreate(fmt._fcreate), _fdestroy(fmt._fdestroy),
   _fcopy(fmt._fcopy), _iscompiled(fmt._iscompiled), _couldbecompiled(fmt._couldbecompiled)
{
}



// --------------------------------------------------------------------
Format& Format::operator=( const Format& fmt )
{
  new (this) Format(fmt);
  return *this;
}



// --------------------------------------------------------------------
ObjectInitFormat::ObjectInitFormat()
{
  Format::init_format();
}



// --------------------------------------------------------------------
#if !defined(INUKTITUT_CXX_IRIX_CC) && !defined(PTH_USE_PTHREAD_LINUX_IA64)
#define ACOM_DECL_SPECIALIZED_FORMAT(TYPE,OBJ)\
template<>\
const Format& ACOMWrapperFormat<TYPE>::theformat = OBJ;\
template<>\
const Format* const ACOMWrapperFormat<TYPE>::format \
  = &ACOMWrapperFormat<TYPE>::theformat;
#else
#define ACOM_DECL_SPECIALIZED_FORMAT(TYPE,OBJ)\
const Format& ACOMWrapperFormat<TYPE>::theformat = OBJ;\
const Format* const ACOMWrapperFormat<TYPE>::format \
  = &ACOMWrapperFormat<TYPE>::theformat;
#endif



// --------------------------------------------------------------------
#define DECL_FORMAT(T,NN)\
const Format NN(\
       ACOM_VALTOSTR(T),\
       sizeof(T),\
       (Format::f_write)&ACOMWrapperFormat<T>::write,\
       (Format::f_read)&ACOMWrapperFormat<T>::read,\
       (Format::f_allocate)&ACOMWrapperFormat<T>::allocate,\
       (Format::f_deallocate)&ACOMWrapperFormat<T>::deallocate,\
       (Format::f_cstor)&ACOMWrapperFormat<T>::cstor,\
       (Format::f_dstor)&ACOMWrapperFormat<T>::dstor,\
       (Format::f_create)&ACOMWrapperFormat<T>::create,\
       (Format::f_destroy)&ACOMWrapperFormat<T>::destroy,\
       (Format::f_copy)&ACOMWrapperFormat<T>::copy\
     );\
     ACOM_DECL_SPECIALIZED_FORMAT(T,NN)


// --------------------------------------------------------------------
class ByteFormat: public Format {
public:
  void write( OStream& s, const void* val, size_t count ) const
  { 
    s.write( &FormatDef::Byte, OStream::DA, val, count); 
  }

  void read( IStream& s, void* val, size_t count ) const
  { 
    s.read( &FormatDef::Byte, IStream::DA, val, count);
  }

  void* allocate( size_t count ) const
  { return acom_inuk_malloc(count); }
  
  void deallocate( void* d, size_t count ) const
  { acom_inuk_free(d); }
  
  void cstor( void* d, size_t count) const
  { ::memset(d,0, count); }

  void dstor( void* d, size_t count ) const
  { }

  void* create( size_t count ) const
  { return acom_inuk_calloc(count,1); }

  void destroy( void* d, size_t count ) const
  { ::free(d); }

  void copy( void* dest, const void* src, size_t count ) const
  { memcpy(dest, src, count); }
};


// --------------------------------------------------------------------
Format FormatDef::Null;
const Format FormatDef::Byte (
       sizeof(char),
       (Format::f_write)&ByteFormat::write,
       (Format::f_read)&ByteFormat::read,
       (Format::f_allocate)&ByteFormat::allocate,
       (Format::f_deallocate)&ByteFormat::deallocate,
       (Format::f_cstor)&ByteFormat::cstor,
       (Format::f_dstor)&ByteFormat::dstor,
       (Format::f_create)&ByteFormat::create,
       (Format::f_destroy)&ByteFormat::destroy,
       (Format::f_copy)&ByteFormat::copy
);

/* DECL_FORMAT(bool, FormatDef::Bool)
DECL_FORMAT(char, FormatDef::Char)
DECL_FORMAT(unsigned char, FormatDef::UChar)
DECL_FORMAT(int, FormatDef::Int)
DECL_FORMAT(unsigned int, FormatDef::UInt)
DECL_FORMAT(short, FormatDef::Short)
DECL_FORMAT(unsigned short, FormatDef::UShort)
DECL_FORMAT(long, FormatDef::Long)
DECL_FORMAT(unsigned long, FormatDef::ULong)
DECL_FORMAT(long long, FormatDef::LLong)
DECL_FORMAT(unsigned long long, FormatDef::ULLong)
DECL_FORMAT(float, FormatDef::Float)
DECL_FORMAT(double, FormatDef::Double)
DECL_FORMAT(long double, FormatDef::LDouble)
*/

// --------------------------------------------------------------------
void Format::init_format()
{
  static bool isinit = false;
  if (isinit) return;
  FormatDef::Null._id =0;
  isinit = true;
}



} // namespace
