// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_netserver.cpp,v $
// Copyright(c)'2001 by Apache Team
// acom_network.cpp : 
// Authors: Jacques Briat, Thierry Gautier
// ==========================================================================
#include <inuk/acom_netserver.h>
#include <inuk/acom_network.h>

namespace ACOM_NET_NAMESPACE {


// --------------------------------------------------------------------
NetworkServer::NetworkServer()
{
}



// --------------------------------------------------------------------
NetworkServer::~NetworkServer()
{
// pure virtual call ????  clear();
}



// --------------------------------------------------------------------
void NetworkServer::clear()
{
  CallContext* call_cc = _listcc.pop();
  while (call_cc !=0) 
  {
    deallocate_callcontext(call_cc);
    call_cc = _listcc.pop();
  }
}



// --------------------------------------------------------------------
NetworkServer::CallContext::~CallContext()
{
  _access.destroy();
}



// --------------------------------------------------------------------
void NetworkServer::CallContext::init(Network* net, size_t capacity )
{
  network = net;
  header  = 0;
  receipt = 0;
  cc      = 0;
  iov        = 0;
  iov_sz     = 0;
  ext_header = 0;
  error_no = ComFailure::OK;
  if (capacity >0) 
    _base = network->allocate_data(capacity);
  else 
    _base = 0;
  _stack_size = capacity;
  _top       = _base;
//  reset();  
}  



// --------------------------------------------------------------------
void NetworkServer::CallContext::destroy()
{
  if (_stack_size >0) 
    network->deallocate_data( _base );
  ext_header = 0;
  cc = 0;
  _stack_size = 0;
  _base = 0;
  _top  = 0;
  network = 0;
}



// --------------------------------------------------------------------
void NetworkServer::CallContext::reserve(size_t sz )
{
  void* newbase = network->allocate_data( sz );
  network->deallocate_data( _base );
  _stack_size = sz;
  _base = newbase;
  _top  = _base;
}



// --------------------------------------------------------------------
void NetworkServer::CallContext::reset()
{
  _im_data_seq.destroy();
  if (receipt !=0) receipt->finish();
  receipt->~Receipt();
  _top       = _base;
  iov        = 0;
  iov_sz     = 0;
  cc         = 0;
  header     = 0;
  receipt    = 0;
  ext_header = 0;
  error_no = ComFailure::OK;
}



// --------------------------------------------------------------------
NetworkServer::CallContext* NetworkServer::acquire_callcontext() throw (BadAlloc)
{
  //std::cout << "IN : " << __PRETTY_FUNCTION__ << std::endl;
  CallContext* call_cc = _listcc.pop();
  if (call_cc !=0) {
    --_szlistcc;
  }
  else {
    call_cc = allocate_callcontext();
  }
  call_cc->_refcount = 1;
//  std::cout << (long)call_cc << "-a:1" << std::endl;
  //std::cout << "OUT : " << __PRETTY_FUNCTION__ << std::endl;
  return call_cc;
}
  


// --------------------------------------------------------------------
void NetworkServer::CallContext::acquire( int rc )
{
  ACOM_ASSERT_DEBUG(network !=0, "bad assertion");
  PTH::CriticalGuard critical( _access );
  _refcount +=rc;
//  std::cout << (long)this << "-a:" << _refcount << std::endl;
}



// --------------------------------------------------------------------
void NetworkServer::release_callcontext(NetworkServer::CallContext* call_cc) throw (InvalidArgument)
{
  { PTH::CriticalGuard critical(call_cc->_access );
    --call_cc->_refcount;
//    std::cout << (long)call_cc << "-r:" << call_cc->_refcount << std::endl;
    if (call_cc->_refcount >0) return;
  }
  call_cc->reset();
  ACOM_ASSERT_DEBUG( call_cc->_refcount ==0, "bad assertion" );
  if ((_szlistcc >=8) || ( call_cc->size() > 1384))
    deallocate_callcontext(call_cc);
  else {
    _listcc.push( call_cc );
    ++_szlistcc;
  }
}
  

 
} // - namespace ACOM_...
