// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_service.cpp,v $
// Copyright(c)'2001 by Apache Team
// acom_api.H : this part contains the  API to be include by users
// Authors: Jacques Briat, Thierry Gautier
// $Id: acom_service.cpp,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ==========================================================================
#include <inuk/acom_service.h>

namespace ACOM_NET_NAMESPACE {

// --------------------------------------------------------------------------
ServiceManager Service::Manager;

// --------------------------------------------------------------------------
#define ACOM_GEN_SERVICEID(lns, sid) Id((lns<<16)|sid)
#define ACOM_MAX_LNS (~0 & (1<<16)) 

ServiceManager::ServiceManager()
 : _services(), _nextlns(DEFAULT_LNS), _freelns()
{
}


// --------------------------------------------------------------------------
int ServiceManager::allocate_lns() throw (OutOfRange)
{
  int retval =0;
  if (!_freelns.empty()) {
    retval = _freelns.back();
    _freelns.pop_back();
  } else {
    if (_nextlns >= ACOM_MAX_LNS) throwit( OutOfRange() );
    retval = _nextlns++;
  }
  return retval;
}



// --------------------------------------------------------------------------
void ServiceManager::deallocate_lns( int lns ) throw (InvalidArgument)
{
  _freelns.push_front( lns );
}



// --------------------------------------------------------------------------
  /// defines a new service 
void ServiceManager::bind( int lns, Id sid, Service* ptobj) 
  throw(InvalidArgument)
{
  if ((lns<0) || (lns>=256))
    throwit(
      InvalidArgument("[Service::bind] bad libray identifier value")
    );
  if (sid>=(1<<24-1))
    throwit(
      InvalidArgument("[Service::bind] bad service identifier value")
    );

  Id id = ACOM_GEN_SERVICEID(lns,sid);
//std::cout << "Register service: " << id << "=(" << lns << "," << sid << ")" << std::endl;

  if (_services.find(id) != _services.end())
    throwit(
      InvalidArgument("[Service::bind] already bind")
    );

  _services[id] = ptobj;
  ptobj->set_id(id);
}


// --------------------------------------------------------------------------
  /// resolv a service stub
Service* ServiceManager::resolv( Service::Id id ) throw(InvalidArgument)
{
   
  std::map<Id, Service*>::const_iterator curr;
  curr = _services.find( id );
  if (curr == _services.end()) {
    //return 0;
    throwit(
      InvalidArgument("[Service::resolv] bad identifier")
    );
  }
  return (*curr).second; 
}


// --------------------------------------------------------------------------
Service::~Service()
{
}


// --------------------------------------------------------------------------
Service::Service()
 : _id(~0)
{}

} // - namespace ATH...
