// =========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_stream.cpp,v $
// (c) INRIA, projet APACHE, 2001
// Author: T. Gautier
//
//
// =========================================================================
#include <inuk/acom_stream.h>
#include <inuk/acom_format.h>
#include <inuk/acom_system.h>


namespace ACOM_NET_NAMESPACE {

bool ByteSeq::flip()
{ int sz=_bufsz-_sz;
  if (_buf && (sz>0))
  {
    IOVectSeq::put(_buf,sz );
  }
  _buf=_bufA->alloc();
  if (_buf) 
  {
    BytePacket::init(_buf,_bufsz);
    return true;
  }
  BytePacket::init(0,0);_bufsz=0;
  return false;
}


bool Map::overflow( Mode m, size_t add_len ) 
{ 
  std::cerr << "[Map::overflow] cannot put or pack more data" << std::endl;
  return false; 
}

void OStream::write ( const Format* f, const void* data, size_t count )
{
  size_t len = count * f->get_size();
  size_t retval = _datastream->pack( data, len );
#if 0
  if (retval != len) {
    bool iscont = overflow( Map::IA, len );
    if (!iscont) {
      std::cerr << "No more buffer to pack data" << std::endl;
      Acom::abort(-1);
    }
  }
#endif
}

void OStream::write ( const Format* f, Mode m, const void* data, size_t count )
{
  size_t len = count * f->get_size();
  if ((m == Map::IA ) || (len <= THRESHOLD_IA))
  {
    size_t retval = _datastream->pack( data, len );
#if 0
    if (retval != len) {
      bool iscont = overflow( Map::IA, len );
      if (!iscont) {
        std::cerr << "No more buffer to pack data" << std::endl;
        Acom::abort(-1);
      }
#if 0
      for (int i=0; i<count; ++i)
        std::cout << "pack[" << i << "]=" << (int)((char*)data)[i] << std::endl;
#endif
    }
#endif
  }
  else if (m == Map::DA) {
    bool retval = _iovstream->put( (void*)data, len );
#if 0
    if (!retval) 
    {
      bool iscont = overflow( Map::DA, len );
      if (!iscont) {
        std::cerr << "No more available iovect entry to put" << std::endl;
        Acom::abort(-1);
      }
    }
#endif
  }
}

void IStream::read ( const Format* f, void* const data, size_t count )
{
  size_t len = count * f->get_size();
  size_t retval = _datastream->unpack( data, len );
#if 0
  if (retval != len) {
    bool iscont = overflow( Map::IA, len );
    if (!iscont) {
      std::cerr << "No more data to unpack" << std::endl;
      throwit( NoMoreResource() );
    }
  }
#endif
}

void IStream::read ( const Format* f, Mode m, void* const data, size_t count )
{
  size_t len = count * f->get_size();
  if ((m == Map::IA ) || (len <= THRESHOLD_IA))
  {
    size_t retval = _datastream->unpack( data, len );
#if 0
    if (retval != len) {
      bool iscont = overflow( Map::IA, len );
      if (!iscont) {
        std::cerr << "No more data to unpack" << std::endl;
        throwit( NoMoreResource() );
      }
#if 0
      for (int i=0; i<count; ++i)
        std::cout << "unpack[" << i << "]=" << (int)((char*)data)[i] << std::endl;
#endif
    }
#endif
  }
  else if (m == Map::DA) {
    if (_iovsz ==0) {
      std::cerr << "No more iovect entry to unput" << std::endl;
      throwit( NoMoreResource() );  
    }
    if (_iov->len != len) {
      std::cerr << "size of iovect entry differs" << std::endl;
      throwit( Length() );
    }
    _iov->data = data;
    --_iovsz; _iov += 1;
  }
}


} // namespace ACOM_NET_NAMESPACE

