// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_system.cpp,v $
// Copyright(c)'97 by Apache Team
// see the copyright file
// Author: T. Gautier
// acom_system.cpp : 
//
//
//
// ==========================================================================
#include <inuk/inuk_conf.h>
#include <inuk/acom_system.h>
//#include <inuk/ics>
#include <fstream>
#include <string>
#include <list>
#include <exception>
#include <netdb.h>  // get host by name
#include <string.h> // strcmp, strdup
#include <stdlib.h> // system
#include <stdio.h>  // sprintf
#include <sys/types.h>   // getpid, wait
#include <sys/wait.h>    // wait
#include <unistd.h>      // fork, getpagesize
#include <signal.h>      // signal


extern "C" 
{
  //#define ACOM_USE_LTDLOPEN
#ifdef ACOM_USE_DLOPEN
#  ifdef ACOM_USE_LTDLOPEN
#    include <ltdl.h>// libtool dlopen/ dlclose
#    define dlclose lt_dlclose
#    define dlsym lt_dlsym
#    define dlerror lt_dlerror
  typedef lt_dlhandle dlhandle_t;
#  else
#    include <dlfcn.h>  // dlopen/ dlclose
  typedef void * dlhandle_t;
#  endif
#else
#ifdef ACOM_USE_TCPNET
  extern void TCPNET_factory_initialize();
  extern void TCPNET_factory_destroy();
#endif
#ifdef ACOM_USE_TAKTUK
  extern void TAKTUK_factory_initialize();
  extern void TAKTUK_factory_destroy();
#endif
#ifdef ACOM_USE_CORNET
  extern void CORNET_factory_initialize();
  extern void CORNET_factory_destroy();
#endif
#ifdef ACOM_USE_SOCKNET
  extern void SOCKNET_factory_initialize();
  extern void SOCKNET_factory_destroy();
#endif
#endif
}


#if defined(INUKTITUT_USE_LINUX_IA64)
#define GETOPTRESET() 
#endif

extern "C" {
typedef void (*signal_handler)(int); 
void acom_sv_signal_hanler(int v) // signal handler for SIG_VEG
{
  ACOM_NET_NAMESPACE::Acom::sv_signal_hanler(v);
}
}

namespace ACOM_NET_NAMESPACE {

// ----------------------------------------------------------------------
// Node description: node rank of the current process, node count.
// ----------------------------------------------------------------------

#ifdef ACOM_USE_DLOPEN
static std::list<dlhandle_t> dynlib_load;  // - list of dynamic loaded library
#endif

static std::string self_canonical_name;     // - canonical name of the host
static int my_pid;     // pid of the current process

Acom::AllNetworks Acom::all_networks; // - all the net connected to it
std::vector<Network*> Acom::all_nids; // - all the net connected to it with key nid

static std::string defaultnetname;   // - name of the default network 
static Network*    defaultnet =0;    // - the default network 


// ----------------------------------------------------------------------
// Terminate handler
// STATE: stable
// ----------------------------------------------------------------------
static void Terminate_handler()
{
#ifdef ACOM_TRACE_DEBUG
  std::cout << "In terminate handler" << std::endl;
#endif
}

// ----------------------------------------------------------------------
// Return informations about the command line
// STATE: stable
// ----------------------------------------------------------------------
int*    Acom::argc =0 ;
char*** Acom::argv = 0;
ACOM::Arg*  Acom::arg;
ACOM::Arg::ArgInfo*  Acom::unparsed_args_info;
ACOM::Arg::ArgInfo*  Acom::all_args_info;
char* Acom::get_argv(int i) throw (InvalidArgument)
{
  if ((i <0) || (i >= *Acom::argc))
    throwit( InvalidArgument("[Acom::get_argv]: index out of bounds"));
  return (*Acom::argv)[i];
}
    
int Acom::get_argc()
{
  return *Acom::argc;
}


// ----------------------------------------------------------------------
// Attach a debugger to the current process.
// Should be improve in order to be "portable" among the set of 
// OS....
// STATE: instable
// ----------------------------------------------------------------------
static char* pid_name = NULL; // name of the current process

void Acom::attach_debugger() throw (Runtime)
{
  // -- Spawn a process dbx with the program name and the pid of the
  // current process.
  // - fork a child on which the debugger run ?
  pid_t pid = getpid();
  if (pid !=0 && pid_name != NULL) {
    char cmd_line[128];
    sprintf(cmd_line, "inukdbx %s %i\n", pid_name, getpid() );
    std::cout << "dbx spawned for:" << pid_name << ", pid:" << getpid() 
              << std::endl;
    if (system(cmd_line) == -1)
      throwit(Runtime("Cannot start dbx"));
    std::cout << "dbx finish:" << pid_name << ", pid:" << getpid() 
              << std::endl;
  }
}


// ----------------------------------------------------------------------
// Set up a new signal handler on sigveg
// STATE: stable
// ----------------------------------------------------------------------
// - static variable of this file
typedef void (*type_old_signal_handler)(int); 
static void (*old_signal_handler)(int);

void Acom::sv_signal_hanler (int v ) 
{
  ACOM_NET_NAMESPACE::Acom::attach_debugger();
  old_signal_handler( v );
}


// ----------------------------------------------------------------------
// Abort the current process and returns.
// STATE: stable
// ----------------------------------------------------------------------
void Acom::abort(int err) throw (ErrorAbort)
{ 
  throwit(ErrorAbort());
}


// ----------------------------------------------------------------------
// Initialise the A0 kernel, the underlayer MPI
// - read options on the commande line
// STATE: stable
// ----------------------------------------------------------------------
static long is_init = 0;
static Network* primary_net =0;

void Acom::initialize(int* _argc, char*** _argv) throw (ErrorAbort,InvalidArgument)
{
  //std::cout << "In: " << __PRETTY_FUNCTION__ << std::endl;
  // to avoid multiple call of this method
  if (is_init >0) return;
  ++is_init;

  if ((!_argc) || (!_argv))
    throwit(InvalidArgument("Acom::initialize: invalid null pointer to argc or argv"));
  


  PTH::Pth::initialize(argc, argv);
  PTH::Pth::commit();
  //ICS::Ics::initialize(argc, argv);

  Format::compile_all();

  Properties prop;
  try 
  {
#ifdef ACOM_USE_DLOPEN
    const char* error;
#endif
    Acom::argc = _argc;
    Acom::argv = _argv;
    Acom::arg = new ACOM::Arg((*argv)[0]);
    Acom::arg->new_module("inuk");
    Acom::arg->parse(*argc - 1, (const char **) &(*argv)[1]);
    Acom::all_networks.clear();

    std::set_unexpected( Terminate_handler );
    //std::set_terminate( Terminate_handler );
    const char* conffile = 0; // - name of the configuration file

    // - setup global variable
    my_pid = getpid();
    pid_name = (*_argv)[0];

    // - setup self_canonical_name 
    char* hostname = getenv("HOST");
    if (hostname ==0) hostname= getenv("HOSTNAME");
    if (hostname !=0)
    {
#if 0
      struct hostent* tmp = gethostbyname( hostname );
      if (tmp ==0)
      {
        //hstrerror( h_errno );
        throwit( InvalidArgument("Cannot get canonical hostname"));
      }
#endif
      self_canonical_name = hostname; //tmp->h_name; 
    }

    // -- parse arguments
    int i, 
        use_dbx,
        rank_dbx =-1; // -- rank_dbx ==0: then dbx on all processes
                      // -- rank_dbx ==-1: then no dbx on this process
                      // -- rank_dbx ==1: then dbx this process

    int curr2set =0;
// #ifdef ACOM_DEBUG_ARGV
//     std::cout << "ACOM, input argv" << std::endl;
//     for (i=0; i<*_argc; i++) std::cout << "ACOM, argv[" << i << "]=" << (*_argv)[i] << std::endl;
// #endif
//    TAKTUK_factory_initialize(); // to remove

#if 0 /* TG: replace this code by initialization if -inuknet is on command line */
#ifdef ACOM_USE_TAKTUK
    TAKTUK_factory_initialize();
#endif
#ifdef ACOM_USE_TCPNET
    TCPNET_factory_initialize();
#endif
#ifdef ACOM_USE_CORNET
    CORNET_factory_initialize();
#endif
#ifdef ACOM_USE_SOCKNET
    SOCKNET_factory_initialize();
#endif
#endif
    ACOM::Arg::ArgVect av = (*Acom::arg)["inuk"];
    for (i=0; i<av.size(); i++) 
    {
      const char* arg = av[i].c_str();
  
      if ((strcmp(arg,"-h") ==0) || (strcmp(arg,"--help") ==0))
      {
        std::cout << "Options define for inuk programs:\n";
        std::cout << "  --usage     : display usage of the library\n"; 
        std::cout << "  -n <#nodes>: set the number of nodes\n"; 
        std::cout << "  -rtdbg     : attach a debugger on each node on signal\n";
        std::cout << "  -debug     : attach a debugger on each node\n";
        std::cout << "  -debug <i> : attach a debugger on the node with rank i\n";
        std::cout << "  -conf <filename>: choose <filename> as configuration file\n";
        std::cout << "  -net <NET> : use <NET> implementation\n";
        std::cout << "  -defaultnet <NET> : use <NET> as default network\n";
        std::cout << std::endl;
        Acom::abort(0);
      }
      else if (strcmp(arg, "-rtdbg") ==0) 
      {
#ifndef INUKTITUT_USE_IRIX
        old_signal_handler = (type_old_signal_handler)::signal( SIGSEGV, (signal_handler)::acom_sv_signal_hanler);
#else        
        //TG: ca ne compile pas ! old_signal_handler = std::signal( SIGSEGV, ::acom_sv_signal_hanler);
#endif
        if ((void*)old_signal_handler == (void*)SIG_ERR)
        {
          std::cerr << "Cannot setup signal handler on SIGSEGV" << std::endl;
        }
        std::cout << "Handler " << std::endl;
      }
      else if (strcmp(arg, "-conf") ==0) 
      {
        if (av.size() <= i+1)
        {
          std::cerr << "-conf <filename>: please give the filename" << std::endl;
          Acom::abort(0);
        }
        conffile = av[i+1].c_str();
        ++i;
        std::ifstream toto(conffile);
        if (!toto.is_open())
        {
          std::cerr << "-conf '" << conffile <<"': filename cannot be open" << std::endl;
          Acom::abort(0);
        }
        // The following does not work because of 'lowercasing' and unicity
        // of properties
        // prop.load(conffile);
        // Acom::arg->parse("inuk", prop, "inukarg_");
      }
      else if (strcmp(arg, "-defaultnet") ==0) 
      {
#ifdef ACOM_TRACE_DEBUG
        std::cerr << "arg = " << argc << std::endl;
        std::cerr << "av[i] = " << av[i] << std::endl;
        std::cerr << "i= " << i << std::endl;
#endif
        if (av.size() <= i+1)
        {
          std::cerr << "-defaultnet <NET>: please give the network name" << std::endl;
          Acom::abort(0);
        }
        defaultnetname = av[i+1];
std::cout << "[ACOM] Default net name=" << defaultnetname << std::endl;
        ++i;
      }
      else if (strcmp(arg, "-net") ==0) 
      {
#ifdef ACOM_TRACE_DEBUG
        std::cerr << "arg = " << av.size() << std::endl;
        std::cerr << "av[i] = " << av[i] << std::endl;
        std::cerr << "i= " << i << std::endl;
#endif
        if (av.size() <= i+1)
        {
          std::cerr << "-net <NET>: please give the network name" << std::endl;
          Acom::abort(0);
        }
        std::string netname = av[i+1];
        ++i;

        AllNetworks::const_iterator curr;
        curr = all_networks.find(netname);
        if (curr != all_networks.end()) 
          continue;
        
#ifndef ACOM_USE_DLOPEN

#ifdef ACOM_USE_TAKTUK
        if (netname == "TAKTUK")
          TAKTUK_factory_initialize();
#endif
#ifdef ACOM_USE_TCPNET
        if (netname == "TCPNET") 
          TCPNET_factory_initialize();
#endif
#ifdef ACOM_USE_CORNET
        if (netname == "CORNET") 
          CORNET_factory_initialize();
#endif
#ifdef ACOM_USE_SOCKNET
        if (netname == "SOCKNET") 
          SOCKNET_factory_initialize();
#endif
#else  /* ACOM_USE_DLOPEN */
        std::string libnetname = "libinuknet_"+netname+".so";
        //        std::cout << "dlopen " << libnetname << std::endl;
#ifdef ACOM_USE_LTDLOPEN
        lt_dlinit();
        dlhandle_t handle = lt_dlopen( libnetname.c_str()); 
#else
        dlhandle_t handle = dlopen( libnetname.c_str(), RTLD_LAZY | RTLD_GLOBAL ); 
#endif
        if (handle ==0)
        {
          std::cerr << dlerror() << std::endl;
          //          std::cerr << dlerror() << std::endl;
          Acom::abort(1);
        }
        void (*fnc_initialize)() = (void (*)())dlsym(handle, (netname+"_factory_initialize").c_str()); 
        if ((error = dlerror()) != 0)
        {
          std::cerr << error << std::endl;
          Acom::abort(1);
        }
        else
        {
          fnc_initialize();
          dynlib_load.push_back( handle );
        }
#endif  // ACOM_USE_DLOPEN

        // add new argument module
        Acom::arg->new_module(netname.c_str());        
      }
      else
      {
        // - compact the array of args to the beginning
        if (i > curr2set) av[curr2set] = av[i];
        ++curr2set;
      }
    }
    //C *_argc = curr2set;
    
    curr2set =0;
    for (i=0; i<av.size(); i++) 
    {
      const char* arg = av[i].c_str();
      int used_arg =0;
  
      if ((strcmp(arg,"-dbg") ==0) || (strcmp(arg,"-debug") ==0))
      {
        used_arg = 1;
        use_dbx = 1;
        if (i+1 >= av.size()) rank_dbx =0;
        else 
        {
          int tmp = atoi(av[i+1].c_str());
          if (tmp !=0)
          {
            i++;
            //            if (tmp == Acom::self_node+1) rank_dbx = 1;
          }
          else rank_dbx =0;
        }
      }
  //C else
  //C {
  // - compact the array of args to the beginning
  //C if (i > curr2set) (*_argv)[curr2set] = (*_argv)[i];
  //C ++curr2set;
  // }
  
  // - suppress argv[currval] to argv[i]
  //C if (used_arg ==1)
  //C {
  //C   int offset = 1+i-currval;
  //C   for (int j=currval; j<*_argc; j++)
  //C   {
  //C     (*_argv)[j] = (*_argv)[j+offset];
  //C   }
  //C  *_argc -= offset;
  // - restore a correct i for next iteration 
  //C  i-=offset; 
  //C }
    }
    //C *_argc = curr2set;
    
#ifdef ACOM_DEBUG_ARGV
    std::cout << "ACOM, output argv" << std::endl;
    for (i=0; i<av.size(); i++) std::cout << "ACOM, argv[" << i << "]=" << av[i] << std::endl;
    std::cout << std::endl;
#endif
    
    //  -- 
    if (rank_dbx >=0)
    {
      Acom::attach_debugger(); 
      std::cout << "attached_debugger " << std::endl;
    }
    //Acom::attach_debugger(); 
    
#if 0
    if (conffile !=0) Acom::compute_topology(conffile);
#endif

    Acom::arg->reset();
    Acom::arg->parse(*argc - 1, (const char **) &(*argv)[1]);
    Acom::unparsed_args_info = new ACOM::Arg::ArgInfo(*Acom::arg,"_unparsed_");
    Acom::all_args_info = new ACOM::Arg::ArgInfo(*Acom::arg,"_all_");
    *_argv =  Acom::unparsed_args_info->argv;
    *_argc =  Acom::unparsed_args_info->argc;

    
    /* initialize primary network */
    Acom::const_iterator iend = Acom::end_network();  
    Acom::const_iterator icurr = Acom::begin_network();

    /* default is the first one */
    if (icurr != iend)
      defaultnet = *icurr;

    while (icurr != iend) 
    {
      if ((*icurr)->get_property("primary") == "true")
      {
        if (primary_net ==0) 
          primary_net=*icurr;
        ACOM::Arg::ArgInfo ai(*arg, (*icurr)->get_name().c_str());
        //       std::cout<<"Initialize primary network " << (*icurr)->get_name() <<std::endl;
        (*icurr)->initialize( &(ai.argc), &(ai.argv));

        GETOPTRESET();
        break;
      }
      /* default is the network with name given by -defaultnet */
      if ((*icurr)->get_name() == defaultnetname)
        defaultnet = *icurr;
      ++icurr;
    }
    
#if 0 /* mv in commit : secondary network needs primary initialized ? */
    /* initialize secondary network */
    iend = Acom::end_network();  
    icurr = Acom::begin_network();  
    while (icurr != iend) 
    {
      if ((primary_net!=0) && ((*icurr)->get_property("secondary") == "true"))
      {
        ACOM::Arg::ArgInfo ai(*arg, (*icurr)->get_name().c_str());
        (*icurr)->initialize( &(ai.argc), &(ai.argv), primary_net);
        //std::cout<<"Secondary network " << (*icurr)->get_name() << " initialized"<<std::endl;
        GETOPTRESET();
      }
      ++icurr;
    }
#endif
  }
  catch (const PosixError& E)
  {
    std::cerr << "[acom::Acom::init]: Posix error:" << E.what() << std::endl;
    throwit(E);
  }
  catch (const RuntimeError& E)
  {
    std::cerr << "[acom::Acom::init]: Runtime error:" << E.what() << std::endl;
    throwit(E);
  }
  catch (const std::exception& E)
  {
    std::cerr << "[acom::Acom::init]: std exception catched" << std::endl;
    throw E;
  }
//std::cout << "Out: " << __PRETTY_FUNCTION__ << std::endl;
}


// ----------------------------------------------------------------------
// Second stage of the initialisation process
// STATE: stable
// ----------------------------------------------------------------------
void Acom::commit() throw ()
{ 
  // to avoid multiple call of this method
  static bool is_init = false;
  if (is_init) return;
  is_init = true;

  if (primary_net !=0) 
    primary_net->commit();

  /* initialize secondary network */
  Acom::const_iterator iend = Acom::end_network();  
  Acom::const_iterator icurr = Acom::begin_network();  
  while (icurr != iend) 
  {
    if ((primary_net!=0) && ((*icurr)->get_property("secondary") == "true"))
    {
      ACOM::Arg::ArgInfo ai(*arg, (*icurr)->get_name().c_str());
      (*icurr)->initialize( &(ai.argc), &(ai.argv), primary_net);
      //std::cout<<"Secondary network " << (*icurr)->get_name() << " initialized"<<std::endl;
      GETOPTRESET();
    }
    /* default is the network with name given by -defaultnet */
    if ((*icurr)->get_name() == defaultnetname)
      defaultnet = *icurr;
    ++icurr;
  }

  /* commit secondary network */
  iend = Acom::end_network();  
  icurr = Acom::begin_network();  
  while (icurr != iend) {
    if ((primary_net!=0) && ((*icurr)->get_property("secondary") == "true"))
        (*icurr)->commit( );
    //std::cout<<"Commit network "<<(*icurr)->get_name()<<std::endl;
    ++icurr;
  }
  //ICS::Ics::commit();
}

// ----------------------------------------------------------------------
// Terminaison of the virtual machine 
// STATE: stable
// ----------------------------------------------------------------------
void Acom::terminate() throw (ErrorAbort)
{ 
  // to avoid multiple call of this method
  if (is_init ==0) return;
  is_init--;

  { // - terminate all modules
    Acom::const_iterator iend = Acom::end_network();  
    Acom::const_iterator icurr = Acom::begin_network();  
    while (icurr != iend) {
      if ((primary_net!=0) && ((*icurr)->get_property("secondary") == "true"))
        (*icurr)->terminate( );
      ++icurr;
    }
    if (primary_net!=0)
      primary_net->terminate( );
  }

  { // - close dynamic libraries
#ifdef ACOM_USE_DLOPEN
    const char* error;
    //    std::list<void*>::iterator iend = dynlib_load.end();
    //    std::list<void*>::iterator icurr = dynlib_load.begin();
    std::list<dlhandle_t>::iterator iend = dynlib_load.end();
    std::list<dlhandle_t>::iterator icurr = dynlib_load.begin();
    while (icurr != iend) {
      void (*fnc_terminate)() = (void (*)())dlsym(*icurr, "factory_destroy");
      if ((error = dlerror()) != 0) 
        std::cerr << error << std::endl;
      else 
        fnc_terminate();
      dlclose((*icurr));
      ++icurr;
      } 
    /*    Acom::const_iterator iend  = Acom::end_network();  
          Acom::const_iterator icurr = Acom::begin_network();  
          while (icurr != iend) {
          void (*fnc_terminate)() =
          (void (*)())dlsym(*icurr, ((*icurr)->get_name()+"_factory_destroy").c_str());
          if ((error = dlerror()) != 0) 
          std::cerr << error << std::endl;
          else 
          fnc_terminate();
          //dlclose((*icurr));
          ++icurr;
      } */
#else
  { // - destroy all modules
    Acom::const_iterator iend  = Acom::end_network();  
    Acom::const_iterator icurr = Acom::begin_network();  
    while (icurr != iend) {
      delete *icurr;
      ++icurr;
    }
  }
/** Replaced by the previous lines of code  
#ifdef ACOM_USE_TAKTUK
    TAKTUK_factory_destroy();
#endif
#ifdef ACOM_USE_TCPNET
    TCPNET_factory_destroy();
#endif
#ifdef ACOM_USE_CORNET
    CORNET_factory_destroy();
#endif
#ifdef ACOM_USE_SOCKNET
    SOCKNET_factory_destroy();
#endif
*/
#endif
  }
  //ICS::Ics::terminate();
  PTH::Pth::terminate();
}


// ----------------------------------------------------------------------
// Iterator over the networks
// ----------------------------------------------------------------------
Acom::const_iterator Acom::begin_network()
{
  return Acom::const_iterator( Acom::all_networks.begin() );
}

Acom::const_iterator Acom::end_network()
{
  return Acom::const_iterator( Acom::all_networks.end() );
}

// ----------------------------------------------------------------------
// Registration/bind/resolv 
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
void Acom::bind_network( const std::string& name, Network* net )
    throw (InvalidArgument)
{
#ifdef ACOM_TRACE_DEBUG
std::cout << "In " << __PRETTY_FUNCTION__ << ", name=" << name << std::endl;
#endif
  AllNetworks::const_iterator curr;
  curr = all_networks.find(name);
  if (curr != all_networks.end()) 
  {
    throwit( InvalidArgument("[Acom::bind] name already registered") );
  }
  all_networks[name] = net;
  if (all_nids.size() <= net->get_id()) {
    all_nids.resize(net->get_id()+1);
  }
  all_nids[net->get_id()] = net;
  
#ifdef ACOM_TRACE_DEBUG
std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif
}


// ----------------------------------------------------------------------
void Acom::rebind_network( const std::string& name, Network* net )
    throw (InvalidArgument)
{
#ifdef ACOM_TRACE_DEBUG
std::cout << "In " << __PRETTY_FUNCTION__ << ", name=" << name << std::endl;
#endif
  all_networks[name] = net;
#ifdef ACOM_TRACE_DEBUG
std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif
}


// ----------------------------------------------------------------------
Network* Acom::resolv_network( const std::string& name )
    throw (InvalidArgument)
{
#ifdef ACOM_TRACE_DEBUG
std::cout << "In " << __PRETTY_FUNCTION__ << ", name = " << name << std::endl;
#endif
  AllNetworks::const_iterator curr = all_networks.find( name );
  if (curr == all_networks.end()) {
    return 0;
    //throwit( InvalidArgument("[Service::resolv] bad network name") );
  }
#ifdef ACOM_TRACE_DEBUG
std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif
  return (*curr).second;
}


void Acom::unbind_network( const std::string& name )
    throw (InvalidArgument)
{
#ifdef ACOM_TRACE_DEBUG
std::cout << "In " << __PRETTY_FUNCTION__ << ", name=" << name << std::endl;
#endif
  AllNetworks::const_iterator curr;
  curr = all_networks.find(name);
  if (curr== all_networks.end()) 
  {
    throwit( InvalidArgument("[Acom::unbind] name do not exist") );
  }
  all_networks.erase(name);  
#ifdef ACOM_TRACE_DEBUG
std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif
}



// ----------------------------------------------------------------------
const std::string& Acom::get_canonical_name() 
{ return self_canonical_name; }



// ----------------------------------------------------------------------
pid_t Acom::get_pid()
{ return my_pid; }



// ----------------------------------------------------------------------
// The size of a page
size_t Acom::get_pagesize()
{
#ifndef WIN32
  return getpagesize();
#else
  SYSTEM_INFO info;
  GetSystemInfo(&info);
  return info.dwPageSize;
#endif
}


// ----------------------------------------------------------------------
/** Return the default network
*/
Network* Acom::get_default_network( )
{
  return defaultnet;
}


} // - namespace
