 // ==========================================================================
// $Source: /home/cvs/Taktuk2/Psocket/src/sock_addr.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file for more information
// Author: J Briat T.Gautier
// Header file for Inuktitut/Thread/Pthread
//
//
//
// ==========================================================================
#include <errno.h>
// for memcpy
#include <string.h>
//for gethostbyname, gethostbyaddr,
#include <netdb.h>
//for gethostname
#include <unistd.h>
//for sprintf
#include <stdio.h>
//for inet_pton
#include <sys/types.h>
#include <sys/socket.h>
// for bzero
#include <strings.h>

//for inet_ntoa for htonl
#include <netinet/in.h>
#include <arpa/inet.h>

//for ostream
#include <iostream>

#if SOCK_USE_DARWIN
// for htonl
#include <machine/endian.h>
#endif

#include "sock_addr.h"

namespace SOCK {

// ==========================================================================
/*! Constructor, no initialization
*/
Address::Address()
{
  _valid = false;
  _lastError = 0;
  _host = "";
  _port = 0;
  ::bzero(&_sockaddr,sizeof(_sockaddr));
#ifdef SOCK_USE_LINUX
  _sockaddr.sin_family = AF_INET;
  _sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  _sockaddr.sin_port=htons(0);
//  if(inet_pton(AF_INET,"0.0.0.0",&_sockaddr.sin_addr)==-1){
//    _lastError = errno;
    _valid = true;
//  }
#endif
}

// ==========================================================================
/*! copy-constructor,initialize with specified address
*/
Address::Address(const Address& addr)
{
  *this = addr;
}


// ==========================================================================
/**
Destructor, ?
*/
Address::~Address(){
}


// ==========================================================================
/*! copy-constructor,initialize with specified address
*/
Address& Address::operator=(const Address& addr)
{
  _valid = addr._valid;
  _lastError = addr._lastError;
  _host = addr._host;
  _port = addr._port;
  _sockaddr = addr._sockaddr;
  return *this;
}


// ==========================================================================
/*! Constructor. initialize with specified host
*/
int Address::getFamily() const
{
  if (_valid)
    return _sockaddr.sin_family;
  return 0;
}


// ==========================================================================
/*! initialize with host
*  \return reference to initialized object
*/
Address& Address::setByName( const std::string& host )
{
//std::cout << "IN " << __PRETTY_FUNCTION__ << std::endl;
  _host = host;
  hostent* hp;
  hp = gethostbyname(_host.c_str());
  if (hp == 0) {
//std::cout << "Host=" << host << " no found" << std::endl;
    // not found, get error
    _lastError = h_errno;
    _valid = false;
    bzero(&_sockaddr,sizeof(_sockaddr));
    return *this;
  }
  _valid = true;
  // copy address
  _sockaddr.sin_family = AF_INET;
  memcpy(&_sockaddr.sin_addr, hp->h_addr_list[0], sizeof(struct in_addr));
//std::cout << "OUT " << __PRETTY_FUNCTION__ << std::endl;
  return *this;
}


// ==========================================================================
/** return the address of host given by name
 *  \return reference to initialized object
 */
Address Address::getByName( const std::string& host )
{
  Address retval;
  retval.setByName( host );
  return retval;
}


// ==========================================================================
/*! get hostname of current object
*  \return _host
*/
const std::string& Address::getHostName() const
{
  return _host;
}


// ==========================================================================
/*! get address object for localhost
 *  \return Address object for localhost
 */
const Address& Address::getLocalHost()
{
  static Address localaddr;
  static bool localaddr_ok = false;
  if (localaddr_ok)
    return localaddr;

  std::string hostname;
  char buffer[1024];
  if ( gethostname( buffer, 1024 ) != 0) {
    // gethostname failed, use localhost
//std::cout << "Invalid hostname is=" << buffer << std::endl;
    hostname = "localhost";
  } else {

//std::cout << "Hostname is=" << buffer << std::endl;
   hostname = buffer;
  }
  localaddr.setByName( hostname );
  localaddr_ok =true;
  return localaddr;
}


// ==========================================================================
/** return the adress (ip:port)
 *  \return as string
 */
std::string Address::getURL() const
{
  if (!_valid) return "<invalid address>";
  char tmp[1024];
  sprintf(tmp,"%s:%i", inet_ntoa( _sockaddr.sin_addr ) , _port );
  return std::string(tmp);
}

// ==========================================================================
/*! return if object/ip is valid.
*  \return true if host/IP is valid
*  \warning if isValid() returns false you can't rely on the results of
*           other get functions in Address
*/
bool Address::isValid() const
{
    return _valid;
}


// ==========================================================================
/*! return description of last error
*  \return error as string
*/
std::string  Address::getLastError() const
{
#if defined(SOCK_USE_LINUX) || defined(SOCK_USE_DARWIN)
    return std::string(hstrerror(_lastError));
#elif defined(SOCK_USE_SUNOS)
    return std::string("Error");
#endif
}


// ==========================================================================
int Address::getLastErrorNo() const
{
  return _lastError;
}


// ==========================================================================
/*! return host address information
*  \return sockaddr_in struct (system dependent structure to store address)
*  \warning this is data inside the class, be carefull not to overwrite !
*/
const sockaddr_in* Address::getSockAddr_in() const
{ return &_sockaddr; }
sockaddr_in* Address::getSockAddr_in() 
{ return &_sockaddr; }

const sockaddr* Address::getSockAddr() const
{ return (const sockaddr*)&_sockaddr; }
sockaddr* Address::getSockAddr() 
{ return (sockaddr*)&_sockaddr; }

// ==========================================================================
/*! set port of address
*  \warning changing the port after connecting will not change the port your
*           connection is on, but modify data wich will lead to inconsistency
*/
void Address::setPort(int p)
{
  _port=p;
  //this->_port = _port & 0xffff;  // valid port range between 0 .. 65535
  _sockaddr.sin_port = htons(this->_port);
}


// ==========================================================================
/*! return port number of address
*  \return _port
*/
int Address::getPort() const
{
  return _port;
}



// ==========================================================================
/**
*/
std::ostream& Address::print(std::ostream& o ) const
{
  o << "[host=" << _host
  	<< ", valid= " << (_valid ? "true" : "false" )
  	<< ", port= " << _port
  	<< ", sin_family= " << _sockaddr.sin_family
  	<< ", sin_port= " << ntohs(_sockaddr.sin_port)
  	<< ", sin_addr= " << inet_ntoa(_sockaddr.sin_addr)
  	<< "]"<<std::endl; 
  return o;
}



// ==========================================================================
/*! return the size of address
* \ return size
*/
int Address::getSockAddrSize() const
{
  return sizeof(sockaddr_in);
}


}//namespace SOCK
