 // ==========================================================================
// $Source: /home/cvs/Taktuk2/Psocket/src/sock_basic.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file for more information
// Author: J Briat T.Gautier
// Header file for Inuktitut/Thread/Pthread
//
//
//
// ==========================================================================

#include "sock_basic.h"
#include <cstdlib>
#include <iostream>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
//for fcntl
#include <fcntl.h>
//for TCP_NODELAY
#include <netinet/in.h>
//for close
#include <unistd.h>

namespace SOCK {

/**
*! Constructor
*/
Socket::Socket()
 : _addr(), _fd(-1), _valid(false)
{
}

  /** Assignment operator
  */
Socket& Socket::operator=(const Socket& s)
{ 
  return *this; 
}



  /*! close socket/disconnect from server
  * \return zero if socket was closed properly, -1 on error
  */
int Socket::close(){
  _valid = false;
  return ::close(_fd);
}

/** initialize from already existing socket  
 *  \arg fd valid file-descriptor
 *  \warning initialization may fail, check isConnected() afterwards
 *  \warning using this constructor will result in loss of the socket origin!
 */
void Socket::set_fd( int fd )
{
  if (fd ==-1)
    _valid = false;
  else {
    _valid = true;
    _fd = fd;
  }
}


/*! permit to manipulate the socket options TCP_NODELAY, FD_CLOEXEC, SO_REUSEADDR
* \arg opt option (no_delay, clo_exec, reuseaddr)
* \arg value
*/
int Socket::setSockOpt(int opt,int value )
{
  int flag;
  switch(opt){
    case no_delay :
      flag = 1;
      if (setsockopt(_fd,IPPROTO_TCP,TCP_NODELAY, &flag,sizeof(flag)) ==-1)
        return -1;
      else return 0;
      flag =fcntl(_fd,F_GETFL,0);
      if (flag < 0)
        return flag;
      if (value !=0)
        flag |= O_NONBLOCK;
      else
        flag &= ~O_NONBLOCK;
      return fcntl(_fd, F_SETFL, flag);

    case clo_exec :
      flag =fcntl(_fd,F_GETFD,0);
      if (flag < 0)
        return flag;
      if (value !=0)
        flag |= FD_CLOEXEC;
      else
        flag &= ~FD_CLOEXEC;
      return fcntl(_fd, F_SETFD, FD_CLOEXEC);

    case reuseaddr :
      int s = setsockopt(_fd,SOL_SOCKET,SO_REUSEADDR,&value,sizeof(value));
      return s;
  }
  return -1;
}


std::ostream& Socket::print( std::ostream& o ) const
{
  o << "[@=" << _addr
    << ", file descr=" << _fd
    << ", " << (_valid ? "valid" : "no valid")
    << "]"<<std::endl; 
  return o;
}

char* Socket::c_str() {
  char * buf_fd;
  int toto=0;
  buf_fd =(char *) malloc(sizeof(char)*256);
  if (_valid)
    toto= sprintf(buf_fd,"file descriptor = %d\n valid  = true\n",_fd);
  else
    toto = sprintf(buf_fd,"file descriptor = %d\n valid = false\n",_fd);
  buf_fd[toto] = 0;
  return buf_fd;
}


bool Socket::valid() const {
  return _valid;
}

/*! \class Socket
 *  \brief a Socket is an endpoint for communication
 *
 * the class Socket encapulates the socket, adding functions for
 * connection, writing, reading, checking for data
 */
}//namespace SOCK
