// ==========================================================================
// $Source: /home/cvs/Taktuk2/Psocket/src/sock_client.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file for more information
// Author: J Briat T.Gautier
// Header file for Inuktitut/Thread/Pthread
//
//
//
// ==========================================================================
#include <iostream>
#include <errno.h>
#include <inuk/pth>
// for perror
#include <stdio.h>
//for close
#include <unistd.h>

// for writev readv
#include <sys/uio.h>

// AF_LOCAL
#include <sys/socket.h>
#include <sys/un.h>

#include <sys/types.h>

#include <netinet/in.h>
//for hostent
#include <netdb.h>

#include <string>
#include "sock_client.h"
#include "sock_addr.h"
#include "inuk/taktuk/debuglog.h"

namespace SOCK{

//========================================================================
/*! connect to server on specified address and port.
 *  \arg addr host address to connect to
 *  \return zero if connect was succesfull, -1 else
 */
int SocketClient::connect()
{
//std::cout << "IN " << __PRETTY_FUNCTION__ << std::endl;
//std::cout << "addr =" << _addr << std::endl;
  if (_fd == -1)
  {
    _fd = ::socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
    if (_fd == -1)
      std::cout << "NEW SOCKET: fd  = " << _fd << std::endl;
    else
      _valid = true;
  }
  int connect_result = ::connect(_fd, _addr.getSockAddr(), _addr.getSockAddrSize());
  if (connect_result == -1){
    _addr._lastError=errno;
    //::perror("connect::");
  }
  _valid = true;
//std::cout << " Address=" << getAddress() << std::endl;
//std::cout << " URL=" << getURL() << std::endl;
//std::cout << "OUT " << __PRETTY_FUNCTION__ << ", this=" << *this 
//          << std::endl << std::endl;
  return connect_result;
}

int SocketClient::connect(const Address& addr)
{
  setAddress( addr );
  return connect();
}

//========================================================================
/*! connect to server on specified port
 *  \arg host hostname to connect to
 *  \arg port port to connect on
 *  \return zero if connect was succesfull, -1 else
 */
int SocketClient::connect(const std::string& host,int port)
{
//std::cout << "IN " << __PRETTY_FUNCTION__ << std::endl;
  Address addr;
  addr.setByName(host);
  addr.setPort(port);
//std::cout << "addr =" << addr << std::endl;
  return connect( addr );
//std::cout << "OUT " << __PRETTY_FUNCTION__ << std::endl;
}



//========================================================================
/*! write/send data to socket
 *  \arg buf data to be send
 *  \arg size length of bytes to be written
 *  \return number of bytes actually written
 */
int SocketClient::writen(const void* buf, int size)
{
  //if (!_valid) return -1;
  const char* buf_inter=(const char*)buf;
  int nbsent=0;
  int nbtosent =size;
  while(nbtosent!=0)
  {
    nbsent = ::send(_fd, buf_inter, nbtosent, MSG_DONTROUTE);
    if((nbsent==-1&&errno!=EINTR)||nbsent==0)
    {
     	break;
    }
    buf_inter =buf_inter + (nbsent*sizeof(char));
    nbtosent  -= nbsent;
  }
  if(nbsent <= 0 && nbtosent != 0)
  {
    //connection lost
    _valid = false;
		if(nbsent==0)
			dbg.log("SOCKET","CLOSED on WRITEN on fd %d\n",_fd);
		if(nbsent==-1)
			dbg.log("SOCKET","ERROR on WRITEN on fd %d, error: %s\n",_fd,strerror(errno));
//std::cout << ", error !" << std::endl;
    return -1;
  }
//std::cout << ", ok !" << std::endl;
  return size;
}

int SocketClient::write(const void* buf, int size)
{
  if (!_valid) return -1;
	int ssize=0;
	bool eintr=false;
	do
	{
		eintr=false;
  	ssize = ::write(_fd, buf, size);
  	if (ssize ==-1&&errno==EINTR)
 		{
        eintr=true;
  	}
	}while(eintr);
	if(ssize==0)
		dbg.log("SOCKET","CLOSED on WRITE on fd %d\n",_fd);
	if(ssize==-1)
		dbg.log("SOCKET","ERROR on WRITE on fd %d, error: %s\n",_fd,strerror(errno));
  return ssize;
}


//========================================================================
/*! write/send data to socket
 *  \arg iov data to be send
 *  \arg count number of bytes to be written
 *  \return number of bytes actually written
 */
int SocketClient::writev(const iov_entry* iov,int count)
{
  if (!_valid) return -1;
  void* buf_inter;
  int nbsent=0;
  int nbtosent =count;
  while (nbtosent!=0) {
    buf_inter= (char*)buf_inter+nbsent;
    nbsent = ::writev(_fd, (iovec *) iov, count);
    if (nbsent == -1)
      _valid = false;
    nbtosent-=nbsent;
    if (nbsent<=0) {
      _valid = false;
      break;
    }
  }
  if (nbsent <= 0 && nbtosent != 0) {
    //connection lost
    return -1;
  }
  return count;
}

//========================================================================
/*! read/receive data from socket.
 *  \arg iov pointer where data will be stored
 *  \arg count maximum number of bytes that will be read
 *  \return number of bytes actually read (may be less than len)
 */
int SocketClient::readv(iov_entry* iov,int count)
{

        if (! _valid)
                return -1;
        void* buf_inter;
        int nbread=0;
        int nbtoread =count;
        while(nbtoread!=0)
        {
                buf_inter= (char*)buf_inter+ nbread;
                nbread = ::readv(_fd,(iovec *) iov, count);
                nbtoread-=nbread;
                if(nbread<=0)
                {
                        break;
                }
        }
        if(nbread <= 0 && nbtoread != 0)
        {
      //connection lost
                _valid = false;
                return -1;
        }
        else
                return count;


        /*int ssize = ::readv(_fd,(iovec*) iov,count);
        if (ssize == -1) {
                _addr._lastError=errno;

        }
        return ssize;*/
}

//========================================================================
/*! read/receive data from socket.
 *  \arg buf pointer where data will be stored
 *  \arg size maximum number of bytes that will be read
 *  \return -1 if the number of bytes read is different from size
 */
int SocketClient::readn(void* buf, int size)
{
  if (!_valid) return -1;
  char* buf_inter= (char*)buf;
  int nbread=0;
  int nbtoread =size;
//std::cout << counter << "::read[" << _fd << "]: len=" << size << std::endl;
  while(nbtoread>0)
  {
    int szread = ( nbtoread < SSIZE_MAX ? nbtoread : SSIZE_MAX );
    nbread = ::read(_fd, buf_inter, szread);
    //nbread = ::recv(_fd, buf_inter, nbtoread, MSG_WAITALL);
    if((nbread==-1&&errno!=EINTR)||nbread==0)
    {
      break;
    }
    buf_inter = buf_inter + (nbread*sizeof(char));
    nbtoread  -= nbread;
  }
  if(nbread <= 0 && nbtoread > 0)
  {
    //connection lost
    _addr._lastError = errno;
    _valid = false;
		if(nbread==0)
			dbg.log("SOCKET","CLOSED on READN on fd %d\n",_fd);
		if(nbread==-1)
			dbg.log("SOCKET","ERROR on READN on fd %d, error: %s\n",_fd,strerror(errno));
//std::cout << "Lost connection, error errno=" << errno << std::endl;
    return -1;
  }
//std::cout << ", ok !" << std::endl;
  return size;
}

/*! read a line terminated by carrier return
	*  \arg reference to a string string read will be added to it
	*  \return -1 if connection closed , else size read
	*/
int SocketClient::readline(std::string& buf)
{
	char cur='t'; // to pass the first loop
	int nbread=0;
	if (!_valid) return -1;
	while(_valid && cur!='\n')
	{
		if(read(&cur,1)>0)
		{
			buf+=cur;
			nbread++;
		}
	}
	return (_valid?nbread:-1);
}

int SocketClient::read(void* buf, int size)
{
	bool eintr=false;
	int ssize=0;
  if (!_valid) return -1;
	do
	{
		eintr=false;
  	ssize = ::read(_fd, buf, size);
    if (ssize == 0) {
       // _addr._lastError=errno;
       _valid = false;
    }
		if (ssize==-1)
		{
			if(errno==EINTR)
				eintr=true;
			else
				_valid=false;
		}
	}while(eintr);
	if(ssize==0)
		dbg.log("SOCKET","CLOSED on READ on fd %d\n",_fd);
	if(ssize==-1)
		dbg.log("SOCKET","ERROR on READ on fd %d, error: %s\n",_fd,strerror(errno));
  return ssize;
}


//========================================================================
/*! create a copy of the socket
 * \return the copy
 */
SocketClient* SocketClient::dup()
{
  if(_valid)
  {
  	int fd =  ::dup(_fd);
  	if (fd == -1)
	  {
	    _addr._lastError=errno;
	    return NULL;
	  }
	  else {
      SocketClient* sc = new SocketClient;
      sc->set_fd(fd);
      return sc;
    }
  }
  else
    return NULL;
  
}

//========================================================================
/*! create a copy of the socket
 * \arg new_sock  new_sock is the copy of the socket
 * \return the copy
 */
void SocketClient::dup2( SocketClient& new_sock)
{
  int  fd = ::dup2(_fd,new_sock._fd);
  if (fd == -1)
        _addr._lastError=errno;
  else
        new_sock._valid = true;
}

/** returns the address of the socket that SOCKET is connected to
  */
int SocketClient::getPeerName (Address& addr) const
{
#if defined(SOCK_USE_LINUX) || defined(SOCK_USE_SUNOS) || defined(INUKTITUT_USE_LINUX_IA64)
  socklen_t length;
#elif defined(SOCK_USE_DARWIN) || defined(SOCK_USE_IRIX)
  int length;
#endif
  length = addr.getSockAddrSize();
  return ::getpeername (_fd, addr.getSockAddr(), &length);
}


//========================================================================
/*! connect a pair of sockets
 * \return zero on succes, -1 else
 */
int SocketClient::sock_pair(SocketClient* one, SocketClient* two)
{
  int sv[2];
#if defined(SOCK_USE_LINUX) || defined(SOCK_USE_DARWIN) || defined(SOCK_USE_IRIX) || defined(INUKTITUT_USE_LINUX_IA64)
  int sp = ::socketpair( AF_LOCAL,SOCK_STREAM,0,sv);
#elif defined(SOCK_USE_SUNOS)
  int sp = ::socketpair( AF_INET,SOCK_STREAM,0,sv);
#endif
  one->set_fd(sv[0]);
  one->_valid=true;
  two->set_fd(sv[1]);
  two->_valid=true;
  return sp;
}

//========================================================================
/*! connect a pair of sockets
 * \return zero on succes, -1 else
 */
int SocketClient::sock_pair
  (int d, int type, int proto, SocketClient* one, SocketClient* two)
{
  int sv[2];
  int sp = ::socketpair(d,type,proto, sv);
  one->_fd=sv[0];
  one->_valid=true;
  two->_fd=sv[1];
  two->_valid=true;
  return sp;
}


}//namespace SOCK
