 // ==========================================================================
// $Source: /home/cvs/Taktuk2/Psocket/src/sock_select.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file for more information
// Author: J Briat T.Gautier
// Header file for Inuktitut/Thread/Pthread
// ==========================================================================
/* According to POSIX 1003.1-2001 */

#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#ifdef SOCK_USE_LINUX
#include <sys/select.h>
#endif


#include "sock_select.h"
#include "sock_client.h"
// ==========================================================================


namespace SOCK{
/*! wait for a number of FDSet to change status
 * \return the number of FDSet before the timer expires. On error, -1 is returned
 */
int select(FDSet& readset, FDSet& writeset, FDSet& exeptset)
{
/*        int n = 0;
        errno=0;
        if (readset._fdmax > writeset._fdmax)
                n = readset._fdmax;
        else
                n = writeset._fdmax;
        if (n < exeptset._fdmax)
                n = exeptset._fdmax;
        struct timeval *timeout;
        int  select_result = 0;
        do
        {
        select_result = ::select(n+1,&readset._set,&writeset._set,&exeptset._set, NULL);
        }while(errno == EINTR);
        //if (select_result == -1)_addr.lastError=errno;
        return select_result;
    */
    return select(readset,writeset,exeptset,NULL);
}
// ==========================================================================
/*! wait for a number of FDSet to change status
 * \return the number of FDSet before the timer expires. On error, -1 is returned
 */
int select(FDSet& readset, FDSet& writeset, FDSet& exeptset, struct timeval* Tout)
{
  int n = 0;
  errno=0;
  if (readset._fdmax > writeset._fdmax)
    n = readset._fdmax;
  else
    n = writeset._fdmax;
  if (n < exeptset._fdmax)
    n = exeptset._fdmax;
  int  select_result=0;
  do
  {
    select_result= ::select(n+1,  readset._fdmax==0?NULL:&readset._set,
                                  writeset._fdmax==0 ? NULL : &writeset._set,
                                  exeptset._fdmax==0? NULL : &exeptset._set, Tout );
    
  } while(errno == EINTR);
  //if (select_result == -1)_addr.lastError=errno;
  select_result=errno;
  return select_result;
}
// ==========================================================================
/*! Constructor
*/
FDSet::FDSet()
{
        _fdmax = 0;
}
// ==========================================================================
/*! Destructor
*/
FDSet::~FDSet(){
}

// ==========================================================================
/*! add a given descriptor from a set
 *
 */
void FDSet::set(SocketClient* s){
        FD_SET(s->get_fd(),&_set);
        if (s->get_fd()>_fdmax)
                _fdmax = s->get_fd();

}

// ==========================================================================
/*! remove a given descriptor from a set
*/
void FDSet::clr(SocketClient* s){
  FD_CLR(s->get_fd(),&_set);
}

// ==========================================================================
/*! clear a set
*/
void FDSet::zero(){
  FD_ZERO(&_set);
}

// ==========================================================================
/*! tests if the descriptor of a socket is part of the set
* \return nonzero value if the descriptor is part of the set, and zero otherwise
*/
int FDSet::isready(SocketClient* s)
{
  return FD_ISSET(s->get_fd(),& _set);
}

}//namespace SOCK
