// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_pthreadsync.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Authors: J. Briat & T. Gautier 
// $Id: pth_pthreadsync.cpp,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
//
// ==========================================================================
// the namespace identifier
#include "pth_ns.h"
#include "pth_pthreadsync.h"
#include "pth_system.h" 

#ifdef PTH_USE_TRACER
#include "Pthread_stub_tracer.h"
#endif

namespace PTH
{

pthread_mutexattr_t NTMutex::default_attr;
pthread_mutexattr_t NTMutex::recursive_attr;

void Mutex::init_static()throw (PosixError)
{
  int err;
  static bool once=true;if(once)once=false;else return;
  // default attribute for Mutex
  ASSERT(!(err =pthread_mutexattr_init(&default_attr)), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
  ASSERT(!(err =pthread_mutexattr_init(&recursive_attr)), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
#if !defined(PTH_USE_PTHREAD_LINUX_IA64)
  ASSERT(!(err =pthread_mutexattr_settype(&default_attr,MUTEX_TYPE_NORMAL )), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
  ASSERT(!(err =pthread_mutexattr_settype(&recursive_attr, MUTEX_TYPE_RECURSIVE )), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
#endif
};

pthread_condattr_t  Condition::default_attr;

void Condition::init_static()throw (PosixError)
{
  int err;
  static bool once=true;if(once)once=false;else return;
  // default attribute for Condition
  ASSERT(!(err =pthread_condattr_init(&default_attr)), 
      PosixError(SUFX( "[Condition:: Init]: cannot define condition attributes "),err));

#ifdef PTH_USE_PTHREAD_IRIX
  ASSERT(!(err =pthread_condattr_setpshared(&default_attr, COND_TYPE )), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
#endif

  
};



pthread_attr_t Thread::detached_attr;
pthread_attr_t Thread::joinable_attr;

Thread::Kind Thread::DETACHED;
Thread::Kind Thread::JOINABLE;      

void Thread::init_static()throw (PosixError)
{
  int err;
  static bool once=true;if(once)once=false;else return;

#ifdef PTH_USE_PTHREAD_IRIX
  ASSERT(!( err= pthread_setconcurrency(256)),
    PosixError(SUFX( "[Pthread:: Init]: cannot set concurrency ") ,err));
#endif

  // -- initialize detached attribut for thread creation
  ASSERT(!( err= pthread_attr_init(&detached_attr)),
      PosixError(SUFX( "[Pthread:: Init]: cannot define   pthread attributes ") ,err));
  ASSERT(!( err= pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED)), 
      PosixError(SUFX( "[Pthread:: Init]: cannot define  pthread attributes "), err));


  // -- initialize joinable attribut for thread creation
  ASSERT(!( err= pthread_attr_init(&joinable_attr)),
      PosixError(SUFX( "[Pthread:: Init]: cannot define   pthread attributes ") ,err));
  ASSERT(!( err= pthread_attr_setdetachstate(&joinable_attr, PTHREAD_CREATE_JOINABLE)), 
      PosixError(SUFX( "[Pthread:: Init]: cannot define  pthread attributes "), err));

#ifdef PTH_USE_PTHREAD_IRIX
  // 1MByte / per stack on IRIX
  ASSERT(!( err= pthread_attr_setstacksize(&detached_attr, 2 * 1024*1024)),
    PosixError(SUFX( "[Pthread:: Init]: cannot set concurrency ") ,err));
//  ASSERT(!( err= pthread_attr_setstacksize(&joinable_attr, 1024*1024)),
//    PosixError(SUFX( "[Pthread:: Init]: cannot set concurrency ") ,err));
#endif

  DETACHED= &detached_attr; JOINABLE= &joinable_attr;


}

} // -namespace
