// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_timer.cpp,v $
// Copyright(c)'94-97 by Givaro Team
// see the copyright file.
// Authors: T. Gautier
// Based on Givaro's givtimer.C file.  
// review: [3/3/98]
//
//
//
// ==========================================================================
#include <pth_timer.h>
#include <pth_error.h>
#include <iomanip>

extern "C" {
# include <sys/time.h>
# include <sys/resource.h>
# include <stdio.h>
# include <unistd.h>
#if defined( INUKTITUT_USE_DARWIN )
#include <sys/types.h>
#include <sys/sysctl.h>
#endif
}


namespace PTH {

// --------------------------------------------------------------------
const std::string& HighResTimer::unit()
{
  static std::string u = "s";
  return u;
}

// --------------------------------------------------------------------
const std::string& WallTimer::unit()
{
  static std::string u    = "s";
  return u;
}

// --------------------------------------------------------------------
const std::string& CpuTimer::unit()
{
  static std::string u     = "s";
  return u;
}

// --------------------------------------------------------------------
const std::string& SysTimer::unit()
{
  static std::string u     = "s";
  return u;
}

// --------------------------------------------------------------------
const std::string& LogicalTimer::unit()
{
  static std::string u = "#";
  return u;
}


// --------------------------------------------------------------------
unsigned long long HighResTimer::latency =0;
double HighResTimer::dtick_per_s =0;




// --------------------------------------------------------------------
//#define USE_SOFT_CALIBRATION

void HighResTimer::calibrate()
{
#ifdef USE_SOFT_CALIBRATION
  size_t count;
  double t00, t11;
  HighResTimer::type hrt, hrt0, hrt1, hrt2;
#  if defined( INUKTITUT_USE_DARWIN )
  asm volatile ("mftbu %0" : "=r" (hrt0.u.s.t_u) );
  asm volatile ("mftb %0" : "=r" (hrt0.u.s.t_l) );
#  elif defined( INUKTITUT_USE_LINUX )
  rdtsc( hrt0.u.s.t_l, hrt0.u.s.t_u );
#  endif
  /* loop 1 */
  for (int i=0; i<32; ++i)
    t00 = WallTimer::gettime();
#  if defined( INUKTITUT_USE_DARWIN )
  asm volatile ("mftbu %0" : "=r" (hrt1.u.s.t_u) );
  asm volatile ("mftb %0" : "=r" (hrt1.u.s.t_l) );
#  elif defined(INUKTITUT_USE_LINUX)
  rdtsc( hrt1.u.s.t_l, hrt1.u.s.t_u );
#  endif
  /* loop 2 */
  t00 = WallTimer::gettime();
#if defined( INUKTITUT_USE_DARWIN )
  for (int i=0; i<2000; ++i)
#else
  for (int i=0; i<20000; ++i)
#endif
    for (int j=0; j<100000; ++j)
      count++;
  t11 = WallTimer::gettime();
#  if defined( INUKTITUT_USE_DARWIN )
  asm volatile ("mftbu %0" : "=r" (hrt2.u.s.t_u) );
  asm volatile ("mftb %0" : "=r" (hrt2.u.s.t_l) );
#  elif defined( INUKTITUT_USE_LINUX )
  rdtsc( hrt2.u.s.t_l, hrt2.u.s.t_u );
#  endif

  /* ticks for 32 call to gettimeofday -> hrt0 : loop 1 */
  hrt0.u.t = hrt1.u.t - hrt0.u.t;

  /* ticks to the loops - 2 gettimeofday -> hrt1 : loop 2 */
  hrt1.u.t = hrt2.u.t - hrt1.u.t - hrt0.u.t / 16;

//std::cout << "Delay (loop1): " << hrt1.u.t << std::endl;
//std::cout << "Delay (gettimeofday): " << hrt0.u.t << std::endl;

  /* ticks for loop */
  double d = double(hrt1.u.t);
//  std::cout << "Tick per gettimeofday=" << double(hrt0.u.t) / 32.0 << std::endl;
  HighResTimer::dtick_per_s = d / (t11-t00);
//  std::cout << "Tick per sec=" << HighResTimer::dtick_per_s << std::endl;
#else // NO SOFT_CALIBRATION

#if defined( INUKTITUT_USE_DARWIN )
  int mib[2];
  size_t len;
  int mhz;
  mib[0] = CTL_HW;
  mib[1] = HW_CPU_FREQ;
  len =sizeof(int);
  sysctl(mib, 2, &mhz, &len, NULL, 0);
//  std::cout << "freq:" << mhz << std::endl;
  HighResTimer::dtick_per_s = mhz*1e6*268.07020875; /* it seems that tick freq is about 2/3e9 of the frequency on PowerbOOK*/
#elif defined( INUKTITUT_USE_IRIX )
  struct timespec tp; 
  clock_getres( CLOCK_SGI_CYCLE, &tp );
//  std::cout << "SGI res: " << tp.tv_sec << ", " << tp.tv_nsec << std::endl;
  HighResTimer::dtick_per_s = double(tp.tv_sec) + double(tp.tv_nsec)*1e-9;
  HighResTimer::dtick_per_s = 1/dtick_per_s;
#elif defined( INUKTITUT_USE_LINUX )
  /* open cpuinfo and scan cpu Mhz: */
  FILE* file = fopen("/proc/cpuinfo","r");
  double mhz = 1.0;
  PTH_ASSERT(file !=0, "Cannot open /proc/cpuinfo");

  int retval =0;
  char line[256];
  do {
    retval = fscanf(file,"%[^:]:%lf\n",line,&mhz);
//std::cout << "result fscanf=" << retval << "str=" << line << ", mhz=" << mhz << std::endl;
    if ( strncmp(line,"cpu MHz", 7) ==0) break; 
/*
     if (retval ==0) {
       getline( &line, 256, file );
     }
*/
  } while (retval != EOF);
  HighResTimer::dtick_per_s = mhz*1e6;
  fclose(file);
  //std::cout << "Tick per sec=" << HighResTimer::dtick_per_s << std::endl;
#else
#endif

#endif // END CALIBRATION

  /** Compute latency for call to start/stop */
  long tick_loop;
  unsigned long long t0, t1;
  t0 = HighResTimer::gettick();
  for (int i=0; i<100; ++i) 
    ;
  t1 = HighResTimer::gettick();
  tick_loop = t1-t0;
//std::cout << "Tick loop=" << tick_loop << std::endl;

  HighResTimer hr;
  t0 = HighResTimer::gettick();
  for (int i=0; i<100; ++i) 
  { hr.start(); hr.stop(); } 
  t1 = HighResTimer::gettick();
//std::cout << "Tick loop+start+stop=" << t1-t0 << std::endl;

  HighResTimer::latency = (t1-t0 - tick_loop)/100;
#if 0
  {
    std::cout << "Ticks per second  :" << std::setprecision(10) << HighResTimer::dtick_per_s << std::endl;
    std::cout << "Latency start/stop:" << std::setprecision(10) << HighResTimer::latency << std::endl;
  }
#endif
}



// --------------------------------------------------------------------
long Timer::seed() 
{
  struct timeval tp;
  gettimeofday(&tp, 0) ;
  return(tp.tv_usec+tp.tv_sec);
}



// --------------------------------------------------------------------
std::ostream& Timer::print( std::ostream& o ) const
{
  return 
   o << "Time (elapsed highres) s= " << hrrealtime() << std::endl
     << "Time (elapsed) s= " << realtime() << std::endl
     << "Time (cpu) s= " << usertime() << std::endl
     << "Time (sys) s= " << systime()  << std::endl
     << "#Ticks = " << tickcount() << std::endl
     << "#Call  = " << callcount() << std::endl;
}

} // -namespace
