/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 *
 *
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */
/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */
#include <inuk/sock>
#include <inuk/pth>
#include <inuk/acom>
#include <inuk/taktuk/CoreTCP.h>
#include <mmputenv.h>
#include <pthread.h>
#include <inuk/taktuk/debuglog.h>
#include <inuk/taktuk/panichandler.h>

int InfoTCPnet::env_export()
{
  char env[512];
  sprintf(env,"taktukrank=%d",rank );
  mmputenv(env);
  dbg.log("info","put env --> %s\n",env);
  sprintf(env,"taktukcount_nodes=%d",count_nodes );
  mmputenv(env);
  dbg.log("info","put env --> %s\n",env);
  sprintf(env,"taktukinfosize=%d",all_links.size() );
  mmputenv(env);
  dbg.log("info","put env --> %s\n",env);
  for(int i=0;i<all_links.size();i++)
    {
      sprintf(env,"taktukinfo_%d_fd=%d",i,all_links[i]->_sock->get_fd());
      mmputenv(env);
      dbg.log("info","put env --> %s\n",env);
      sprintf(env,"taktukinfo_%d_peer=%d",i,all_links[i]->_peer );
      mmputenv(env);
      dbg.log("info","put env --> %s\n",env);
      sprintf(env,"taktukinfo_%d_flag=%d",i,all_links[i]->_flag );
      mmputenv(env);
      dbg.log("info","put env --> %s\n",env);
    }
  for(int i=0;i<count_nodes;i++)
    {
      sprintf(env,"taktukinforoute_%d=%d",i,route_table[i]);
      mmputenv(env);
      dbg.log("info","put env --> %s\n",env);
    }
  return 1;
}

int InfoTCPnet::env_recovery()
{
  char* inter;
  char req[512];
  int all_links_size=0;

  if((inter=getenv("taktukinfosize"))!=NULL)
    {
      all_links_size=atoi(inter);
      dbg.log("info","get env all_links size = %d\n",all_links_size);
    }
  else
    return -1;
 
  if((inter=getenv("taktukrank"))!=NULL)
    {
      rank=atoi(inter);
      dbg.log("info","get env rank = %d\n",rank);
    }
  else
    return -1;

  if((inter=getenv("taktukcount_nodes"))!=NULL)
    {
      count_nodes=atoi(inter);
      dbg.log("info","get env count_node = %d\n",count_nodes);
    }
  else
    return -1;
  
  all_links.reserve(all_links_size);
  all_links.resize(all_links_size);

  for(int i=0;i<all_links_size;i++)
    {
      sprintf(req,"taktukinfo_%d_fd",i);
      if((inter=getenv(req))!=NULL)
			{
				all_links[i]=new link();
        all_links[i]->_sock=new SOCK::SocketClient();
	  		all_links[i]->_sock->set_fd(atoi(inter));
				all_links[i]->_mut=new PTH::Mutex();
				all_links[i]->_mut->init();
	  		dbg.log("info","get env fd of all_links[%d] = %d\n",i,atoi(inter));
			}
      else
        return -1;
      sprintf(req,"taktukinfo_%d_peer",i);
      if((inter=getenv(req))!=NULL)
			{
        all_links[i]->_peer=atoi(inter);
	  		dbg.log("info","get env _peer of all_links[%d] = %d\n",i,atoi(inter));
			}
      else
        return -1;
      sprintf(req,"taktukinfo_%d_flag",i);
      if((inter=getenv(req))!=NULL)
			{
        	all_links[i]->_flag=atoi(inter);
	  			dbg.log("info","get env flag of all_links[%d] = %d\n",i,atoi(inter));
			}
      else
        return -1;
    }
	route_table.resize(count_nodes);
	for(int i=0;i<count_nodes;i++)
	{
		sprintf(req,"taktukinforoute_%d",i);
		if((inter=getenv(req))!=NULL)
		{
			route_table[i]=atoi(inter);
			dbg.log("info","get env route_table[%d] = %d\n",i,route_table[i]);
		}
	}
  return 1;
}

routertcp::routertcp()
{
  //initialize some mutex variable
  PTH::Pth::initialize();
  PTH::Pth::commit();
  state_mutex.initnothrow();
  mutex_stop.initnothrow();
  mutex_waiting.initnothrow();
  cond_stop.initnothrow();
  cond_waiting.initnothrow();
  receive_buffer=(void*)malloc(MAXBUF);

  //init the socketpair to wake up the thread
  if(SOCK::sock_pair(&wakeupW,&wakeupR)<0)
    {
      //dbg.log("","ERROR router can't allocate socketpair:\n%s\n",strerror(errno));
    }
  else
    {
      wakeupW.setSockOpt(SOCK::Socket::clo_exec,1);
      wakeupW.setSockOpt(SOCK::Socket::no_delay,1);
      wakeupR.setSockOpt(SOCK::Socket::clo_exec,1);
    }

  state= stopping;
  thread_exist=false;
  //dbg.log("router","constructor init [OK]\n");

  // to launch the thread one
  once_control = 0;
}

routertcp::~routertcp()
{
  
  //dbg.log("router","destructor thread terminate [OK]\n");
	try
	{
	  state_mutex.destroy();
	  //pthread_mutex_destroy(&state_mutex);

	 	mutex_waiting.destroy();
		//pthread_mutex_destroy(&mutex_waiting);
		mutex_stop.destroy();
		//pthread_mutex_destroy(&mutex_stop);
		cond_waiting.destroy();
		//pthread_cond_destroy(&cond_waiting);
		cond_stop.destroy();
		//pthread_cond_destroy(&cond_stop);
        join.destroy();
	}catch(PTH::PosixError &e)
	{
		dbg.log("WARNING","Destructeur Routeur %s\n",e.what());
	}
	wakeupW.close();
	wakeupR.close();
	free(receive_buffer);
  //PTH::Pth::terminate();
  //dbg.log("router","destructor [OK]\n");
}

void thread_router (routertcp* router)
{
  router->routing();
  --(router->join);
}
int routertcp::initialize()
{
  //initialize(NULL);
  return 1;
}

/*
int routertcp::initialize(InfoTCPnet* info)
{
  this->info=info;
  return 1;
}
*/
int routertcp::commit()
{
  /*if(info==NULL)
    return -1;
		*/
  if(once_control==0)
    {
      //create the router thread
      int error=0;
      if((error=routing_thread.createnothrow(&routing_threadId,(void *(*)(void*))thread_router, this, PTH::Thread::JOINABLE)) != 0)
//      if((error=pthread_create(&routing_thread,NULL,(void *(*)(void*))thread_router, this)) != 0)
			{
	  		dbg.log("WARNING","module TCP router can't start his thread\n");
	  		return -1;
			}
      join.init(1);
      thread_exist=true;
      //waiting for routing thread stop ack
      wait_stop();
      once_control = 1;
   }
	return 0;
}

void routertcp::LoadInfo(InfoTCPnet& _info)
{

	int state = NET_off();
	dbg.log("","LOCK load info\n");
	_info.all_links_mut.lock();
  this->info=_info;
	_info.all_links_mut.unlock();
	dbg.log("","UNLOCK load info \n");
	dbg.log("setinfo","TCP router Load new info structure \n");
	for(int i=0;i<this->info.all_links.size();i++)
		dbg.log("setinfo","NEW INFO link %d ->> node %d %s state : %s \n",i,this->info.all_links[i]->_peer,this->info.all_links[i]->open()?" open ":" closed ",this->info.all_links[i]->_flag == InfoTCPnet::link::free?" FREE ":" NOT FREE " );
	if(state!=-1)
		NET_on();
}

void routertcp::RefreshInfo(InfoTCPnet& _info)
{
	
	dbg.log("","LOCK Refresch info\n");
	_info.all_links_mut.lock();
	this->info=info;
	_info.all_links_mut.unlock();
	dbg.log("","UNLOCK Refresch info\n");
	wakeupW.writen((void*)"W",1);
}

void routertcp::SetHandler(PanicHandler& h)
{
  _handler=h;
}

void routertcp::NET_on()
{
  set_router_state(running);
}

int routertcp::NET_off()
{
  if(!thread_exist || check_router_state()==stopped )
    return -1;

  set_router_state(stopping);
  wakeupW.writen((void*)"W",1);
  wait_stop();
  dbg.log("router","Thread ROUTER STOPPED\n");
  return 1;
}

void routertcp::kill()
{
  dbg.log("router","Thread ROUTER KILL\n");
  set_router_state(killed);
  wakeupW.writen((void*)"W",1);
  if(thread_exist)
    {
      join.wait();
      //PTH::Thread::join(routing_threadId,NULL);
      //pthread_join(routing_thread,NULL);
	  
      thread_exist=false;
    }
  dbg.log("router","Thread ROUTER KILLED\n");
}
/*
void routertcp::shutdown()
{

  dbg.log("router","Thread ROUTER STANDBY request\n");
	SignalTime.B();
  set_router_state(standby);
  wakeupW.writen((void*)"W",1);
  dbg.log("router","Thread ROUTER STANDBY request [OK]\n");
}
*/
void routertcp::terminate()
{
  dbg.log("router","Thread ROUTER STANDBY request\n");
	SignalTime.B();
  set_router_state(standby);
  wakeupW.writen((void*)"W",1);
  dbg.log("router","Thread ROUTER STANDBY request [OK]\n");
  if(thread_exist)
    {
      join.wait();
      PTH::Thread::join(routing_threadId,NULL);
      //pthread_join(routing_thread,NULL);
      //      join.wait();
      thread_exist=false;
    }
  dbg.log("router","Thread ROUTER TERMINATED [OK] Signal Time %f sec %s %s\n",SignalTime.Time_sec(),SignalTime.BOK()?"":"No Begin Time", SignalTime.EOK()?"":"No End Time");

}

void routertcp::waiting_to_restart()
{
  dbg.log("router","thread [WAITING state = stopped]\n");
  set_router_state(stopped);
  wait_run();
  dbg.log("router","thread [RESTARTING]\n");
}

void routertcp::wait_run()
{

  //waiting for running or standby
  state_mutex.locknothrow();
  //pthread_mutex_lock(&state_mutex);
  while (state==stopped)
    {
      cond_stop.waitnothrow(state_mutex);
      //pthread_cond_wait(&cond_stop,&state_mutex);
      dbg.log("router","wait_run after wake up state= %d\n",state);
    }
  state_mutex.unlocknothrow();
  //pthread_mutex_unlock(&state_mutex);
}

void routertcp::wait_stop()
{
  //waiting for routing thread stop ack
  state_mutex.locknothrow();
  //pthread_mutex_lock(&state_mutex);
  while (state != stopped)
    cond_stop.waitnothrow(state_mutex);
  //pthread_cond_wait(&cond_stop,&state_mutex);
  state_mutex.unlocknothrow();
  //pthread_mutex_unlock(&state_mutex);
}

void routertcp::set_router_state(short state)
{
  state_mutex.locknothrow();
  //pthread_mutex_lock(&state_mutex);
  this->state=state;
  cond_stop.signalnothrow();
  //pthread_cond_signal(&cond_stop);
  state_mutex.unlocknothrow();
  //pthread_mutex_unlock(&state_mutex);
  dbg.log("router","thread state set to [ %d ] with running=1, stopped=2, standby=4, alone=5\n",state);
}

int routertcp::check_router_state()
{
  int inter;
  dbg.log("router","check thread state!!! \n");

  state_mutex.locknothrow();
  //pthread_mutex_lock(&state_mutex);
  inter=state;
  state_mutex.unlocknothrow();
  //pthread_mutex_unlock(&state_mutex);
  dbg.log("router","check thread state = %d \nwith running=1, stopped=2, standby=4, alone=5\n",inter);
  return inter;
}

/*
InfoTCPnet* routertcp::set_InfoTCP(InfoTCPnet* info)
{
  InfoTCPnet* old_info;

  if(check_router_state()!=stopped)
    return NULL;
  old_info=this->info;
  this->info=info;
  //  dbg.log("router","set new topology [OK]\n");
  return old_info;
}
*/

void routertcp::routing()
{


  int nb_open = 0;

  char wakeupbuf;
  SOCK::FDSet fds_null;
  SOCK::FDSet fds_null2;
  fds_null.zero();
  fds_null.zero();

  dbg.log("thread_router","thread begin init\n");

  while(1)
    {

      switch(check_router_state())
			{
 				case running:
 	  			break;
 				case standby:
					SignalTime.E();
	  			dbg.log("thread_router","Thread ROUTER  standby..............\n");
 	  			break;
 				case alone:
	  			//dbg.log("PANIC:","routing thread alone PANIC\nCall PANIC HANDLER\n");
				  _handler.handle();
 	  			return;
 	  			break;
			 	case stopping:
			 	  //dbg.log("thread_router","stopping\n");
			 	  waiting_to_restart();
			 	  continue;
			 	  break;
			 	case killed:
			 	  return;
			 	  break;
			 	default:
			 	  return;
			}


			// read set init with the fd_net set and all local in,out,err set
      //      dbg.log("thread_router","thread connection set init\n");
      SOCK::FDSet read_set;
      //       fd_set read_set;
      int select_out;

      //reset read set & control variable
      read_set.zero();
      //       FD_ZERO(&read_set);
      read_set.set(&wakeupR);
      //       FD_SET(wakeup[1],&read_set);
      //        if(max_fd<=wakeupR.SOCK::Socket::get_fd())
      //  	max_fd = wakeupR.SOCK::Socket::get_fd();
      nb_open=0;

      //look for taktuk interconnect
      for(int i=0; i<info.all_links.size(); i++)
			{
			  if(info.all_links[i]->open())
	 	    {
		      //dbg.log("thread_router","set in fdset link number %d sock :%s \n",i,info.all_links[i]._sock->c_str());
					//dbg.log("thread_router","link %d flag	%s\n",i,info.all_links[i]._flag==InfoTCPnet::link::normal?"normal":"not normal");
	 	      read_set.set(info.all_links[i]->_sock);
		      //dont wait for the up node finish
	 	      if(info.all_links[i]->_flag==InfoTCPnet::link::normal)
						nb_open++;
	 	    }
 			}


//      dbg.log("thread_router","nb connection open %d\n",nb_open);
      if((nb_open <=0) && check_router_state()==standby)
			{
				SignalTime.E();
    		return;
			}

//      dbg.log("thread_router","perform select\n");
      //select step
      select_out = SOCK::select(read_set,fds_null,fds_null2);


      if(select_out < 0)
			{
//			  dbg.log("","THREAD ROUTER: ERROR in the select %s PANIC\n",strerror(errno));
			    _handler.handle();
			  return;
			  //continue;
			}
      //check the read socket set
//      dbg.log("thread_router","check connections set after select\n");
      //wakeup
      if (wakeupR.valid()&& read_set.isready(&wakeupR))
		 	{
//			  dbg.log("thread_router","select WAKE UP\n");
		 	  wakeupR.readn(&wakeupbuf,1);
		 	}

      //active message & write signal connections set

      for(int i=0; i<info.all_links.size(); i++)
		 	{
		 	  if(info.all_links[i]->open())
		 	    if(read_set.isready(info.all_links[i]->_sock))
	 	      {
//						dbg.log("thread_router","receive on the net connection number %d\n",i);

						if(process(info.all_links[i],i)<=0)
					  {
			 		    //connection have been lost during receiving data
	//				    dbg.log("thread_router","connection number %d have been lost during the process step\n",i);
					    if(info.all_links[i]->_flag==InfoTCPnet::link::critical)
				      {
								set_router_state(alone);
				      }
					    // one child crashed or lost his connection
		  			}

		//				dbg.log("thread_router","receive and process completed........\n");

 	      	}
 			}// for
    }// while
  	dbg.log("thread_router","thread router terminate...\n");
	}



