/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 * 
 * 
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#include <unistd.h>         // for close and write function
#include <sys/types.h>
#include <errno.h>
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <vector>
#include <inuk/taktuk/IOredir.h>
#include <inuk/acom>
#include <inuk/sock>
#include <inuk/pth>
#include <mmputenv.h>
#include <inuk/taktuk/debuglog.h>
#include <signal.h>
#ifdef PTH_USE_PTHREAD_IRIX
#include <strings.h> // bzero on IRIX
#endif

#define MAXIOBUF 4000


IOredir::IOredir()
{
  PTH::Pth::initialize();
  state_mutex.initnothrow();
  cond_state.initnothrow();
  //  sem.init(0);
  dbg.log("IOredir","IOredir constructor\n");
	HostName = (char*)malloc (512);
  gethostname((char *)HostName,512);
  
	_in = new SOCK::SocketClient();
  _in->set_fd(0);
  _out = new SOCK::SocketClient();
  _out->set_fd(1);
  _err = new SOCK::SocketClient();
  _err->set_fd(2);
  
	state=none;
  once_control=0;

  _net=NULL;


}

int IOredir::initialize(acom::Network* net)
{
  _net=net;
  if(_net->rank() ==0)
        com_channel = net->get_default_route(-1u);
  else
        com_channel = net->get_default_route(0);
 
  //add i/o services to the service manager WARNING must change this later (no free)
  dbg.log("redirIO","commit try to add Services\n");
  try
    {
      oservice = new output(this);
      iservice = new input(this);
      acom::Service::Manager.bind(acom::ServiceManager::TAKTUK_LNS,1, (acom::Service*)oservice );
      acom::Service::Manager.bind(acom::ServiceManager::TAKTUK_LNS,2, (acom::Service*)iservice );
    }
  catch(...)
    {
      dbg.log("WARNING","IO redir failed to add some Services in the manager\n");;//error
    }


}


IOredir::~IOredir()
{
  wakeupW.close();
  wakeupR.close();
  delete _in;
  delete _out;
  delete _err;
	free (HostName);
  state_mutex.destroy();
  cond_state.destroynothrow();
}

int IOredir::env_export()
{
  //build the env array
  char env[512];

	mmputenv("taktukioredir=OK");
  if(konsole_out!=NULL)
    {
      sprintf(env,"taktukkonsole_out=%d",konsole_out->get_fd());
      mmputenv(env);
  	  dbg.log("IOredir","put env --> %s\n",env);
    }
  if(konsole_in!=NULL)
    {
      sprintf(env,"taktukkonsole_in=%d",konsole_in->get_fd());
      mmputenv(env);
      dbg.log("IOredir","put env --> %s\n",env);
    }
  if(konsole_err!=NULL)
    {
      sprintf(env,"taktukkonsole_err=%d",konsole_err->get_fd());
      mmputenv(env);
      dbg.log("IOredir","put env --> %s\n",env);
    }
  if(fd_out!=NULL)
    {
      sprintf(env,"taktukfd_out=%d",fd_out->get_fd() );
      mmputenv(env);
      dbg.log("IOredir","put env --> %s\n",env);
    }

  if(fd_in!=NULL)
    {
      sprintf(env,"taktukfd_in=%d",fd_in->get_fd() );
      mmputenv(env);
      dbg.log("IOredir","put env --> %s\n",env);
    }

  if(fd_err!=NULL)
    {
      sprintf(env,"taktukfd_err=%d",fd_err->get_fd() );
      mmputenv(env);
      dbg.log("IOredir","put env --> %s\n",env);
    }

  return 1;
}

int IOredir::env_recovery()
{
  char* inter;
  char req[512];

  dbg.log("IOredir","TRY ENV RECOVERY\n");
	if((inter=getenv("taktukioredir"))==NULL)
		return -1;
  if((inter=getenv("taktukkonsole_out"))!=NULL)
  {
    konsole_out=new SOCK::SocketClient();
	konsole_out->set_fd(atoi(inter));
	dbg.log("IOredir","get env  konsole_out fd =%d\n",atoi(inter));
  }
  else
    konsole_out=NULL;
  if((inter=getenv("taktukkonsole_in"))!=NULL)
  {
    konsole_in=new SOCK::SocketClient();
	konsole_in->set_fd(atoi(inter));
	dbg.log("IOredir","get env  konsole_in fd =%d\n",atoi(inter));
  }
  else
    konsole_in=NULL;
  if((inter=getenv("taktukkonsole_err"))!=NULL)
  {
    konsole_err=new SOCK::SocketClient();
	konsole_err->set_fd(atoi(inter));
	dbg.log("IOredir","get env  konsole_err fd =%d\n",atoi(inter));
  }
  else
    konsole_err=NULL;

	if((inter=getenv("taktukfd_out"))!=NULL)
  {
    fd_out=new SOCK::SocketClient();
		fd_out->set_fd(atoi(inter));
		dbg.log("IOredir","get env  fd_out fd =%d\n",atoi(inter));
  }
  else
    fd_out=NULL;
	if((inter=getenv("taktukfd_in"))!=NULL)
  {
    fd_in=new SOCK::SocketClient();
		fd_in->set_fd(atoi(inter));
		dbg.log("IOredir","get env  fd_in fd =%d\n",atoi(inter));
  }
  else
    fd_in=NULL;
	if((inter=getenv("taktukfd_err"))!=NULL)
  {
    fd_err=new SOCK::SocketClient();
		fd_err->set_fd(atoi(inter));
		dbg.log("IOredir","get env  fd_err fd =%d\n",atoi(inter));
  }
  else
    fd_err=NULL;
  dbg.log("IOredir","ENV RECOVERY COMPLETED.......\n");
  return 1;
}

 void thread_IO(IOredir* IO)
 {
   IO->run();
 }



void IOredir::terminate()
{
  dbg.log("IOredir","terminate \n");

  if(_out!=NULL && _out->valid())
  	_out->close();
  if(_err!=NULL && _err->valid())
  	_err->close();
  waiting(in);
  dbg.log("IOredir","IOredir thread state OK to die \n");
}

void IOredir::terminatecommit()
{
  dbg.log("IOredir","IOredir terminate\n");
  IO_kill();
  restore();
  dbg.log("IOredir","IOredir restore IO OK\n");
  dbg.log("IOredir","IOredir terminate OK\n");
 
}

int IOredir::commit()
{
  if(_net==NULL)
    return -1;
  return commit(1,1);
}


int IOredir::commit(int cloexec,int _verbose)
{
  if(_net==NULL)
    return -1;
  
  verbose=_verbose;
  _rank= _net->rank();
  // socket pair allocation
  SOCK::SocketClient* loc_inW=new SOCK::SocketClient();
  SOCK::SocketClient* loc_inR=new SOCK::SocketClient();
  SOCK::SocketClient* loc_outW=new SOCK::SocketClient();
  SOCK::SocketClient* loc_outR=new SOCK::SocketClient();
  SOCK::SocketClient* loc_errW=new SOCK::SocketClient();
  SOCK::SocketClient* loc_errR=new SOCK::SocketClient();


	if(env_recovery()<0)
	//after a resume we have already do it
	{
		if(SOCK::sock_pair(loc_inW,loc_inR)<0)
		{
			dbg.log("ERROR","ERROR redirIO can't allocate socketpair:\n%s\n",strerror(errno));
			return -1;
		}
		if(SOCK::sock_pair(loc_outW,loc_outR)<0)
		{
			dbg.log("ERROR","ERROR redirIO can't allocate socketpair:\n%s\n",strerror(errno));
			return -1;
		}
		if(SOCK::sock_pair(loc_errW,loc_errR)<0)
		{
			dbg.log("ERROR","ERROR redirIO can't allocate socketpair:\n%s\n",strerror(errno));
			return -1;
		}

		//dbg.log("redirIO","socket pair allocation [OK]\n");
		//dbg.log("redirIO","loc_in: %d %d\n",loc_in[0],loc_in[1]);
		//dbg.log("redirIO","loc_out: %d %d\n",loc_out[0],loc_out[1]);
		//dbg.log("redirIO","loc_err: %d %d\n",loc_err[0],loc_err[1]);

		//redirect standard I/O to the up node
		if(_rank ==0)
		{
			dbg.log("redirIO","this node is root redirect the konsole\n");
			konsole_in =_in->dup();
			konsole_out =_out->dup();
			konsole_err =_err->dup();
			//konsole_in->setSockOpt(SOCK::Socket::no_delay,1);
			//konsole_out->setSockOpt(SOCK::Socket::no_delay,1);
			//konsole_err->setSockOpt(SOCK::Socket::no_delay,1);
		}
		else
		{
			//this node is not the root and no konsole initialized
			konsole_in=NULL;
			konsole_out=NULL;
			konsole_err=NULL;
		}

		//set option TCP NODELAY on all sockets
		/*
		loc_inW->setSockOpt(SOCK::Socket::no_delay,1);
		loc_inR->setSockOpt(SOCK::Socket::no_delay,1);
		loc_outW->setSockOpt(SOCK::Socket::no_delay,1);
		loc_outR->setSockOpt(SOCK::Socket::no_delay,1);
		loc_errW->setSockOpt(SOCK::Socket::no_delay,1);
		loc_errR->setSockOpt(SOCK::Socket::no_delay,1);
		*/


		dbg.log("redirIO","IO duplication [OK]\n");
		// replace standard I/O
		loc_inR->dup2(*_in);
		loc_outW->dup2(*_out);
		loc_errW->dup2(*_err);
		if(loc_inR->close()<0)
			dbg.log("ERROR","redirIO close ERROR \n");
		else
			delete(loc_inR);
		if(loc_outW->close()<0)
			dbg.log("ERROR","redirIO close ERROR \n");
		else
			delete(loc_outW);
		if(loc_errW->close()<0)
			dbg.log("ERROR","redirIO close ERROR \n");
		else
			delete(loc_errW);

		dbg.log("redirIO","IO redirection [OK]\n");



		if(cloexec)
		{
			//set option on socket pair & fd_net[0] (the father)
			// when execute an extern programm close router's sockets
			//only 0,1,2 must survive (CF spawner::connect same option on fd_net[i] children)
			loc_inW->setSockOpt(SOCK::Socket::clo_exec,1);
			loc_outR->setSockOpt(SOCK::Socket::clo_exec,1);
			loc_outR->setSockOpt(SOCK::Socket::clo_exec,1);
		}

		fd_in=loc_inW;// IO thread router write on
		fd_out=loc_outR;// IO thread router read on
		fd_err=loc_errR;// IO thread router read on
	}
	//init the socketpair to wake up the thread
	if(SOCK::sock_pair(&wakeupW,&wakeupR)<0)
	{
		dbg.log("","ERROR router can't allocate socketpair:\n");
		return -1;
	}
	else
	{
		wakeupW.setSockOpt(SOCK::Socket::clo_exec,1);
		wakeupW.setSockOpt(SOCK::Socket::clo_exec,1);
		/*wakeupR.setSockOpt(SOCK::Socket::no_delay,1);
		wakeupR.setSockOpt(SOCK::Socket::no_delay,1);
		*/
		//dbg.log("redirIO","socketpair wakeup use: %d %d\n",wakeup[0],wakeup[1]);
	}

	dbg.log("redirIO","IO redirection OK try to create the thread\n");
	//create the IO thread
	if(once_control==0)
	{
		//create the router thread
		int error=0;
		try
		{
			IO_thread.create(&IO_threadId,(void *(*)(void*))thread_IO, this, PTH::Thread::JOINABLE);
		}
		catch(...)
		{
			dbg.log("WARNING","IO redir modules can't create a thread to redirect IO\n");
			return -1;
		}

		dbg.log("IOredir","thread created\n");
		once_control = 1;
	}


	return 1;


}

void IOredir::IO_on()
{
  state_mutex.locknothrow();
  dbg.log("IOredir","IO_on\n");
  state=all;
  cond_state.signalnothrow();
  state_mutex.unlocknothrow();
}

void IOredir::IO_kill()
{
  dbg.log("IOredir","Thread IO KILL\n");
  state_mutex.locknothrow();
  state=killed;
  cond_state.signalnothrow();
  state_mutex.unlocknothrow();
  wakeupW.write((void*)"W",1);
	PTH::Thread::join(IO_threadId,NULL);
  dbg.log("IOredir","Thread IO KILLED\n");
}

void IOredir::IO_off()
{
  state_mutex.locknothrow();
  dbg.log("IOredir","IO_off\n");
  state=none;
  wakeupW.write((void*)"W",1);
  state_mutex.unlocknothrow();
}



void IOredir::restore()
{
	dbg.log("IOredir","RESTORE IO\n");
	if(_out!=NULL && _out->valid())
	  _out->close();
	if(_err!=NULL && _err->valid())
	  _err->close();

  if(_rank==0)
	{
	 	if(_in!=NULL && konsole_in!=NULL)
        konsole_in->dup2(*_in);
	  if(_out!=NULL && konsole_out!=NULL)
        konsole_out->dup2(*_out);
	  if(_err!=NULL && konsole_err!=NULL)
        konsole_err->dup2(*_err);
 	}
  else
	{
      if(konsole_in!=NULL)
				konsole_in->dup2(*_in);
      if(konsole_out!=NULL)
				konsole_out->dup2(*_out);
      if(konsole_err!=NULL)
				konsole_err->dup2(*_err);
	}
}

IOredir::flag IOredir::status()
{
  IOredir::flag ret;

  state_mutex.locknothrow();
  ret=state;
  state_mutex.unlocknothrow();
  return ret;
}

void IOredir::waiting(flag _f)
{
  state_mutex.locknothrow();
  while (state!=_f)
	{
		cond_state.waitnothrow(state_mutex);
		dbg.log("IOredir","waiting for state %d after wake up state= %d\n",_f,state);
	}
  state_mutex.unlocknothrow();
}

void IOredir::waitingwhile(flag _f)
{
  state_mutex.locknothrow();
  while (state==_f)
	{
		cond_state.waitnothrow(state_mutex);
		dbg.log("IOredir","waiting while state = %d after wake up state= %d\n",_f,state);
	}
  state_mutex.unlocknothrow();
}

void IOredir::run()
{
  int nb_open = 0;
	int locout_open =0;
	int locerr_open =0;
	int konsole_open =0;
  char wakeupbuf;



  dbg.log("IOredir","thread begin init\n");

  SOCK::FDSet fds_null;
  SOCK::FDSet fds_null2;
  fds_null.zero();
  fds_null2.zero();

  while(1)
	{
		if(status()==none)
	  {
			dbg.log("IOredir","None connection checked waiting order IO_on\n");
	    waitingwhile(none);
			dbg.log("IOredir","IO Thread IO_on request OK restart\n");
	  }
		else
			if(status()==killed)
			{
				dbg.log("IOredir","Kill status checked thread terminate\n");
				return;
			}
		// read set init with the fd_net set and all local in,out,err set
		dbg.log("IOredir","thread connection set init\n");
		SOCK::FDSet read_set;

		int select_out;

		//int max_fd = 0;
		//reset read set & control variable
		dbg.log("IOredir","set read_set to zero\n");
		read_set.zero();
	  
		dbg.log("IOredir","read_set try to add wakeup\n");
		read_set.set(&wakeupR);

		dbg.log("IOredir","read set wakeup\n");
		nb_open =0;
		locout_open=0;
		locerr_open=0;
		konsole_open=0;

		// look for a Konsole
		if(konsole_in!=NULL && konsole_in->valid() && status() !=none)
		{
			read_set.set(konsole_in);
			//dbg.log("IOredir","fd read set + konsole_in %s\n",konsole_in->c_str());
			dbg.log("IOredir","fd read set + konsole_in \n");
			konsole_open=1;
		}
	 	//look for local std out
 		if(status() == all)
		{
			dbg.log("IOredir","state = ALL\n");
	    if(fd_out!=NULL && fd_out->valid())
			{
				read_set.set(fd_out);
				dbg.log("IOredir","fd read set out \n%s",fd_out->c_str());
				locout_open=1;
			}

	    if(fd_err!=NULL&& fd_err->valid())
			{
				read_set.set(fd_err);
				dbg.log("IOredir","fd read set err \n%s",fd_err->c_str());
				locerr_open=1;
			}
		}
    
		if((locout_open + locerr_open)<=0/*&&!konsole_in->valid()*/)
		{
		 	state_mutex.locknothrow();
			state=in;
			cond_state.signalnothrow();
			state_mutex.unlocknothrow();
		}
																
		
		dbg.log("IOredir","perform the select\n");
	  select_out=SOCK::select(read_set, fds_null, fds_null2);

		if(select_out < 0)
		{
			dbg.log("WARNING","IOTHREAD: ERROR in the select %s \n",strerror(errno));
			return;
		}
	 	//check the read socket set
 		dbg.log("IOredir","check connections set after select\n");

		if(read_set.isready(&wakeupR))
		{
			dbg.log("IOredir","IOREDIR  WAKE UP\n");
			wakeupR.read(&wakeupbuf,1);
			dbg.log("IOredir","select WAKE UP data read\n");
		}

	 	//konsole
 		if(konsole_in!=NULL&& konsole_in->valid())
			if(read_set.isready(konsole_in))
			{
				//konsole in : --> broadcast this to all
				dbg.log("IOredir","receive somthing from the konsole\n");
				if(konsoleprocess()<=0)
	      {
					dbg.log("IOredir","Konsole was closed\n");
					//		read_set.clr(konsole_in);
	      }
			}
		//local stdout set
	  if(fd_out!=NULL&&fd_out->valid())
	    if(read_set.isready(fd_out))
			{
				dbg.log("IOredir","receive somthing from local stdout\n");
				if(localIOprocess(IOout,fd_out)<=0)
				{
					//the connection was lost
					// one local stdout finished or closed his connection
					dbg.log("IOredir","local stdout closed \n");
					locout_open=0;
				}
			}
		//local stderr set
	  if(fd_err!=NULL&&fd_err->valid())
		{
	 		if(read_set.isready(fd_err))
			{
				dbg.log("IOredir","receive somthing from local stderr \n");
				if(localIOprocess(IOerr,fd_err)<=0)
		    {
		      //the connection was lost
		      // one local stderr finished or closed his connection
					dbg.log("IOredir","local stderr closed \n");
					locerr_open=0;
		    }
			}
		}
	}// while
  dbg.log("I0redir","thread IO router terminate... never dispayed\n");
}




void IOredir::localIOprint(local_iobuf* iobuf)
{
  char buffer[64];
  bzero(buffer,64);
 
  if(_rank!=0)
    return; // error

  dbg.log("output_service","iobuf:\nemet:%d\ntype:%d\nsize:%d\nbuf:%s\n",iobuf->emet_id,iobuf->iotype,iobuf->size,iobuf->buffer);
  dbg.log("output_service","iobuf: @buffer %x\n",iobuf->buffer);
  switch (iobuf->iotype)
	{
    case IOout:
      if ( verbose )
			{
				sprintf(buffer,"<%s> [rank:%d]:",iobuf->emet_name,iobuf->emet_id);
				//	  sprintf(buffer,"\n%s> ",iobuf->emet_name);
				if(konsole_out!=0&&konsole_out->valid())
					konsole_out->write(buffer,strlen(buffer));
			}
			if(konsole_out!=0&&konsole_out->valid())
				konsole_out->write(iobuf->buffer,iobuf->size);
   	  break;
		case IOerr:
			if (verbose )
			{
				sprintf(buffer,"<%s> [rank:%d]:",iobuf->emet_name,iobuf->emet_id);
				//sprintf(buffer,"\n%s> ",iobuf->emet_name);
				if(konsole_err!=0&&konsole_err->valid())
					(konsole_err)->write(buffer,strlen(buffer));
			}
			if(konsole_err!=0&&konsole_err->valid())
	   	   (konsole_err)->write(iobuf->buffer,iobuf->size);
    	break;
		default:
      break;//nothing
	}
}

void IOredir::localIOrecv(local_iobuf* iobuf)
{
 // iobuf->buffer[iobuf->size]='\0';
 // dbg.log("input_service","iobuf:\nemet:%d\ntype:%d\nsize:%d\nbuf:%s\n",iobuf->emet_id,iobuf->iotype,iobuf->size,iobuf->buffer);

	sigset_t mask;
	sigemptyset(&mask);
	sigaddset(&mask, SIGPIPE);
	sigprocmask(SIG_BLOCK,&mask,NULL);

  if(iobuf->size>0)
	{
 		dbg.log("input_service","broadcast on all local thread\n");
		//braodcast this input message on all local threads or processes
		if(fd_in!=0 && fd_in->valid())
		{
			dbg.log("input_service","broacast to fd_in  size to send %d\n",iobuf->size);

			fd_in->write(iobuf->buffer, iobuf->size);
			dbg.log("input_service","broacast to fd_in finished\n");
		}
	}
  else
	{
		dbg.log("input_service","close local input\n");
		if(fd_in!=0 && fd_in->valid())
			fd_in->close();
	}

}

int IOredir::localIOprocess(IOTypes type, SOCK::SocketClient* local_s)
{
  local_iobuf msg;
	msg.buffer=NULL;
	size_t sizebuf=0;
  int nbread=0;
	//size_t nbread=0;

	std::string buf;
	if(local_s->readline(buf)<=0)
	{
		nbread=0;
	}else
	{
		msg.buffer=(char*)buf.c_str();
		nbread=buf.size();
	}
	/*FILE* input=fdopen(local_s->get_fd(),"r");
	if (input == NULL)
		dbg.log("IOredir","CAN'T OPEN FILE*\n");
	if((nbread=getline(&msg.buffer, &sizebuf, input))>=0)
		nbread+=1;
	else
	{
		nbread=0;
		local_s->close();
	}
	*/
//	msg.buffer=(char*)alloca(MAXIOBUF*sizeof(char));
//  nbread=local_s->read(&msg.buffer,MAXIOBUF);
	
  dbg.log("IOredir","localIOprocess read %d: [%s]\n",nbread,msg.buffer);
  msg.emet_id=_net->rank();
  msg.emet_name=HostName;
  msg.iotype=type;
  msg.size=nbread;
  if(konsole_out!=0&&konsole_out->valid()&&nbread>0)
  	localIOprint(&msg);
  else
    if(nbread>0)
		{
			try{
				acom::ActiveMessage<output> am(_net, oservice, 0);
				am->put(&msg.emet_id,sizeof(acom::Network::LocalNodeNo));
				am->put(msg.emet_name,strlen(msg.emet_name)+1); // NULL terminated
				am->put(&msg.iotype,sizeof(int8_t));
				am->put(&msg.size,sizeof(int32_t));
				if(msg.size>0)
					am->put(msg.buffer, nbread);
				am.post(com_channel);
      }
			catch(...){
        dbg.log("WARNING","IOredir module can't send message one specified Net\n");
      }
 		}
	if(sizebuf!=0)
		free(msg.buffer);
  return nbread;
}

int IOredir::konsoleprocess()
{
	local_iobuf msg;
	msg.buffer=NULL;
  //size_t nbread=0;
  int nbread=0;
	size_t sizebuf=0;
	IOTypes iot=IOin;

	char* buf=(char*)alloca(MAXIOBUF*sizeof(char));
  if((nbread=konsole_in->read((void*)buf,MAXIOBUF))<0)
	{
		nbread=0;
	}else
	{
		msg.buffer=(char*)buf;
		//nbread=buf.size();
	}
											

//	msg.buffer=(char*)alloca(MAXIOBUF*sizeof(char));
//  nbread=konsole_in->read(msg.buffer,MAXIOBUF);

/*	FILE* fdinput=fdopen(konsole_in->get_fd(),"rb");
  if((nbread=getline(&msg.buffer, &sizebuf, fdinput))>=0)
		nbread+=1;
	else
	{
		nbread=0;
		konsole_in->close();
	}
*/


  msg.emet_id=_net->rank();
	msg.emet_name=HostName;
  msg.iotype=IOin;
  if(nbread<=0)
    msg.size=0;
  else
    msg.size=nbread;

  dbg.log("IOredir","konsole process read %d send a msg \n",nbread);
  try{
    acom::ActiveMessage<input> am (_net, iservice, 0);
		am->put(&msg.emet_id,sizeof(acom::Network::LocalNodeNo));
		am->put(msg.emet_name,strlen(msg.emet_name)+1); //NULL terminated
		am->put(&msg.iotype,sizeof(int8_t));
		am->put(&msg.size,sizeof(int32_t));
		if(msg.size>0)
			am->put(msg.buffer, nbread);
    am.post(com_channel);
  }
  catch(...){
    dbg.log("WARNING","IOredir can't send message on specified net\n");
  }
  //send to my local thread too 
  localIOrecv(&msg);
	if(sizebuf!=0)
		free(msg.buffer);
  return nbread;
}







//active message section

input::input(IOredir* IO)
{
  _IO=IO;
}


void input::call(
			 acom::ComFailure::Code e,      // if any
			 acom::Network::Receipt* receipt,
			 acom::Network::Header* h,
			 acom::IOVectEntry* iov,
			 acom::IOVectSize   iov_sz,
       void* ctxt
			 ) throw (acom::ComFailure)
{
  dbg.log("", "input service called \n");
	if(e != acom::ComFailure::OK)
    return;

  local_iobuf iobuf;
	iobuf.emet_id=*(acom::Network::LocalNodeNo*)(iov[0].data);
	iobuf.emet_name=(char*)(iov[1].data);
	iobuf.iotype=IOin;
	iobuf.size=*(int32_t*)(iov[3].data);
	if(iobuf.size>0)
		iobuf.buffer=(char*)(iov[4].data);
	
  _IO->localIOrecv(&iobuf);
  dbg.log("input_service","FINISHED\n");
}


output::output(IOredir* IO)
{
  _IO=IO;
}
void output::call(
			  acom::ComFailure::Code e,      // if any
			  acom::Network::Receipt* receipt,
			  acom::Network::Header* h,
			  acom::IOVectEntry* iov,
			  acom::IOVectSize   iov_sz,
       void* ctxt
			  ) throw (acom::ComFailure)
{
  if(e != acom::ComFailure::OK)
    return;

//  local_iobuf* iobuf= (local_iobuf*)(iov[0].data);
	local_iobuf iobuf;
	iobuf.emet_id=*(acom::Network::LocalNodeNo*)(iov[0].data);
	iobuf.emet_name=(char*)(iov[1].data);
  dbg.log("IOredir","output msg : emet name = %s\n",iobuf.emet_name);
	iobuf.iotype=*(int8_t*)(iov[2].data);
	iobuf.size=*(int32_t*)(iov[3].data);
	if(iobuf.size>0)
		iobuf.buffer=(char*)(iov[4].data);
	
	_IO->localIOprint(&iobuf);
  //_IO->localIOprint(&iobuf);
  dbg.log("ouput_service","FINISHED\n");
}
