/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */


#include <vector>
#include <string>
#include <inuk/taktuk/connector.h>
#include <connector_rsh.h>
#include <connector_rsh_async.h>
#include <connector_ssh.h>
#include <connector_ssf.h>
#include <connector_rsh_fork.h>
#include <connector_fork.h>
#include <connector_rsh_thread.h>
#include <connector_envrsh.h>
//TG#include <connector_rexec.h>
#include <inuk/taktuk/debuglog.h>
#include <inuk/taktuk/scheduler.h>
#include <inuk/taktuk/taktukconf.h>
#include <iostream>
#include <inuk/pth>
#include <sys/time.h>
#include <sys/socket.h>

//#include <pthread.h>

connector_collector::connector_collector()
{
  //save different availlable connectors
  add<rsh>("rshs");
  add<envrsh>("env");
  add<rsh_async>("rsha");
  add<rsh_fork>("rshf");
  add<rsh_fork>("rsh");
  //add<Srexec>("rexec");
  add<rsh_thread>("rsht");
  add<ssh>("ssh");
  add<ssf>("ssf");
	add<confork>("fork");

  //YOU CAN ADD CONNECTORS HERE....
  //....
}

connector_collector::~connector_collector()
{
  std::map<std::string,WrapperFactoryBase*,ltstr>::iterator cur  = collector.begin();
  // maybe delete connectors saved
  for(;cur!=collector.end();cur++)
    delete (*cur).second;
}

void runnable(connector* con)
{
  dbg.log("connector","Thread for connector launched try to call connector's run\n");
	Profile SpawnTime;
	SpawnTime.B();
  con->run();
  con->getalive();
	SpawnTime.E();
	//CM con->_time=SpawnTime.Time_msec();
	//CM gettimeofday(&(con->_end),NULL);
  dbg.log("connector","Spawn time %f sec \n",SpawnTime.Time_sec());
	//record new nodes completed in the scheduler
	//control job correctly done & and supress rejected host
	for(std::vector<topology::host*>::iterator iter=con->remote_hosts.begin();
																							iter!=con->remote_hosts.end();)
	{
		if((*iter)->fd==-1)
		{
			iter=con->remote_hosts.erase(iter);
		}
		else
		{

			iter++;
		}
	}
	// if empty don't have to record it
	if(!con->remote_hosts.empty())
	{
		con->_schd->add_childs(con->remote_hosts);
	}else
	{
		// time record aborted connector failed
		SpawnTime.erase();
	}
  //implemente the join
	dbg.log("connector","before set time jobid : %i\n",con->_id);

  con->_rt->set_time(SpawnTime,con->_id);
  --(*con->_cd);
	dbg.log("connector","after decremente count down \n");

	delete con;
}

//int connector::connect(std::vector<topology::host*>*reject_hosts, std::string l_user,PTH::CountDown* cd,scheduler* schd)
//int connector::connect(std::vector<topology::host*>*reject_hosts, std::string l_user)
int connector::connect(synch_vector<topology::host*>*reject_hosts, std::string l_user,PTH::CountDown* cd,scheduler* schd,int id,run_time* rt)
{
	_schd=schd;
  _cd=cd;
  _id=id;
  _rt=rt;

  local_user=l_user;
  reject=reject_hosts;
  //call connect in a thread
  PTH::ThreadId pthid=connect_thread.getSelf();
  try
  {
  	connect_thread.create(&pthid,(void *(*)(void*))runnable,(void*) this);
  }
  catch(...)
  {
		dbg.log("WARNING","connector module can't create Thread\n");
		std::cerr<<"Connector module can't allocate new threads"<<std::endl;
		for(int i=0;i<remote_hosts.size();i++)
		{
			remote_hosts[i]->fd=-1;
			reject_hosts->push_back(remote_hosts[i]);
		}
    return -1;
  }
	/*
  if(pthread_create(&connect_thread,NULL,(void*(*)(void*))runnable,(void*)this)<0)
  {
     dbg.log("WARNING","connector module can't create Thread\n");
     return -1;
  }
	*/
  return 0;

}

void connector::getalive()
{

  fd_set ackset;
  FD_ZERO(&ackset);
  int max=-1;
  int nbhosts = 0;
  int hostsack[remote_hosts.size()];
	int timeout=0;
  int remaintime=0;
	int selectout=0;
	int errorread=0;
	struct timeval time;
	time.tv_sec=0;
	time.tv_usec=0;
  struct timeval elapsedtime;
	struct timeval starttime;
	int nodelayon=1;
  dbg.log("connector", "GETALIVE Begin of getalive \n");

  for (int i=0; i<remote_hosts.size(); i++){
    if (remote_hosts[i]->fd != -1){
			setsockopt(remote_hosts[i]->fd,IPPROTO_TCP,TCP_NODELAY,&nodelayon,sizeof(nodelayon));
      hostsack[i] = 1;
      nbhosts++;
      if (remote_hosts[i]->timeout > timeout)
				timeout = remote_hosts[i]->timeout;
			dbg.log("connector","GETALIVE Time out for host %s  =	%d\n",remote_hosts[i]->name.c_str(),timeout);
    }
    else
      hostsack[i] = 0;
  }

	remaintime=timeout*1000;

  while(nbhosts > 0)
	{
    max=-1;
    FD_ZERO(&ackset);
    for (int i=0; i<remote_hosts.size(); i++)
		{
      if ((remote_hosts[i]->fd != -1) && (hostsack[i] != 0))
			{
				FD_SET(remote_hosts[i]->fd, &ackset);
				if (remote_hosts[i]->fd > max)
					max = remote_hosts[i]->fd;
      }
    }
		time.tv_sec=remaintime/1000000;
		time.tv_usec=remaintime%1000000;
		gettimeofday(&starttime,NULL);
		dbg.log("connector", "GETALIVE SELECT waiting for %i more hosts Time out = %d sec and %d micros\n", nbhosts,time.tv_sec,time.tv_usec);
		do 
		{
    	selectout=select(max+1, &ackset, NULL, NULL,((time.tv_sec==0&&time.tv_usec==0)?NULL:&time)); ///mettre un timeout ici pour connecteur recalsitrant
		}while(selectout==-1&&errno==EINTR);
		
		dbg.log("connector","GETALIVE select out = %d\n",selectout);
		if(selectout==0)
    {
			break;
   	}
		if(timeout!=0)
    {
			dbg.log("connector","GETALIVE Time out dring and !=0 \n");
    	gettimeofday(&elapsedtime,NULL);
    	//compute remaining time to read the CHAR_CTRL
			dbg.log("connector","GETALIVE Timeout Elapsed Time %d\n",((elapsedtime.tv_usec+elapsedtime.tv_sec*1000000)-(starttime.tv_usec+starttime.tv_sec*1000000)));
    	remaintime=remaintime-((elapsedtime.tv_usec+elapsedtime.tv_sec*1000000)-(starttime.tv_usec+starttime.tv_sec*1000000));
    }
    for(int i=0; i<remote_hosts.size(); i++)
		{
      if (remote_hosts[i]->fd != -1)
			{
				if (FD_ISSET(remote_hosts[i]->fd, &ackset))
        {
					char ack;
					char version[3];
					do
					{
						errorread=read(remote_hosts[i]->fd, &ack, 1);
					}while(errorread==-1&&errno==EINTR);
					if ( errorread <=0)
          {
						dbg.log("connector", "GETALIVE Error when reading ack from socket in taktuk remote execution request  connection closed by peer\n");
						remote_hosts[i]->fd = -1;
            reject->push_back(remote_hosts[i]);
            hostsack[i] = 0;
						nbhosts--;
						this->cancel();
					}
					else
							if (ack != CHAR_CTRL)
              {
								remote_hosts[i]->badtxt+=ack;
                //read all the sequence before the CHAR_CTRL
							}
							else
              {
              	//this node is OK CHAR_CTRL received
								dbg.log("connector", "GETALIVE OK for host	%s\n",remote_hosts[i]->name.c_str());
								do
								{
									errorread=read(remote_hosts[i]->fd, version, 3);
								}while(errorread==-1&&errno==EINTR);
								if ( errorread < 3)
								{
									dbg.log("connector", "GETALIVE Error when reading version\n");
									remote_hosts[i]->fd = -1;
									reject->push_back(remote_hosts[i]);
									hostsack[i] = 0;
									nbhosts--;																								
									this->cancel();
								}else
								{	
									if(strncmp(TAKTUKVERSION,version,3)!=0)
									{
										dbg.log("connector", "GETALIVE Error when reading version\n");
										std::cerr<<"Bad Taktuk version detected with node "<<remote_hosts[i]->name <<" with verison "<< version <<" aborded, (use same Taktuk version on all requested nodes) local version is "<<TAKTUKVERSION<<std::endl;
    	              remote_hosts[i]->fd = -1;
  	                reject->push_back(remote_hosts[i]);
      	            hostsack[i] = 0;
        	          nbhosts--;
										this->cancel();
									}else
									{
										hostsack[i] = 0;
										nbhosts--;
									}
								}	
              }



					dbg.log("connector", "GETALIVE Waiting for %i more hosts \n", nbhosts);
				}
			}
		}
	}
	//if(selectout==0) // total time elapsed
	//{
	for(int i=0; i<remote_hosts.size(); i++)
	{
		if(hostsack[i]==1)
		{
			if(remote_hosts[i]->fd!=-1)
			{
				// Timeout rejected
				dbg.log("connector"," GETALIVE Child %d is rejected named %s on TIME OUT fd=%d\n",i,remote_hosts[i]->name.c_str(),remote_hosts[i]->fd);
				close(remote_hosts[i]->fd);
				remote_hosts[i]->fd=-1;
				reject->push_back(remote_hosts[i]);
			}
		}
		if(!remote_hosts[i]->badtxt.empty())
		{
			dbg.log("connector", "GETALIVE Reading string from child : [%s] \n", remote_hosts[i]->badtxt.c_str());
			std::cout << remote_hosts[i]->badtxt << std::endl;
		}
	}
	//}
  dbg.log("connector", "GETALIVE End of getalive \n");
}
