#include <inuk/taktuk/mput_file_reader.h>
#include <inuk/taktuk/mput_wrappers.h>
#include <inuk/taktuk/mput_buffers.h>
#include <string.h>

int current_read_buffer;
int current_read_position;
int last_read_buffer;
static int nbfile;
static int nblnk;
static int nbdir;

void init_file_reader()
{

   nbfile=0;
   nblnk=0;
   nbdir=0;

   current_read_buffer=0;
   current_read_position=0;
   last_read_buffer=NBBUFF-1;

   /* get first buffer */   
   Pthread_mutex_lock(&(buffers[current_read_buffer].buffer_full_mutex));
   while(buffers[current_read_buffer].buffer_full != 0)
      Pthread_cond_wait(&(buffers[current_read_buffer].buffer_full_cond),&(buffers[current_read_buffer].buffer_full_mutex));
   Pthread_mutex_unlock(&(buffers[current_read_buffer].buffer_full_mutex));

   /* lock this buffer*/
   Pthread_mutex_lock(&(buffers[current_read_buffer].prod_mutex));
   buffers[current_read_buffer].prod = 1;
   Pthread_mutex_unlock(&(buffers[current_read_buffer].prod_mutex));

   /* unlock precedent buffer*/  
   Pthread_mutex_lock(&(buffers[last_read_buffer].prod_mutex));
   buffers[last_read_buffer].prod = 0;
   Pthread_cond_broadcast(&(buffers[last_read_buffer].prod_cond));
   Pthread_mutex_unlock(&(buffers[last_read_buffer].prod_mutex));

}

void switch_read_buffer()
{
   /* free buffer */
   Pthread_mutex_lock(&(buffers[current_read_buffer].buffer_full_mutex));
   buffers[current_read_buffer].buffer_full = nb_consumers;
   Pthread_mutex_unlock(&(buffers[current_read_buffer].buffer_full_mutex));

   /* choose new buffer */
   last_read_buffer = current_read_buffer;
   current_read_buffer++;
   if(current_read_buffer == NBBUFF)
      current_read_buffer = 0;
         
   /* get new buffer */
   Pthread_mutex_lock(&(buffers[current_read_buffer].buffer_full_mutex));
   while(buffers[current_read_buffer].buffer_full != 0)
      Pthread_cond_wait(&(buffers[current_read_buffer].buffer_full_cond),&(buffers[current_read_buffer].buffer_full_mutex));
   Pthread_mutex_unlock(&(buffers[current_read_buffer].buffer_full_mutex));
   current_read_position = 0;
   
   /* lock this buffer*/
   Pthread_mutex_lock(&(buffers[current_read_buffer].prod_mutex));
   buffers[current_read_buffer].prod = 1;
   Pthread_mutex_unlock(&(buffers[current_read_buffer].prod_mutex));

   /* unlock precedent buffer*/  
   Pthread_mutex_lock(&(buffers[last_read_buffer].prod_mutex));
   buffers[last_read_buffer].prod = 0;
   Pthread_cond_broadcast(&(buffers[last_read_buffer].prod_cond));
   Pthread_mutex_unlock(&(buffers[last_read_buffer].prod_mutex));
    
}

int get_file_name_start(const char * file_name)
{
   int i,rank,length;
   
   rank=-1;
   length=0;
   
   length=strlen(file_name);
   /* ignore trailing '/'*/
   while( (length != 1) && (file_name[length -1]=='/') )
      length--;
   
   for(i=0;i<length;i++)
   {
      if( file_name[i] == '/' )
         rank=i;
   }
   rank++;
   return rank;
}

void set_file_name_size(uint32_t file_name_size)
{
   int i;
   uint8_t * conv;
   conv = (uint8_t *) &file_name_size;
   for(i=0;i<sizeof(file_name_size);i++)
   {
      buffers[current_read_buffer].buff[current_read_position] = conv[i];
      current_read_position++;

      if(current_read_position == MAXBUFF)
      {  
         switch_read_buffer();
      }
   }
  
}

void set_file_name(const char * file_name, const char * directory)
{
   int file_name_size;
   int directory_size;
   int file_name_pos=0;
   int directory_pos=0;
   int temp_size;
   
   file_name_size = strlen(file_name)+1;
   directory_size = strlen(directory);
   set_file_name_size(file_name_size+directory_size);

   while(directory_size > 0)
   {
      if(current_read_position + directory_size <= MAXBUFF)
      {
         strncpy(&(buffers[current_read_buffer].buff[current_read_position]), &(directory[directory_pos]), directory_size);
         current_read_position += directory_size;
	 directory_size = 0;
      }
      else
      {
         temp_size = MAXBUFF - current_read_position;
         strncpy(&(buffers[current_read_buffer].buff[current_read_position]), &(directory[directory_pos]), temp_size);
         current_read_position = MAXBUFF;
         directory_size -= temp_size;
         directory_pos += temp_size;
      }
      if(current_read_position == MAXBUFF)
      {
         switch_read_buffer();
      }
   }

   while(file_name_size > 0)
   {
      if(current_read_position + file_name_size <= MAXBUFF)
      {
         strncpy(&(buffers[current_read_buffer].buff[current_read_position]), &(file_name[file_name_pos]), file_name_size);
         current_read_position += file_name_size;
	 file_name_size = 0;
      }
      else
      {
         temp_size = MAXBUFF - current_read_position;
         strncpy(&(buffers[current_read_buffer].buff[current_read_position]), &(file_name[file_name_pos]), temp_size);
         current_read_position = MAXBUFF;
         file_name_size -= temp_size;
         file_name_pos += temp_size;
      }
      if(current_read_position == MAXBUFF)
      {
         switch_read_buffer();
      }
   }
}

void set_stat(struct stat * file_stat)
{
   int i;
   int struct_stat_size;
   int struct_stat_pos=0;
   int temp_size;
   uint8_t * conv;
   
   conv = (uint8_t *) file_stat;

   struct_stat_size = sizeof(struct stat);

   while(struct_stat_size>0)
   {
      if(current_read_position + struct_stat_size <= MAXBUFF)
      {
         memcpy(&(buffers[current_read_buffer].buff[current_read_position]), &(conv[struct_stat_pos]), struct_stat_size);
         current_read_position += struct_stat_size;
	 struct_stat_size = 0;
      }
      else
      {
         temp_size = MAXBUFF - current_read_position;
         memcpy(&(buffers[current_read_buffer].buff[current_read_position]), &(conv[struct_stat_pos]), temp_size);
         current_read_position = MAXBUFF;
         struct_stat_size -= temp_size;
         struct_stat_pos += temp_size;
      }
      if(current_read_position == MAXBUFF)
      {
         switch_read_buffer();
      }
   }
}

void read_data(int input, struct stat * file_stat)
{
   off_t temp_size;
   off_t data_size;
   int n;
   
   data_size = file_stat->st_size;
   
   while(data_size > 0)
   {
      if(current_read_position + data_size <= MAXBUFF)
      {
         n=0;
	 while( (n += Read(input,&(buffers[current_read_buffer].buff[current_read_position+n]),data_size-n)) < data_size)
  	 {
	    ;
	 }
         current_read_position += data_size;
	 data_size=0;
      }
      else
      {
         temp_size = MAXBUFF - current_read_position;
         n=0;
	 while( (n += Read(input,&(buffers[current_read_buffer].buff[current_read_position+n]),temp_size-n)) < temp_size)
  	 {
	    ;
	 }
         current_read_position = MAXBUFF;
         data_size -= temp_size;
      }
      if(current_read_position == MAXBUFF)
      {
         switch_read_buffer();
      }
   }
}

int file_read_relative(const char * file_name, const char * directory, const char * base)
{
   DIR *source_dir;
   char *dir_name;
   char *sym_target;
   struct stat file_stat;
   int n;
   int result=0;
   int dir_length;
   struct dirent *file;
   int input;
   
   dir_name = (char *)Malloc( (strlen(file_name) + strlen(directory) + strlen(base) + 1)*sizeof(char));
   dir_name[0]='\0';
   strcat(dir_name, base);
   strcat(dir_name, directory);
   strcat(dir_name, file_name);

   if(verbose>0)
   {
      printf("--> %s\n",dir_name);
   }   
   memset(&file_stat,0,sizeof(struct stat)); 
   if( lstat(dir_name, &file_stat) < 0)
   {
      std::cerr<<"stat_error : file "<<dir_name<<" : "<<strerror(errno)<<"\n";
      free(dir_name);
      return 0;
   }
   
   if(S_ISLNK(file_stat.st_mode))
   {
      sym_target = (char *)Malloc( (file_stat.st_size + 1)*sizeof(char));
      if(readlink(dir_name, sym_target, file_stat.st_size + 1) <0)
      {
         std::cerr<<"readlink_error : could not read "<<dir_name<<" : "<<strerror(errno)<<"\n";
         free(dir_name);
         free(sym_target);
         return 0;
      }
      sym_target[file_stat.st_size]='\0';
      set_file_name(file_name, directory);
      set_stat(&file_stat);
      set_file_name(sym_target, "");
      free(sym_target);
      free(dir_name);
      nblnk++;
      return 1;
   }
   else if(S_ISREG(file_stat.st_mode))
   {

      if( (input = open(dir_name,O_RDONLY)) < 0)
      {
         std::cerr<<"open_error : file "<<base<<directory<<file_name<<" : "<<strerror(errno)<<"\n";
	 free(dir_name);
         return 0;
      }
      set_file_name(file_name, directory);
      set_stat(&file_stat);
      read_data(input, &file_stat);
      Close(input);
      free(dir_name);
      nbfile++;
      return 1;
   }
   else if(S_ISDIR(file_stat.st_mode))
   {
      if( (source_dir = opendir(dir_name)) == NULL)
      {
         std::cerr<<"opendir_error : directory "<<base<<directory<<file_name<<" : "<<strerror(errno)<<"\n";
         return 0;
      }
      set_file_name(file_name, directory);
      set_stat(&file_stat);
      free(dir_name);
      /*read dir*/
      dir_length = strlen(file_name) + strlen(directory) + 2;
      dir_name = (char *)Malloc( dir_length * sizeof(char));
      dir_name[0]='\0';
      strcat(dir_name, directory);
      strcat(dir_name, file_name);
      dir_name[dir_length - 2] = '/';
      dir_name[dir_length - 1] = '\0';
      
      while ( (file = readdir(source_dir)) != NULL )
      {
         if (strcmp(file->d_name, ".") == 0) 
            continue;   /* current dir */
         if (strcmp(file->d_name, "..") == 0)
            continue;  /* parent dir  */
         result += file_read_relative( file->d_name, dir_name, base);
      }
      free(dir_name);
      Closedir(source_dir);
      nbdir++;
      return result;
   }
   free(dir_name);
   return 0;
}

int file_read_absolute( char * file_name)
{
   char *base;
   char *file_name_start;
   int start;
   int result;
   
   start = get_file_name_start(file_name);
   
   base = (char *)malloc( sizeof(char) * (start+1) );
   
   strncpy(base,file_name,start);
   base[start]='\0';
   
   file_name_start = &(file_name[start]);
   
   result = file_read_relative(file_name_start,"\0",base);
   
   free(base);
   
   return result;

}
int file_read_source_dest(char * source)
{
   int fd;
   struct stat file_stat;
   int start;
   int result=0;
   DIR *source_dir;
   struct dirent *file;
   char * dir_name;
   int dir_length;
   
   memset(&file_stat,0,sizeof(struct stat));  
   if( stat(source,&file_stat) < 0 )
   {
      std::cerr<<"stat_error : source "<<source<<" could not be opened : "<<strerror(errno)<<"\n";
      return 0;
   }
   
   if(S_ISREG(file_stat.st_mode))
   {
      return file_read_absolute( source);
   }
   else if(S_ISDIR(file_stat.st_mode))
   {
      if( (source_dir = opendir(source)) == NULL)
      {
         std::cerr<<"opendir_error : directory "<<source<<" could not be opened : "<<strerror(errno)<<"\n";
         return 0;
      }
      start = get_file_name_start(source);
      set_file_name(&(source[start]),"");
      set_stat(&file_stat);
      
      dir_length = strlen(source) + 2;
      dir_name = (char *)Malloc( dir_length * sizeof(char));
      dir_name[0]='\0';
      strcat(dir_name, source);
      dir_name[dir_length - 2] = '/';
      dir_name[dir_length - 1] = '\0';
      
      while ( (file = readdir(source_dir)) != NULL )
      {
         if (strcmp(file->d_name, ".") == 0) 
            continue;   /* current dir */
         if (strcmp(file->d_name, "..") == 0)
            continue;  /* parent dir  */
         result += file_read_relative( file->d_name, "", dir_name);
      }
      nbdir++;
      Closedir(source_dir);
      free(dir_name);
      return result;
   }
   return 0;

}

void * file_reader(void *input_files)
{
   struct files *in;
   int file_count=0;
   int i,n;

   in = (struct files *) input_files;
   
   init_file_reader();
   
   if(source_dest)
   {
      file_count = file_read_source_dest(in->file_name[0]);
   }
   else
   {
      for( i = 0 ; i < in->file_count ; i++ )
      {
         if( (n = file_read_absolute( in->file_name[i] )) > 0 )
            file_count += n;
      }
   }
   
   set_file_name_size(0);
   
   if(current_read_position>0)
   {
      memset(&(buffers[current_read_buffer].buff[current_read_position]),0,MAXBUFF-current_read_position);
      switch_read_buffer();
   }
   if(verbose > 0 || reportflag > 0)
   {
      printf("%d files sent\n",file_count);
      printf("%d regular files sent\n",nbfile);
      printf("%d symlink sent\n",nblnk);
      printf("%d directory sent\n",nbdir);
   }
   return NULL;
}
