#include <inuk/taktuk/mput_file_writer.h>
#include <inuk/taktuk/mput_wrappers.h>
#include <inuk/taktuk/mput_buffers.h>
#include <string.h>

int current_write_buffer;
int current_write_position;
static int nbdir;
static int nbfile;
static int nblnk;

uint32_t get_file_name_size()
{
   uint32_t file_name_size;
   int i;
   uint8_t * conv;
   conv = (uint8_t *) &file_name_size;
   for(i=0;i<sizeof(file_name_size);i++)
   {
      conv[i] = buffers[current_write_buffer].buff[current_write_position];
      current_write_position++;
      
      if(current_write_position == MAXBUFF)
      {  
         switch_write_buffer();
      }
   }
   return file_name_size;
}

void get_file_name(char * file_name, int file_name_size)
{
   int temp_size;
   int file_name_pos=0;
   
   while(file_name_size > 0)
   {
      if(current_write_position + file_name_size <= MAXBUFF)
      {
         memcpy(&(file_name[file_name_pos]), &(buffers[current_write_buffer].buff[current_write_position]), file_name_size);
         current_write_position += file_name_size;
	 file_name_size = 0;
      }
      else
      {
         temp_size = MAXBUFF - current_write_position;
         memcpy(&(file_name[file_name_pos]), &(buffers[current_write_buffer].buff[current_write_position]), temp_size);
         current_write_position = MAXBUFF;
         file_name_size -= temp_size;
         file_name_pos += temp_size;
      }
      if(current_write_position == MAXBUFF)
      {
         switch_write_buffer();
      }
   }
}

void get_stat(struct stat * file_stat)
{
   int i;
   int struct_stat_size;
   int struct_stat_pos=0;
   int temp_size;
   uint8_t * conv;
   
   conv = (uint8_t *) file_stat;
   
   struct_stat_size = sizeof(struct stat);
   
   while(struct_stat_size>0)
   {
      if(current_write_position + struct_stat_size <= MAXBUFF)
      {
         memcpy(&(conv[struct_stat_pos]), &(buffers[current_write_buffer].buff[current_write_position]), struct_stat_size);
         current_write_position += struct_stat_size;
	 struct_stat_size=0;
      }
      else
      {
         temp_size = MAXBUFF - current_write_position;
         memcpy(&(conv[struct_stat_pos]), &(buffers[current_write_buffer].buff[current_write_position]), temp_size);
         current_write_position = MAXBUFF;
         struct_stat_size -= temp_size;
         struct_stat_pos += temp_size;
      }
      if(current_write_position == MAXBUFF)
      {
         switch_write_buffer();
      }
   }
}

void dump_file_data(struct stat * file_stat)
{
   off_t temp_size;
   off_t data_size;
   
   data_size = file_stat->st_size;

   while(data_size > 0)
   {
      if(current_write_position + data_size <= MAXBUFF)
      {
         current_write_position += data_size;
	 data_size = 0;
      }
      else
      {
         temp_size = MAXBUFF - current_write_position;
         current_write_position = MAXBUFF;
         data_size -= temp_size;
      }
      if(current_write_position == MAXBUFF)
      {
         switch_write_buffer();
      }
   }
}

void dump_data(off_t size)
{
   off_t temp_size;

   while(size > 0)
   {
      if(current_write_position + size <= MAXBUFF)
      {
         current_write_position += size;
         size = 0;
      }
      else
      {
         temp_size = MAXBUFF - current_write_position;
         current_write_position = MAXBUFF;
         size -= temp_size;
      }
      if(current_write_position == MAXBUFF)
      {
         switch_write_buffer();
      }
   }
}

int write_data(int fd, struct stat * file_stat)
{
   off_t temp_size;
   off_t data_size;
   int n,result;
   
   data_size = file_stat->st_size;
   
   while(data_size > 0)
   {
      if(current_write_position + data_size <= MAXBUFF)
      {
         n=0;
         while( n < data_size)
         {
            if( (result = write(fd,&(buffers[current_write_buffer].buff[current_write_position+n]),data_size-n)) < 0 )
            {
               dump_data(data_size - n);
	       return -1;
            }
	    n+=result;
         }
         current_write_position += data_size;
         data_size = 0;
      }
      else
      {
         temp_size = MAXBUFF - current_write_position;
         n=0;
         while( n < temp_size)
         {
	    if( (result = write(fd,&(buffers[current_write_buffer].buff[current_write_position+n]),temp_size-n)) < 0 )
            {
               dump_data(data_size - n);
	       return -1;
            }
	    n+=result;
         }
         current_write_position = MAXBUFF;
         data_size -= temp_size;
      }
      if(current_write_position == MAXBUFF)
      {
         switch_write_buffer();
      }
   }
   return 0;
}

void init_file_writer()
{
   current_write_buffer=0;
   current_write_position=0;
   nbdir=0;
   nbfile=0;
   nblnk=0;

   Pthread_mutex_lock(&(buffers[current_write_buffer].prod_mutex));
   while(buffers[current_write_buffer].prod == 1)
   {
      Pthread_cond_wait(&(buffers[current_write_buffer].prod_cond),&(buffers[current_write_buffer].prod_mutex));
   }
   Pthread_mutex_unlock(&(buffers[current_write_buffer].prod_mutex));
}

int switch_write_buffer()
{
   /* free buffer */
   Pthread_mutex_lock(&(buffers[current_write_buffer].buffer_full_mutex));
   buffers[current_write_buffer].buffer_full--;
   if(buffers[current_write_buffer].buffer_full == 0)
      Pthread_cond_signal(&(buffers[current_write_buffer].buffer_full_cond));
   Pthread_mutex_unlock(&(buffers[current_write_buffer].buffer_full_mutex));

   /* choose new buffer */
   current_write_buffer++;
   if(current_write_buffer == NBBUFF)
      current_write_buffer = 0;
         
   /* get new buffer */
   Pthread_mutex_lock(&(buffers[current_write_buffer].prod_mutex));
   while(buffers[current_write_buffer].prod == 1)
   {
      Pthread_cond_wait(&(buffers[current_write_buffer].prod_cond),&(buffers[current_write_buffer].prod_mutex));
   }
   Pthread_mutex_unlock(&(buffers[current_write_buffer].prod_mutex));
         
   current_write_position = 0;
}

int file_dump()
{
   int file_name_size;
   struct stat file_stat;
   int fd;
   
   if( (file_name_size = get_file_name_size()) == 0)
      return -1;
   
   dump_data(file_name_size);
   
   get_stat(&file_stat);
   
   if(S_ISLNK(file_stat.st_mode))
   {
      file_name_size = get_file_name_size();
      dump_data(file_name_size);
   }
   else if( S_ISREG(file_stat.st_mode) )
   {
      dump_file_data(&file_stat);
   }
   
   return 0;
   
}

int file_write_source_dest(const char * target)
{
   int file_count=0;
   int file_name_size;
   struct stat file_stat;
   char * sym_target;
   char * file_name;
   int fd;
   int directory_size;
   int result;

   directory_size = strlen(target);
   
   if( (file_name_size = get_file_name_size()) == 0)
      return 0;
  
   file_name = (char *) Malloc( (directory_size+file_name_size+1) * sizeof(char));

   strncpy(file_name,target,directory_size);
   file_name[directory_size]='/';
   directory_size++;
   
   get_file_name(&(file_name[directory_size]), file_name_size);
   
   get_stat(&file_stat);
   
   if(S_ISREG(file_stat.st_mode))
   {
      if( (fd = creat(target,file_stat.st_mode)) < 0)
      {
         if(errno != EISDIR)
         {
            std::cerr<<"create_eror : file "<<target<<" could not be created : "<<strerror(errno)<<"\n";
            dump_file_data(&file_stat);
            free(file_name);
            get_file_name_size();
            return 0;
         }
         if( (fd = creat(file_name,file_stat.st_mode)) < 0)
         {
            std::cerr<<"create_eror : file "<<file_name<<" could not be created : "<<strerror(errno)<<"\n";
            dump_file_data(&file_stat);
            free(file_name);
            get_file_name_size();
            return 0;
         }
         else if(verbose>0)
         {
            printf(">-- %s\n",file_name);
         }
      }
      else if(verbose>0)
      {
         printf(">-- %s\n",target);
      }
      if( write_data(fd, &file_stat) < 0)
      {
         std::cerr<<"write_error : file "<<file_name<<" could not be written : "<<strerror(errno)<<"\n";
         Close(fd);
         Unlink(file_name);
         free(file_name);
         get_file_name_size();
         return 0;
      }
      nbfile++;
      Close(fd);
      free(file_name);
      get_file_name_size();
      return 1;
   }
   else if( S_ISDIR(file_stat.st_mode) )
   {
      if( mkdir(target, file_stat.st_mode) < 0)
      {
         if(errno != EEXIST)
         {
            std::cerr<<"mkdir_eror : directory "<<target<<" could not be created : "<<strerror(errno)<<"\n";
            while(file_dump() >= 0)
            {
               ;
            }
            free(file_name);
            return 0;
         }
         else if( mkdir(file_name, file_stat.st_mode) < 0)
         {
            if(errno != EEXIST)
            {
               std::cerr<<"mkdir_eror : directory "<<file_name<<" could not be created : "<<strerror(errno)<<"\n";
               while(file_dump() >= 0)
               {
                  ;
               }
               free(file_name);
               return 0;
            }
         }
         while( (result = file_write(file_name)) >= 0)
         {
            if( result > 0)
            file_count++;
         }
	 nbdir++;
         free(file_name);
         return file_count;
      }     
      while( (result = file_write(target)) >= 0)
      {
         if( result > 0)
         file_count++;
      }
      nbdir++;
      free(file_name);
      return file_count;
   }
   else
   {
      free(file_name);
      while(file_dump() >= 0)
      {
         ;
      }
      return 0;
   }
}

int file_write(const char * target_directory)
{
   uint32_t file_name_size;
   uint8_t file_type;
   char * file_name;
   char * sym_target;
   int directory_size;
   FILE *output;
   struct stat file_stat;
   int n;
   int fd;
   
   directory_size = strlen(target_directory);

   if( (file_name_size = get_file_name_size()) == 0)
      return -1;
      
   file_name = (char *) Malloc( (directory_size+file_name_size+1) * sizeof(char));
   strncpy(file_name,target_directory,directory_size);
   file_name[directory_size]='/';
   directory_size++;
   
   get_file_name(&(file_name[directory_size]), file_name_size);

   if(verbose>0)
   {
      printf(">-- %s\n",file_name);
   }

   get_stat(&file_stat);

   if(S_ISLNK(file_stat.st_mode))
   {
      sym_target = (char *)Malloc( (file_stat.st_size + 1)*sizeof(char));
      file_name_size = get_file_name_size();
      get_file_name(sym_target, file_name_size);
      if( symlink(sym_target,file_name) < 0 )
      {
         if(errno==EEXIST )
         {
            unlink(file_name);
            if( symlink(sym_target,file_name) < 0 )
            {
               std::cerr<<"symlink_eror : link "<<file_name<<" could not be created : "<<strerror(errno)<<"\n";
               free(sym_target);
               free(file_name);
               return 0;
            }
         }
         else
         {
            std::cerr<<"symlink_eror : link "<<file_name<<" could not be created : "<<strerror(errno)<<"\n";
            free(sym_target);
            free(file_name);
            return 0;
         }
      }
      nblnk++;
      free(sym_target);
      free(file_name);
      return 1;
   }
   else if( S_ISREG(file_stat.st_mode) )
   {
      if( (fd = creat(file_name,file_stat.st_mode)) < 0)
      {
         std::cerr<<"create_eror : file "<<file_name<<" could not be created : "<<strerror(errno)<<"\n";
         dump_file_data(&file_stat);
         free(file_name);
         return 0;
      }
      if( write_data(fd, &file_stat) < 0 )
      {
         std::cerr<<"write_error : file "<<file_name<<" could not be written : "<<strerror(errno)<<"\n";
         Close(fd);
         Unlink(file_name);
         free(file_name);
         return 0;
      }
      nbfile++;
      Close(fd);
      free(file_name);
      return 1;
   }
   else if( S_ISDIR(file_stat.st_mode) )
   {
      if( mkdir(file_name, file_stat.st_mode) < 0)
      {
         if(errno != EEXIST)
         {
            std::cerr<<"mkdir_eror : directory "<<file_name<<" could not be created : "<<strerror(errno)<<"\n";
            free(file_name);
            return 0;
         }          
      }
      nbdir++;
      free(file_name);
      return 0;
   }
   free(file_name);
   return 0;
}

void * file_writer(void * target_directory)
{
   int file_count=0;
   int result;
   struct stat file_stat;
   
   init_file_writer();

   if(source_dest)
   {
      file_count = file_write_source_dest((const char *)target_directory);
      if(current_write_position>0)
      {
         current_write_buffer++;
         if(current_write_buffer == NBBUFF)
            current_write_buffer = 0;

      }
      if( verbose > 0 || reportflag >0)
      {
         printf("%d files written\n",file_count);
         printf("%d regular files written\n",nbfile);
         printf("%d symlink written\n",nblnk);
         printf("%d directory written\n",nbdir);
      }
      return NULL;
   }
   if( stat((const char *)target_directory,&file_stat) < 0)
   {
      std::cerr<<"stat_error : target directory : "<<target_directory<<" could not be opened : "<<strerror(errno)<<"\n";
      while(file_dump() >= 0)
      {
         ;
      }
      if(current_write_position>0)
      {
         current_write_buffer++;
         if(current_write_buffer == NBBUFF)
            current_write_buffer = 0;
      }
      if( verbose > 0)
      {
         printf("%d files written\n",file_count);
         printf("%d regular files written\n",nbfile);
         printf("%d symlink written\n",nblnk);
         printf("%d directory written\n",nbdir);
      }
      return NULL;	 
   }
   else if( ! ( S_ISDIR(file_stat.st_mode) ) )
   {
      std::cerr<<"Target directory : "<<target_directory<<" is not a directory"<<strerror(errno)<<"\n";

      while(file_dump() >= 0)
      {
         ;
      }

      if(current_write_position>0)
      {
         current_write_buffer++;
         if(current_write_buffer == NBBUFF)
            current_write_buffer = 0;
      }
      if( verbose > 0 || reportflag >0)
      {
         printf("%d files written\n",file_count);
         printf("%d regular files written\n",nbfile);
         printf("%d symlink written\n",nblnk);
         printf("%d directory written\n",nbdir);
      }
      return NULL;
   }

   if(verbose>0)
   {
      printf("Target directory : %s\n",target_directory);
   }
   while( (result = file_write((const char *)target_directory)) >= 0)
   {
      if( result > 0)
         file_count++;
   }
   if(current_write_position>0)
   {
      current_write_buffer++;
      if(current_write_buffer == NBBUFF)
         current_write_buffer = 0;
   }
   if( verbose > 0 || reportflag >0)
   {
      printf("%d files written\n",file_count);
      printf("%d regular files written\n",nbfile);
      printf("%d symlink written\n",nblnk);
      printf("%d directory written\n",nbdir);
   }
   return NULL;
}
