#include <inuk/taktuk/mput_network.h>
#include <inuk/taktuk/mput_wrappers.h>
#include <string.h>

int connect_to(struct target_mput * target)
{
   struct sockaddr_in servaddr;
   target->sockfd = Socket(AF_INET,SOCK_STREAM,0);

   bzero(&servaddr,sizeof(servaddr));
   servaddr.sin_family = AF_INET;
   servaddr.sin_port = target->port;
   servaddr.sin_addr.s_addr = target->addr;

   Connect(target->sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr));

   return target->sockfd;   
}

char ** my_addrs(int *addrtype)
{
   struct hostent *hptr;
   struct utsname myname;
   
   if( uname(&myname) < 0 )
      return(NULL);

   if( (hptr = gethostbyname(myname.nodename)) == NULL )
      return(NULL);

   *addrtype = hptr->h_addrtype;
   return(hptr->h_addr_list);
}

uint32_t get_ip_addr(char *ip_addr_str)
{
   char **ip_addrs,**ip_addrs2;
   int addr_type;
   uint32_t ip_addr;
   /*find our ip adress*/
   if( (ip_addrs = my_addrs(&addr_type)) == NULL )
   {
      std::cerr<<"cannot_find_adress_error\n";
      exit(1);
   }
   if( addr_type != AF_INET )
   {
      std::cerr<<"protocol_not_supported_error\n";
      exit(1);
   }
   ip_addrs2=ip_addrs;
   if(verbose>1)
   {
      printf("Possible adresses :\n");
      while( *ip_addrs != NULL )
      {
         Inet_ntop(addr_type,*ip_addrs,ip_addr_str,MAXADDRSIZE);
         printf("Adresse : %s\n",ip_addr_str);
         ip_addrs++;
      }
   }
   /*use the first ip adress*/
   if( *ip_addrs2 != NULL )
   {
      if(verbose>1)
      {
         printf("Using first adress in the list.\n");
         fflush(stdout);
      }
      Inet_ntop(addr_type,*ip_addrs2,ip_addr_str,MAXADDRSIZE);
   }
   else
   {
      std::cerr<<"no_network_interface_found_error\n";
      exit(1);
   }
   Inet_pton(AF_INET,ip_addr_str,&ip_addr);
   return ntohl(ip_addr);
}

int create_listening_socket(struct sockaddr_in *servaddr)
{
   int listenfd;
   
   /*create a socket to listen on the network */   
   listenfd = Socket(AF_INET,SOCK_STREAM,0);

   bzero(servaddr, sizeof(*servaddr));
   servaddr->sin_family = AF_INET;
   servaddr->sin_addr.s_addr =  htonl(INADDR_ANY);
   /* socket is assigned dynamically */
   servaddr->sin_port = htons(0);

   Bind(listenfd,(struct sockaddr *) servaddr, sizeof(*servaddr));
   
   /*then set it to listen*/
   Listen(listenfd, LISTENQUEUE);
   
   return listenfd;
}

uint16_t get_port_assigned(struct sockaddr_in *servaddr,int listenfd)
{
   uint16_t port_number;
   socklen_t addr_len;
   /*get info about the sport assigned*/
   addr_len = sizeof(*servaddr);
   
   Getsockname(listenfd, (struct sockaddr *) servaddr, &addr_len);
   
   port_number=ntohs(servaddr->sin_port);
   if(verbose>1)
   {
      printf("Port Number Assigned : %u\n",port_number);
   }
   return port_number;
}
