#include <inuk/taktuk/mput_utils.h>
#include <inuk/taktuk/mput_wrappers.h>
#include <inuk/taktuk/mput_receiver.h>
#include <inuk/taktuk/mput_sender.h>
#include <inuk/taktuk/mput_file_writer.h>
#include <inuk/taktuk/mput_file_reader.h>
#include <inuk/taktuk/mput_buffers.h>
#include <inuk/taktuk/mput_network.h>
#include <dirent.h>
#include <string.h>
#include <ctype.h>

void print_usage(void)
{
   std::cerr<<"arguments : [-v level] [-s] [-r] -- arguments_2\n";
   std::cerr<<"\t\tv : level of verbosity, default is 0\n";
   std::cerr<<"\t\ts : sync() at end of program\n";
   std::cerr<<"\t\tr : report job done at end of process\n";
   std::cerr<<"\targuments_2 :\n";
   std::cerr<<"\t\tfile ... file directory\n";
   std::cerr<<"\t\tfile_source file_dest\n";
}

int parse_optional_arguments(int argc, char **argv)
{
   int c;

   syncflag = 0;
   verbose = 0;
   reportflag = 0;
   
   while( (c = getopt (argc, argv, "rsv:")) != -1 )
   {
      switch(c)
      {
      case 's':
         syncflag = 1;
         break;
      case 'r':
         reportflag = 1;
         break;
      case 'v':
         verbose = atoi(optarg);
         break;
      case '?':
         if (isprint (optopt))
            std::cerr<<"Unknown option `-"<<(char)optopt<<"'.\n";
         else
            std::cerr<<"Unknown option character `\\x"<<optopt<<"'.\n";
         net->usage();
         print_usage();
         net->terminate();
         exit(1);
         break;
     default:
         abort();
      }
   }
   return optind;
}

void verify_arguments(int index, int argc, char** argv)
{
   source_dest = 0;
   if( argc - index < 2 )
   {
      std::cerr<<"Not enough parameters\n";
      net->usage();
      print_usage();
      net->terminate();
      exit(1);
   }
   else if(argc - index == 2)
   {
      source_dest = 1;
   }
   
}

void print_target(struct target_mput * target)
{
   char addrstr[MAXADDRSIZE];

   Inet_ntop(AF_INET,&(target->addr),addrstr,MAXADDRSIZE);
   printf("%s %d %d\n",addrstr,ntohs(target->port),target->father);
    
}

void print_targets(struct target_mput * targets)
{
   int i=0;

   struct target_mput vide = {0,0,0};
   while(compare_targets(&vide,&targets[i]) != 0)
   {
      print_target(&targets[i]);
      i++;
   }
}

int compare_targets(struct target_mput * target1,struct target_mput * target2)
{
   if( htonl(target1->addr) < htonl(target2->addr) )
      return -1;
   else if ( htonl(target1->addr) > htonl(target2->addr) )
      return 1;
   else
   {
      if( htons(target1->port) < htons(target2->port) )
         return -1;
      else if( htons(target1->port) > htons(target2->port) )
         return 1;
      else
         return 0;      
   }
}

void _quick_sort_targets(struct target_mput * start, struct target_mput * end)
{
   int high=1;
   struct target_mput *s,*e,p;
   if(start==end)
      return;
   
   s=start;
   e=end;
   p=*s;
   e--;
   
   while(e!=s)
   {
      if(high)
      {
         if(compare_targets(e,&p)>0)
         {
            e--;
         }
         else
         {
            *s=*e;
            s++;
            high=0;
         }
      }
      else
      {
         if(compare_targets(s,&p) <= 0)
         {
            s++;
         }
         else
         {
            *e=*s;
            e--;
            high=1;
         }      
      }
   }
   *e=*end;
   s++;
   _quick_sort_targets(start,e);
   _quick_sort_targets(s,end);
   *e=p;   
}

void quick_sort_targets(struct target_mput * targets)
{
   struct target_mput vide={0,0,0,0,0};
   
   int count=0;
   while(compare_targets(&(targets[count]),&vide) != 0)
   {
      count++;
   }
   _quick_sort_targets(targets,targets+count);
     
}

struct target_mput * create_host_sorted_list(struct target_mput **input,int target_count)
{
   struct target_mput *output=NULL;
   struct target_mput vide = {0,0,0,0,0};
   int i;

   output = (struct target_mput *)Malloc((target_count+1)*sizeof(struct target_mput));
   
   for(i=0;i<target_count;i++)
   {
      output[i]=*(input[i]);
      free(input[i]);
   }

   free(input);

   output[target_count] = vide;
   
   quick_sort_targets(output);
   
   if(verbose>1)
   {
      printf("List of targets :\n");
      print_targets(output);
   }   
   
   return output;
}

void find_relatives(struct target_mput * targets)
{
   int i=0;
   int last=0;
   struct target_mput vide = {0,0,0,0,0};
   
   father = new(struct target_mput);
   bbrother = new(struct target_mput);
   lbrother = new(struct target_mput);
   
   father->addr = 0;
   bbrother->addr = 0;
   lbrother->addr = 0;
   
   while( compare_targets(&vide,&targets[i]) != 0 )
   {
      if(targets[i].father == 1)
      {
         if(father->addr == 0)
            *father = targets[i];
         else
         {
            std::cerr<<"multiple_father_error\n";
            net->terminate();
            exit(1);
         }
      }
      if( compare_targets(me,&targets[i]) == 0 )
      {
         if( compare_targets(&vide,&targets[i+1]) == 0 )
            *lbrother = targets[0];
         else
            *lbrother = targets[i+1];

         if( i == 0 )
            last = 1;
         else
            *bbrother = targets[i-1];
      }
      i++;
   }
   if(last)
      *bbrother = targets[i-1];
   if( compare_targets(&vide,father) == 0 )
   {
      std::cerr<<"no_father_error\n";
      net->terminate();
      exit(1);
   }
}

void print_relatives()
{
   printf("I am :\n");
   print_target(me);
   printf("My father is :\n");
   print_target(father);
   printf("My big brother is : \n");
   print_target(bbrother);
   printf("My litle brother is : \n");
   print_target(lbrother);
}

struct target_mput * get_targets(struct target_mput * hosts_list)
{
   struct target_mput vide={0,0,0,0,0};
   struct target_mput *result;
   /* temporaire */
   
   result = (struct target_mput *)Malloc(2*sizeof(struct target_mput));
   
   result[0] = vide;
   result[1] = vide;
   
   find_relatives(hosts_list);

   if(father_flag)
   {
      if( compare_targets(me,lbrother) == 0 )
      {
         std::cerr<<"error : only one node\n";
	     net->terminate();
         exit(1);
      }
      accept_connection = 0;
      result[0] = *lbrother;
   }
   else if( compare_targets(father,lbrother) == 0 )
   {
      accept_connection = 1;
   }
   else
   {
      accept_connection = 1;
      result[0] = *lbrother;
   }
   return result;
}

void send_receive_files(struct target_mput * targets, struct files * sources, char * target_directory)
{
   int i=0;
   struct target_mput vide={0,0,0,0,0};
   pthread_t file_tid,receiver_tid;
   int connfd;

   check_point = new(struct end_point);

   check_point->finished = 0;
   check_point->buff_number = 0;
   
   init_buffers();
  
   while( compare_targets(&(targets[i]), &(vide)) != 0)
   {
      nb_consumers++;
      connect_to(&(targets[i]));
      Pthread_create(&(targets[i].tid), NULL, sender, &(targets[i].sockfd));
      i++;
   }
   if(!father_flag)
   {
      nb_consumers++;
      Pthread_create(&file_tid, NULL, file_writer, target_directory);
   }
   if(accept_connection)
   {
      connfd = Accept(me->sockfd, (struct sockaddr *) NULL, NULL);
      Pthread_create(&receiver_tid, NULL, receiver, &connfd);
   }
   if(father_flag)
   {
      Pthread_create(&file_tid, NULL, file_reader, sources);
   }

   Pthread_join(file_tid, NULL);
   
   
   if(father_flag)
   {
      Pthread_mutex_lock(&(buffers[current_read_buffer].prod_mutex));
      check_point->buff_number = current_read_buffer;
   }
   else
   {
      Pthread_mutex_lock(&(buffers[current_write_buffer].prod_mutex));
      check_point->buff_number = current_write_buffer;
   }
   check_point->finished = 1;
   Pthread_mutex_unlock(&(buffers[check_point->buff_number].prod_mutex));
   
   Pthread_cond_broadcast(&(buffers[check_point->buff_number].prod_cond));   
  
   if(accept_connection)
   {
      receiver_finished=1;
      Close(connfd);
   }
   i=0;   
   while( compare_targets(&(targets[i]), &(vide)) != 0)
   {
      Pthread_join(targets[i].tid, NULL);
      Close(targets[i].sockfd);
      i++;
   }      
   
}
