#include <inuk/taktuk/mput_wrappers.h>
#include <string.h>

void error_report(const char * message, int error_number)
{
   std::cerr<<message<<" : "<<strerror(error_number)<<"\n";
   net->terminate();
   exit(1);
}

void *Malloc(size_t size)
{
   void *result;
   if( (result = malloc(size)) == NULL)
   {
      error_report("malloc_error",errno);
   }
   return result;   
}

FILE *Fopen(const char *path, const char *mode)
{
   FILE *result;
   if( (result = fopen(path,mode)) == NULL)
   {
      error_report("fopen_error",errno);
   }
   return result;
}

int Creat(const char *pathname, mode_t mode)
{
   int result;
   if( (result = creat(pathname, mode)) < 0 )
   {
      error_report("creat_error",errno);
   }
   return result;
}

int Stat(const char *pathname, struct stat *buf)
{
   int result;
   if( (result = stat(pathname, buf)) < 0 )
   {
      error_report("stat_error",errno);
   }
   return result;
}

int Fclose(FILE *stream)
{
   int result;
   if( (result = fclose(stream)) == EOF )
   {
      error_report("fclose_error",errno);
   }
   return result;
}

DIR *Opendir(const char *name)
{
   DIR *result;
   if( (result = opendir(name)) == NULL )
   {
      error_report("opendir_error",errno);
   }
   return result;
}

int Closedir(DIR *dir)
{
   int result;
   if( (result = closedir(dir)) < 0)
   {
      error_report("closedir_error",errno);
   }
   return result;
}

const char *Inet_ntop(int af, const void *src, char *dst, socklen_t cnt)
{
   const char *result;
   if( (result = inet_ntop(af, src, dst, cnt)) == NULL)
   {
      error_report("socket_error",errno);
   }
   return result;
}

int Inet_pton(int af, const char *src, void *dst)
{
   int result;
   if( (result = inet_pton(af, src, dst)) < 0)
   {
      error_report("socket_error",errno);
   }
   else if (result == 0)
   {
      std::cerr<<"inet_pton_error : Invalid Adress\n";
      exit(1);
   }
   return result;
}

int Socket(int domain, int type, int protocol)
{
   int result;
   if( (result = socket(domain, type, protocol)) < 0)
   {
      error_report("socket_error",errno);
   }
   return result;
} 

int Connect(int  sockfd,  const  struct sockaddr *serv_addr, socklen_t addrlen)
{
   int result;
   if( (result = connect(sockfd, serv_addr, addrlen)) < 0)
   {
      error_report("connect_error",errno);
   }
   return result;
}

int Bind(int sockfd, struct sockaddr *my_addr, socklen_t addrlen)
{
   int result;
   if( (result = bind(sockfd, my_addr, addrlen)) < 0)
   {
      error_report("bind_error",errno);
   }
   return result;
}

int Listen(int s, int backlog)
{
   int result;
   if( (result = listen(s, backlog)) < 0)
   {
      error_report("listen_error",errno);
   }
   return result;
}

int Accept(int s, struct sockaddr *addr, socklen_t *addrlen)
{
   int result;
   if( (result = accept(s, addr, addrlen)) < 0)
   {
      error_report("accept_error",errno);
   }
   return result;
}

int Close(int fd)
{
   int result;
   if( (result = close(fd)) < 0)
   {
      error_report("close_error",errno);
   }
   return result;
}

int Getsockname(int s, struct sockaddr *name, socklen_t *namelen)
{
   int result;
   if( (result = getsockname(s, name, namelen)) < 0)
   {
      error_report("getsockname_error",errno);
   }
   if( *namelen != sizeof(*name) )
   {
      std::cerr<<"getsockname_error : Sockaddr Overwritten Error\n";
      exit(1);
   }
   return result;
}

ssize_t Read(int fd, void *buf, size_t count)
{
   ssize_t result;
   if( (result = read(fd, buf, count)) < 0 )
   {
      error_report("read_error",errno);
   }
   return result;
}

ssize_t Write(int fd, const void *buf, size_t count)
{
   ssize_t result;
   if( (result = write(fd, buf, count)) < 0 )
   {
      error_report("write_error",errno);
   }
   return result;
  
}

int Pthread_mutex_init(pthread_mutex_t  *mptr,  const  pthread_mutexattr_t *maptr)
{
   int result;
   if( (result = pthread_mutex_init(mptr,maptr)) != 0)
   {
         error_report("pthread_mutex_init_error",result);
   }
   return result;
}

int Pthread_mutex_lock(pthread_mutex_t *mptr)
{
   int result;
   if( (result = pthread_mutex_lock(mptr)) != 0)
   {
         error_report("pthread_mutex_lock_error",result);
   }
   return result;
}

int Pthread_mutex_unlock(pthread_mutex_t *mptr)
{
   int result;
   if( (result = pthread_mutex_unlock(mptr)) != 0)
   {
         error_report("pthread_mutex_unlock_error",result);
   }
   return result;
}

int Pthread_cond_init(pthread_cond_t *cptr, pthread_condattr_t *captr)
{
   int result;
   if( (result = pthread_cond_init(cptr,captr)) != 0)
   {
         error_report("pthread_cond_init_error",result);
   }
   return result;
}

int Pthread_cond_wait(pthread_cond_t *cptr,pthread_mutex_t *mptr)
{
   int result;
   if( (result = pthread_cond_wait(cptr,mptr)) != 0)
   {
         error_report("pthread_cond_wait_error",result);
   }
   return result;
}

int Pthread_cond_signal(pthread_cond_t *cptr)
{
   int result;
   if( (result = pthread_cond_signal(cptr)) != 0)
   {
      error_report("pthread_cond_signal_error",result);
   }
   return result;
}

int Pthread_cond_broadcast(pthread_cond_t *cptr)
{
   int result;
   if( (result = pthread_cond_broadcast(cptr)) != 0)
   {
      error_report("pthread_cond_broadcast_error",result);
   }
   return result;
}

int Pthread_create(pthread_t *tid, const pthread_attr_t *attr, void *(*start_routine)(void *), void *arg)
{
   int result;
   if( (result = pthread_create(tid, attr, start_routine, arg)) != 0)
   {
      error_report("pthread_create_error",result);
   }
   return result;
}

int Pthread_join(pthread_t tid, void **status)
{
   int result;
   if( (result = pthread_join(tid, status)) != 0)
   {
      error_report("pthread_join_error",result);
   }
   return result;
}

int Readlink(const char * path, char *buf, size_t bufsize)
{
   int result;
   if( (result = readlink(path, buf, bufsize)) < 0)
   {
      error_report("readlink_error",errno);
   }
   return result;
}

int Symlink(const char *oldpath, const char *newpath)
{
   int result;
   if( (result = symlink(oldpath, newpath)) < 0)
   {
      error_report("symlink_error",errno);
   }
   return result;
}

int Unlink(const char * path)
{
   int result;
   if( (result = unlink(path)) < 0)
   {
      error_report("unlink_error",errno);
   }
   return result;
}
