#include <inuk/acom>
#include <inuk/taktuk/reduce.h>
#include <inuk/taktuk/unp.h>
#include <inuk/taktuk/mput.h>
#include <inuk/taktuk/mput_utils.h>
#include <inuk/taktuk/mput_network.h>

int father_flag;
int verbose;
int syncflag;
int reportflag;
int accept_connection;
int source_dest;

volatile struct end_point *check_point;

struct target_mput *me;

struct target_mput *father;
struct target_mput *bbrother;
struct target_mput *lbrother;

struct target_mput_temp *me_temp;

Taktuk::Network *net;

int main(int argc, char* argv[])
{
   std::string args="";
   Reduction R;
   int index;
   struct files file_list;
   int i,err;
   int nbelem;
   struct target_mput *connection_targets;
   struct target_mput *target_list;
   char ip_addr_str[MAXADDRSIZE];
   char path[MAXPATHLENGTH];
   uint16_t port_number;
   uint32_t ip_addr;
   struct sockaddr_in servaddr;
   int listenfd;
   struct target_mput **hosts;

   me = new(struct target_mput);

   memset(me,0,sizeof(struct target_mput));

   net = new Taktuk::Network;

   R.init(net);
   
   if( (err=net->initialize(&argc,&argv,true,true,true,false)) < 0)
   {
      switch (err)
      {
      case -1:
         std::cerr<<"mput";
         net->usage();
         print_usage();
         break;
      case -2:
         std::cerr<<"mput can't redirect standard IO"<<std::endl;
         break;
      case -3:
         std::cerr<<"mput can't launch internal control threads: no more ressources"<<std::endl;
         break;
      }
      net->terminate();
      exit(1);
   }

   index = parse_optional_arguments(argc,argv);

   verify_arguments(index,argc,argv);

   if(argv[argc-1][0]!='/')
   {
      if( getcwd(path,MAXPATHLENGTH) == NULL )
      {
         std::cerr<<"mput current path too long, try using absolute directories\n";
         net->terminate();
         exit(1);
      }
      for(i=1;i<argc-1;i++)
      {
         args = ( args + argv[i] ) +" ";  
      }
      args = (( args + path ) +"/") + argv[i];
      
      net->newarg(args);
   }

   net->commit();

   if(verbose>1)
   {  
      i=0;
      printf("arguments passed :\n");
      while(i<argc)
      {
         printf("%s ",argv[i]);
	 i++;
      }
      printf("\n");
   }
   
   father_flag = (net->rank() == 0);
   
   ip_addr = get_ip_addr(ip_addr_str);

   listenfd = create_listening_socket(&servaddr);
  
   port_number = get_port_assigned(&servaddr, listenfd);
 
   me->addr = htonl(ip_addr);
   me->port = htons(port_number);
   me->father = father_flag;
   me->sockfd = listenfd;
   
   hosts = (struct target_mput **)R.reduce(me,sizeof(struct target_mput),&nbelem);
   
   if(verbose > 1)
   {
      printf("Received hosts :\n");
      for(i=0;i<nbelem;i++)
      {
         print_target((struct target_mput *)(hosts[i]));
      }
   }

   target_list = create_host_sorted_list(hosts,nbelem);
   
   connection_targets = get_targets(target_list);

   file_list.file_count = argc-index-1;
   file_list.file_name = &(argv[index]);
 
   if(verbose > 0)
   {
      printf("Connecting to :\n");
      print_targets(connection_targets);
   }
   
   send_receive_files(connection_targets,&file_list, argv[argc-1]);

   if(syncflag)
   {
      if(verbose > 0)
      {
         printf("Syncing...\n");
      }
      sync();
   }
   fflush(stdout);
   net->terminate(); 
     
   return 0;
}
