/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#include "inuk/taktuk/taktuknetwork.h"
#include "inuk/taktuk/debuglog.h"

#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>  
#include <errno.h>
#include <string.h>
#include <inuk/taktuk/panichandler.h>
#include <inuk/pth>

#define setenv(var,val,overwite)\
do { \
	char tmp_str[256]; \
	sprintf(tmp_str,"%s=%s",var,val); \
	putenv(tmp_str); \
} while (0)

class tukpanic : public PanicHandler
{
  public:
    tukpanic()
    {
      _pid=0;
			mut.init();
    }
		~tukpanic()
		{
			mut.destroy();
		}
    void handle()
    {
			mut.lock();
      if(_pid!=0)
        kill(_pid,SIGKILL);
			mut.unlock();
      exit(0);
    }
    void setpid(int pid)
    {
			mut.lock();
      _pid=pid;
			mut.unlock();
    }
  private:
    int _pid;
		PTH::Mutex mut;
};


int main(int argc, char *argv[])
{
  std::string args="'";
  Taktuk::Network taktuk;
  tukpanic _panic;
  taktuk.SetHandler(_panic);
  int err=0;
  if((err=taktuk.initialize(&argc, &argv,true,true,true,false))<0)
    {
      switch (err)
	{
	case -1:
	  std::cerr<<"rshp";
	  taktuk.usage();
	  break;
	case -2:
	  std::cerr<<"rshp can't redirect standard IO"<<std::endl;
	  break;
	case -3:
	  std::cerr<<"rshp can't launch internal control threads: no more ressources"<<std::endl;
	  break;
      }
			taktuk.terminate();
      exit(1);
    }

	//std::cout<<"nombre d'arg "<<argc<<std::endl;
	//for(int i=0;i<argc;i++)
	//	std::cout<<"argv["<<i<<"] = |"<<argv[i]<<"|"<<std::endl;
		
  if(argc>1)
    {
			for(int i=1;i<argc;i++)
				args= (args + argv[i])+" ";
			args=args + "'";
			taktuk.newarg(args);
		}

  if(taktuk.commit()<0)
    {
      std::cerr<<"rshp can't connect several hosts :"<<std::endl;
      taktuk.print_rejected();
    }


  int child_pid;
  char num_node[10];
  struct passwd *pwd;

	//std::cout<<"argc= "<<argc<<std::endl;

  argv = &argv[1];
  argc-=1;


  snprintf(num_node,10,"%d",taktuk.rank());
  setenv("TAKTUK_RANK",(const char*)num_node,1);


	//std::cout<<"argc-1 = "<<argc<<std::endl;
  if(taktuk.rank()!=0)
    {

      if ((child_pid = fork()) == -1)
	{
	  perror("rshp: fork error");
	  taktuk.terminate();;
	  exit(1);
	}
      
      if (child_pid != 0)
	{
	  _panic.setpid(child_pid);
	  //	  cout<<"pere attend fils pid: "<<child_pid<<endl;
		close(0); // close the in it's redirect to the forked process
	  taktuk.terminate();;
	  kill(child_pid,SIGKILL);
	}
      else
	{
	  setuid(getuid());
	  setgid(getgid());
	  if((pwd=getpwuid(getuid()))== (struct passwd* )NULL)
	    {
	      perror("get user env: ");
	      sleep(5);
	      exit(1);
	    }
	  if (chdir(pwd->pw_dir) < 0)
	    {
	      (void) chdir("/");
	    }
	  if (*pwd->pw_shell == '\0')
	    pwd->pw_shell = "/bin/csh";
	//	pwd->pw_shell = "/bin/bash";

	  char *argu[4];
	  std::string argument;
	  argu[0]=pwd->pw_shell;
	  //argu[0]="/bin/tcsh";
	  argu[1]="-c";

	  if(argc>0)
	    {
	      for (int i=0; i<argc; i++)
					argument= argument + argv[i]+ " ";
	    }
	  else
	    {
	      argument=pwd->pw_shell;
				argument+= " -s";
	    }

		char* cmd;
		cmd=strcpy((char*)malloc(argument.size()+1),(char*)argument.c_str());
		//argu[2]=(char*)argument.c_str();
		argu[2]=cmd;
		argu[3]=NULL;
		//	std::cout<<"exec :["<<argu[2]<<"]"<<std::endl;
	  execvp(argu[0],argu);
	  std::cout<<"exec error: "<<strerror(errno)<<std::endl;
	}
    }
  else
	{
		close(0);// close the input i'm the node 0 i don't forj any processes
    taktuk.terminate();
		shutdown(0,SHUT_RDWR);
		shutdown(1,SHUT_RDWR);
		shutdown(2,SHUT_RDWR);
	}
}
