/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#include <inuk/taktuk/topology.h>
#include <inuk/taktuk/scheduler.h>
#include <inuk/acom>
#include <inuk/taktuk/taktuknetwork.h>
#include <inuk/sock>
#include <inuk/taktuk/debuglog.h>

void scheduler::initialize(std::vector<topology::host*>* hlist, int hlistsize, InfoTCPnet* i, int maxthread, Taktuk::Network* net)
{
	info=i;
	_net=net;
	_maxthread=maxthread;
	_originalhostmap=hlist;
	//record nodes in ToDolist
	
	for(int i=0;i<hlistsize&&i<hlist->size();i++)
	{
		ToDoList.put((*hlist)[i]);
	}
	
}

void scheduler::RefreschInfo(bool force)
{
	GetChildsRank();
	SendChildsInfo();
	dbg.log("scheduler","RefreschInfo rank = %d count nodes =%d\n",info->rank,info->count_nodes);
	if(info->rank==0)
		info->count_nodes=treesize;
	GetRouteTable();
	childs_map_mut.lock();
	*_originalhostmap = childs_map;
	childs_map_mut.unlock();
	if(force)
		_net->RefreshInfo(*info);
	else
		_net->LoadInfo(*info);
	_net->RefreschNetwork();

}


void scheduler::terminate()
{
	--spawncompleted;
	infochilds.wait();
	StealRequestList_mut.lock();
	End=true;
	StealRequestList_cond.signal();
	StealRequestList_mut.unlock();
	join.wait();
	dbg.log("scheduler","Terminate thread stopped !!!\n");
}


int scheduler::get_job(std::vector<topology::host*>& hlist,int _stealnum)
{
	ToDoList_mut.lock();
	
	dbg.log("scheduler","-----------GETJOB----BEGIN--------\n");
	dbg.log("scheduler","get_job!!! todolist size =%d, stealnum = %d \n",ToDoList.size(),_stealnum);
	if(ToDoList.empty()&&!nojob)
	{
		dbg.log("scheduler","ToDoList empty\n");
		//request on up node for job and wait it's reply
		//if we are not the root of the tree
		if(info->rank!=0)
		{
			dbg.log("scheduler","STEAL CALLED\n");
			acom::ActiveMessage<steal> am (_net, &_steal, 0);
			//childs_map_mut.lock();
			//am->pack((void*)&treesize,sizeof(acom::Network::LocalNodeNo));
			//childs_map_mut.unlock();
			_net->_LocalChannel_mut.lock();
			Taktuk::Network::TaktukLocalChannel* reply=_net->LocalChannel[0];
			_net->_LocalChannel_mut.unlock();
			am.post(reply);
			ToDoList_cond_mut.lock();
			while(ToDoList.empty()&&!nojob)
				ToDoList_cond.wait(ToDoList_cond_mut);
			ToDoList_cond_mut.unlock();
		}
	}
	for(int i=0;i<_stealnum&&!ToDoList.empty();i++)
		hlist.push_back(ToDoList.pop());
	

	dbg.log("scheduler","get_job new job hlist size %d\n",hlist.size());
	dbg.log("scheduler","-----------GETJOB----END--------\n");
	ToDoList_mut.unlock();
	return(hlist.size());
}

void scheduler::add_childs(std::vector<topology::host*>& hlist)
{
	//we must add this node in childsmap & in info structure & in
	//_net->LocalChannel
	childs_map_mut.lock();
	dbg.log("scheduler","Thread scheduler: Add %d childs in	childs_map containing %d nodes\n",hlist.size(),childs_map.size());
	for(int i=0;i<childs_map.size();i++)
		dbg.log("scheduler","Thread scheduler: Add childs existing :  size %d\n",childs_map[i]->size);
	for(int i=0;i<hlist.size();i++)
	{
		// build childs_map
		SOCK::SocketClient* sock;
		PTH::Mutex* mut;
		childs_map.push_back(hlist[i]);
		// build info all_links
		sock = new SOCK::SocketClient();
		sock->set_fd(hlist[i]->fd);
        sock->setSockOpt(SOCK::Socket::no_delay,true);
		mut = new PTH::Mutex();
		mut->init();
//ATTENTION MODIF POUR TERMINATE LONG info->all_links.push_back(InfoTCPnet::link(sock,mut,InfoTCPnet::link::normal,hlist[i]->rank));
	
		dbg.log("","LOCK add childs\n");
		info->all_links_mut.lock();
		info->all_links.push_back(new InfoTCPnet::link(sock,mut,InfoTCPnet::link::free,hlist[i]->rank));
		info->all_links_mut.unlock();
		dbg.log("","UNLOCK add childs\n");
		
		// build Taktuk network LocalChannel
		_net->_LocalChannel_mut.lock();
		_net->LocalChannel.push_back(new Taktuk::Network::TaktukLocalChannel(_net,_net->LocalChannel.size(),info));
		_net->_LocalChannel_mut.unlock();
		++infochilds;
	}
	// update local treesize
	treesize=1;
	dbg.log("scheduler","Thread scheduler: before Add childs updated treesize \n");
	for(int i=0;i<childs_map.size();i++)
	{
		dbg.log("scheduler","Thread scheduler: treesize %d\n",i);
		dbg.log("scheduler","Thread scheduler: MAP size %d\n",childs_map.size());
		treesize+=childs_map[i]->size;
	}
	
	dbg.log("scheduler","Thread scheduler: Add childs updated treesize %d\n",treesize);
	// reload new infoTCPnet structure in the router
	_net->LoadInfo(*info);
	childs_map_mut.unlock();
	dbg.log("scheduler","Thread scheduler: Add childs OK\n");
}

void * sched_thread_entry(scheduler * sched)
{
	sched->run();
}

void scheduler::startscheduler()
{
	PTH::ThreadId tid;
	sched_thread.create(&tid, (void *(*)(void*)) sched_thread_entry, this, PTH::Thread::JOINABLE);
	// start();
}

void scheduler::run()
{
	int LocalID=-1;
	std::vector<topology::host*> hlist;
	dbg.log("scheduler","Thread scheduler begin\n");
	int begin=0;
	if(info->rank!=0)
		begin=1;
	while(true)
	{
		dbg.log("scheduler","Thread scheduler: wait for stealer\n");
		LocalID=WaitForStealer();
		childs_map_mut.lock();
		if(LocalID!=-1)
			dbg.log("scheduler","----------PROCEED A STEALER REQUEST child %d : %s --------\n",LocalID,childs_map[LocalID-begin]->name.c_str());
		else
			dbg.log("scheduler","----------PROCEED A STEALER REQUEST child %d  --------\n",LocalID);
		dbg.log("scheduler","Thread scheduler: childs_map size = %d\n",childs_map.size());
		childs_map_mut.unlock();

		if(LocalID!=-1)
		{
			childs_map_mut.lock();
			int inter = childs_map[LocalID-begin]->laststeal;
			childs_map_mut.unlock();
			get_job(hlist,inter);
			childs_map_mut.lock();
			if(childs_map[LocalID-begin]->laststeal<8) // laststael initialized to 1 by the constructor
				childs_map[LocalID-begin]->laststeal=2*childs_map[LocalID-begin]->laststeal;
			childs_map_mut.unlock();
		}
		if(!hlist.empty())
		{
			//send job to the node
			dbg.log("scheduler","Thread scheduler send job to child %d\n",LocalID);
			_net->_LocalChannel_mut.lock();
			Taktuk::Network::TaktukLocalChannel* reply = _net->LocalChannel[LocalID];
			_net->_LocalChannel_mut.unlock();
			acom::AllocWriteAndSignal<dojob> aws(_net, &_dojob, 0);
			dbg.log("scheduler","----------SEND JOB TO CHILD %d  --------\n",LocalID);
			for(int i=0;i<hlist.size();i++)
			{
				aws->put((void*)hlist[i]->name.c_str(),(hlist[i]->name.size()+1)*sizeof(char));
				aws->put((void*)hlist[i]->connector.c_str(),(hlist[i]->connector.size()+1)*sizeof(char));
				aws->put((void*)hlist[i]->taktukargs.c_str(),(hlist[i]->taktukargs.size()+1)*sizeof(char));
				aws->put((void*)hlist[i]->userargs.c_str(),(hlist[i]->userargs.size()+1)*sizeof(char));
				aws->put((void*)hlist[i]->login.c_str(),(hlist[i]->login.size()+1)*sizeof(char));
				aws->put((void*)hlist[i]->passwd.c_str(),(hlist[i]->passwd.size()+1)*sizeof(char));
				aws->put((void*)hlist[i]->cmd.c_str(),(hlist[i]->cmd.size()+1)*sizeof(char));
				aws->put((void*)&hlist[i]->size,sizeof(int));
				aws->put((void*)hlist[i]->gateaccess.c_str(),(hlist[i]->gateaccess.size()+1)*sizeof(char));
				aws->put((void*)&hlist[i]->timeout,sizeof(int));
				aws->put((void*)hlist[i]->optuserargs.c_str(),(hlist[i]->optuserargs.size()+1)*sizeof(char));
			}
			aws.post(reply);
		//don't forget to erase the list for next iteration
			hlist.clear();
		}
		else
		{
			dbg.log("scheduler","Thread scheduler NO MORE JOB after get_job\n");
			break;
		}
	}
	// no more job before send info to childs this node must wait for his
	// spawner and after wait for all his childs request info
	// after that it must reply info to all his childs and stop
	// no more job send nodes info
	
	dbg.log("scheduler","Thread scheduler wait for spawn completed\n");
	spawncompleted.wait();
	dbg.log("scheduler","Thread scheduler wait for child steal request and reply no more job\n");
	if(LocalID!=-1)
		SubscribeToSteal(LocalID);
	AutoReplyStealer(getnbchilds());
	

	dbg.log("scheduler","Thread scheduler get info requested\n");
	GetInfo();

	// we know who we are 
	dbg.log("scheduler","Thread scheduler childs rank computed\n");
	// all done we can die
	dbg.log("scheduler","Thread scheduler: terminated stopped ..........\n");
	--join;
}

void scheduler::GetInfo()
{
	
	//wait for childs getinfo
	nbInfoReq_mut.lock();
	nbInfoReq+=getnbchilds();
	dbg.log("scheduler","WAIT FOR GET INFO REQUEST waiting for %d nbchilds %d\n",nbInfoReq,getnbchilds());
	while(nbInfoReq>0)
		nbInfoReq_cond.wait(nbInfoReq_mut);
	nbInfoReq_mut.unlock();

	int rank;
	//update local info treesize
	childs_map_mut.lock();
	rank=info->rank;
	treesize=1; // it's me
	for(int i=0;i<childs_map.size();i++)
	{
		treesize+=childs_map[i]->size;
		dbg.log("scheduler","childs_map[%d]->size = %d\n",i,childs_map[i]->size);
	}
	dbg.log("scheduler","UPDATED treesize = %d\n",treesize);
	if(rank==0)
		info->count_nodes=treesize;
	childs_map_mut.unlock();
	
	if(rank!=0)
	{
		//send Info to father
		dbg.log("scheduler","---------NODE SEND GET INFO TO ROOT----------\n");
		_net->_LocalChannel_mut.lock();
		Taktuk::Network::TaktukLocalChannel* reply = _net->LocalChannel[0];
		_net->_LocalChannel_mut.unlock();
		acom::ActiveMessage<getinfo> am(_net, &_getinfo, 0);
		am->pack(&treesize,sizeof(acom::Network::LocalNodeNo));
		am.post(reply);
	
		//wait for father reply
		dbg.log("scheduler","---------NODE WAITING FOR INFO TO ROOT----------\n");
		InfoReply.wait();
	}
	else
	{
		RefreschInfo(false);
	}
}

void scheduler::SendChildsInfo()
{
	int nbchilds = getnbchilds();
	int begin=0;
	childs_map_mut.lock();
	if(info->rank!=0)
		begin=1;
	childs_map_mut.unlock();

	for(int LocalID=0;LocalID<nbchilds;LocalID++)
	{
		dbg.log("scheduler","Thread scheduler: no more job send node info to child %d\n",LocalID);
		
		dbg.log("scheduler","Thread scheduler: child rank %d\n",info->all_links[LocalID+begin]->_peer);
		_net->_LocalChannel_mut.lock();
		Taktuk::Network::TaktukLocalChannel* reply =	_net->LocalChannel[LocalID+begin];
		_net->_LocalChannel_mut.unlock();
		acom::AllocWriteAndSignal<setinfo> aws(_net, &_setinfo, 0);
		dbg.log("scheduler","============ SEND INFO TO CHILD %d  ==========\n",LocalID);
		//add info on the message
		aws->pack(&info->all_links[LocalID+begin]->_peer,sizeof(acom::Network::LocalNodeNo));
		aws->pack(&info->rank,sizeof(acom::Network::LocalNodeNo));
		aws->pack(&info->count_nodes,sizeof(acom::Network::LocalNodeNo));
		aws.post(reply);
	}
}

void scheduler::SubscribeToSteal(int local_node)
{
	StealRequestList_mut.lock();
	StealRequestList.put(local_node);
	StealRequestList_cond.signal();
	StealRequestList_mut.unlock();
}

int scheduler::WaitForStealer()
{
	int res=-1;
	StealRequestList_mut.lock();
	while(StealRequestList.empty()&&!End)
	//while(StealRequestList.empty())
		StealRequestList_cond.wait(StealRequestList_mut);
	if(StealRequestList.empty())
		res=-1;
	else
		res=StealRequestList.pop();
	StealRequestList_mut.unlock();
	return res;	
}

void scheduler::AutoReplyStealer(int nb)
{
	int nbreply =nb;
	int LocalID=-1;
	if(nb<=0)
		return;
	
	while(nbreply>0)
	{
		dbg.log("scheduler","Thread scheduler: no more job wait for %d steal's child\n",nbreply);
		LocalID=WaitForStealer();
		dbg.log("scheduler","Thread scheduler: no more job send node info to child %d\n",LocalID);
		
		dbg.log("scheduler","Thread scheduler: child rank %d\n",info->all_links[LocalID]->_peer);
		_net->_LocalChannel_mut.lock();
		Taktuk::Network::TaktukLocalChannel* reply = _net->LocalChannel[LocalID];
		_net->_LocalChannel_mut.unlock();
		acom::AllocWriteAndSignal<dojob> aws(_net, &_dojob, 0);
		dbg.log("scheduler","============ SEND NULL DOJOB TO CHILD %d  ==========\n",LocalID);
		aws.post(reply);
		--nbreply;
		--infochilds;
	}

	
	
	/*StealRequestList_mut.lock();
	dbg.log("scheduler","WaitForStealer waiting for %d have %d\n",nb,StealRequestList.size());
	while(StealRequestList.size()!=nb)
		StealRequestList_cond.wait(StealRequestList_mut);
	StealRequestList_mut.unlock();*/
}

int scheduler::getnbchilds()
{
	int res=0;
	childs_map_mut.lock();
	res=childs_map.size();
	childs_map_mut.unlock();
	return res;
}

void scheduler::GetChildsRank()
{
	childs_map_mut.lock();
	for(int i=0 ;i<childs_map.size();i++)
	{
		if(i==0)
      childs_map[i]->rank=info->rank+1;
		else
      childs_map[i]->rank=childs_map[i-1]->rank+childs_map[i-1]->size;

		//update infoTCPnet structure fiel _peer
		//WARNING if this node is root all_links == childs_map
		//but if is not, all_links size +1 == childs_map size
		// all_links 0 is the link to the up rank
		info->all_links[info->rank==0?i:i+1]->_peer=childs_map[i]->rank;
		dbg.log("scheduler","get childs rank\n");
	}
	for(int i=0;i<childs_map.size();i++)
		dbg.log("scheduler","get childs rank --> childs_map[%d] rank = %d\n",i,childs_map[i]->rank);
	for(int i=0;i<info->all_links.size();i++)
		dbg.log("scheduler","get childs rank --> all_links[%d] peer = %d\n",i,info->all_links[i]->_peer);
	childs_map_mut.unlock();
}

void scheduler::add_job(std::vector<topology::host*>& hlist)
{
	ToDoList_cond_mut.lock();
	if(hlist.empty())
		nojob=true;
	else
	{
		dbg.log("scheduler","ADD JOB\n");
		for(int i=0;i<hlist.size();i++)
			ToDoList.put(hlist[i]);
	}
	ToDoList_cond.signal();
	ToDoList_cond_mut.unlock();
}

void scheduler::set_info(int rank,int uprank, int countnodes)
{
	
	dbg.log("scheduler","Set Info begin\n");
	childs_map_mut.lock();
	info->rank=rank;
	info->all_links[0]->_peer=uprank; //this function is never called on root's
																	//tree 
	info->count_nodes=countnodes; //me
	dbg.log("scheduler","Set Info : updated count_nodes = %d\n",info->count_nodes);
	childs_map_mut.unlock();

	RefreschInfo(true);
	
	--InfoReply;
	dbg.log("scheduler","Set Info end\n");
}

void scheduler::GetRouteTable()
{
	childs_map_mut.lock();
	info->route_table.reserve(info->count_nodes);
  info->route_table.resize(info->count_nodes);
	dbg.log("scheduler","GetRouteTable rank = %d, count nodes = %d, treesize = %d\n",info->rank,info->count_nodes,treesize);
  for(int i =0; i<info->count_nodes; i++)
    {
    if(info->rank!=i)
    {
      //it's not me, what's the link where i must send data to reach the node i
        if( info->rank < i &&  i <= (info->rank+treesize-1) )
        {
					//if i is under me in the tree. ie myrank < i < myrank + my subtree size
					//look for link

          int j=0; // 0 is my top node
          for(;j<childs_map.size();j++)
            {
							// find my child with have a node id > i --> the good child is
							// before :-)
              if(childs_map[j]->rank>i || childs_map[j]->rank==-1)
                break;
            }
					// j == good link + 1 if i'm the root else j == good link
					if(info->rank==0)
						--j;
          info->route_table[i]=j;
					// node 0 have not top node so info.all_links.size have an element less
					// be carefull !!
        }
        else
        {
					// i is not under my node redirect to my top node
          info->route_table[i]=0;
        }
    }
    else
      info->route_table[i]=-1;
    }
	childs_map_mut.unlock();
}
