/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 * 
 * 
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */
#include <inuk/taktuk/topology.h>
#include <stdio.h>
#include <iostream>
#include <string>
#include <mmputenv.h>
#include "debuglog.h"
#include <inuk/taktuk/CoreTCP.h>

topology::topology()
{
  //initialisation of node's info
  connector="ssh"; 
  verbose=0;
  perf=0;
  arity=0;
  steal_num=0;
  cloexec=1;
  gethostname((char*)&HostName,MAX_HOST_NAME );
 
}

topology::~topology()
{
 
}


int topology::env_export()
{
  char env[512];

  sprintf(env,"taktukarity=%d",arity );
  if(mmputenv(env)<0)
  	return -1;
  dbg.log("topology","put env --> %s\n",env);
  sprintf(env,"taktukverbose=%d",verbose );
  if(mmputenv(env)<0)
  	return -1;
  dbg.log("topology","put env --> %s\n",env);
  sprintf(env,"taktukperf=%d",perf );
  if(mmputenv(env)<0)
  	return -1;
  dbg.log("topology","put env --> %s\n",env);
  for(int i=0;i<arity;i++)
  {
    sprintf(env,"taktukhostsmap_%d_rank=%d",i,hostsmap[i]->rank );
    if(mmputenv(env)<0)
  	  return -1;
    dbg.log("topology","put env --> %s\n",env);
    sprintf(env,"taktukhostsmap_%d_name=%s",i,hostsmap[i]->name.c_str());
    if(mmputenv(env)<0)
  	  return -1;
    dbg.log("topology","put env --> %s\n",env);
  }
  return 1;
}

int topology::env_recovery()
{
  char* inter;
  char req[512];

  dbg.log("topology","try env_recovery\n");

  if((inter=getenv("taktukarity"))!=NULL)
    {
      arity=atoi(inter);
			dbg.log("topology","get env arity = %d\n",arity);
      hostsmap.reserve(arity);
      hostsmap.resize(arity);
      for(int i=0;i<arity;i++)
        hostsmap[i]=new(topology::host);
    }
  else
    return -1;

  if((inter=getenv("taktukverbose"))!=NULL)
  {
    verbose=atoi(inter);
		dbg.log("topology","get env verbose = %s\n",verbose?"TRUE":"FALSE");
  }
  else
    return -1;

  if((inter=getenv("taktukperf"))!=NULL)
  {
    perf=atoi(inter);
	dbg.log("topology","get env perf = %s\n",perf?"TRUE":"FALSE");
  }
  else
    return -1;
  for(int i=0;i<arity;i++)
  {
  	 sprintf(req,"taktukhostsmap_%d_rank",i);
	 if((inter=getenv(req))!=NULL)
  		{
		    hostsmap[i]->rank=atoi(inter);
  		}
	  else
	    return -1;

  	 sprintf(req,"taktukhostsmap_%d_name",i);
	 if((inter=getenv(req))!=NULL)
  		{
		    hostsmap[i]->name=inter;
  		}
	  else
	    return -1;

  }
  return 1;
}

void topology::set_args(int argc, char* argv[])
{
  std::string new_arg("");
  
  for(int i=0;i<argc;i++)
    new_arg=new_arg + argv[i];
	for(int i=0;i<hostsmap.size();i++)
	{
		hostsmap[i]->userargs=new_arg;
 	}
}

void topology::set_args(std::string& args)
{
	for(int i=0;i<hostsmap.size();i++)
	{
		hostsmap[i]->userargs=args;
 	}
}

void topology::print()
{
  dbg.log("topo","user: %s\n",user.c_str());
  dbg.log("topo","passwd: %s\n",passwd.c_str());
  dbg.log("topo","connector: %s\n",connector.c_str());
  dbg.log("topo","rank: %d\n",info.rank);
  dbg.log("topo","count_nodes: %d\n",info.count_nodes);
  dbg.log("topo","arity: %d\n",arity);

  dbg.log("topo","verbose mode: ");
  if(verbose)
    dbg.log("topo","[YES]\n");
  else
    dbg.log("topo","[NO]\n");

 dbg.log("topo","keep fd alive mode: ");
  if(!cloexec)
    dbg.log("topo","[YES]\n");
  else
    dbg.log("topo","[NO]\n");

  dbg.log("topo","performance mode: ");
  if(perf)
    dbg.log("topo","[YES]\n");
  else
    dbg.log("topo","[NO]\n");


  

  dbg.log("topo","HOSTSMAP: \n");
  for(int i=0;i<hostsmap.size();i++)
    {
      dbg.log("topo","name=%s; connector=%s; taktukargs=%s ; userargs=%s ; login=%s ; passwd=%s ; rank=%d ; size=%d ; gateaccess=%s ; cmd=%s \n",
					hostsmap[i]->name.c_str(),
					hostsmap[i]->connector.c_str(),
					hostsmap[i]->taktukargs.c_str(),
					hostsmap[i]->userargs.c_str(),
					hostsmap[i]->login.c_str(),
					hostsmap[i]->passwd.c_str(),
					hostsmap[i]->rank,
					hostsmap[i]->size,
					hostsmap[i]->gateaccess.c_str(),
					hostsmap[i]->cmd.c_str());
    }

  dbg.log("topo","ROUTE_TABLE: \n");
  for(int i=0;i<info.route_table.size();i++)
    {
      dbg.log("topo","route to %d : %d\n",i,info.route_table[i]);
    }
  dbg.log("topo","REMOTE_CMDS: \n");
}

void topology::printstdout()
{
  printf("user: %s\n",user.c_str());
  printf("passwd: %s\n",passwd.c_str());
  printf("connector: %s\n",connector.c_str());
  printf("rank: %d\n",info.rank);
  printf("count_nodes: %d\n",info.count_nodes);
  printf("arity: %d\n",arity);

  printf("verbose mode: ");
  if(verbose)
    printf("[YES]\n");
  else
    printf("[NO]\n");

 	printf("keep fd alive mode: ");
  if(!cloexec)
    printf("[YES]\n");
  else
    printf("[NO]\n");

  printf("performance mode: ");
  if(perf)
    printf("[YES]\n");
  else
    printf("[NO]\n");



  printf("HOSTSMAP: \n");
  for(int i=0;i<hostsmap.size();i++)
    {
      printf("name=%s; connector=%s; taktukargs=%s ; userargs=%s ; login=%s ; passwd=%s ; rank=%d ; size=%d ; gateaccess=%s ; cmd =%s\n",
					hostsmap[i]->name.c_str(),
					hostsmap[i]->connector.c_str(),
					hostsmap[i]->taktukargs.c_str(),
					hostsmap[i]->userargs.c_str(),
					hostsmap[i]->login.c_str(),
					hostsmap[i]->passwd.c_str(),
					hostsmap[i]->rank,
					hostsmap[i]->size,
					hostsmap[i]->gateaccess.c_str(),
					hostsmap[i]->cmd.c_str());
    }

  printf("ROUTE_TABLE: \n");
  for(int i=0;i<info.route_table.size();i++)
    {
      printf("route to %d : %d\n",i,info.route_table[i]);
    }
  printf("REMOTE_CMDS: \n");
}
