/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 * 
 * 
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <inuk/acom>
#include "debuglog.h"
#include "topology_analyser.h"
#include "topology.h"
#include "IOredir.h"
#include "suboptions.h"

#define DEFAULTSTEAL 250
#define DEFAULT_TIMEOUT 5000




int topology_analyser::get(int* argc, char** argv[],topology* topo, 
                           bool new_arg_specif)
{
  new_arg_spec = new_arg_specif;
  //get local user name
  if(get_user_env(topo)<0)
    {
      return(-2);
    }
  if(options_parser(argc,argv,topo)<0)
    {
      return(-1);
    }
  
  //get user args
  args_parser(argc,argv,topo);        
  get_childs_cmds(topo);
  return(0);
}


//definition of parser function options 
#define OPTIONS "-hd::a:r:c:l:p:m:o::vPf:F:i[]K"
extern char* optarg;
extern int optind;


int topology_analyser::options_parser(int* argc, char** argv[],topology* topo)
{
  int opt;
  int hookcount=0;
  //int beginhook=0;
  int hooktreesize=0;
  //int endhook=0;
  std::string _host="";
  std::string _connector=topo->connector;
  std::string _login=topo->local_user;
  std::string _passwd=topo->passwd;
  std::string _taktukargs;
	std::string _optuserargs;
	int _timeout=DEFAULT_TIMEOUT;
  

  //read library options
  while((opt = getopt(*argc, *argv, OPTIONS))>0)
		switch(opt)
		{
 			case 'F':
				if(optarg != NULL)
				{
					if(hookcount==0)
					{
						std::ifstream file(optarg); 
						std::string host_name;
						if(file.is_open())
						{
							while(!file.eof())
							{
								file>>host_name>>std::ws;
								topo->hostsmap.push_back(new topology::host(host_name,
																													_connector,
																													"",
																													"",
																													_login,
																													_passwd,
																													"",
																													_timeout,
																													_optuserargs));
							}
							file.close();
						}
						else
						{
									std::cerr<<"remote file specified not found"<<std::endl;
									return -1;
						}
					}
					else
					{
						_taktukargs=_taktukargs + " " + "-F" + optarg;
					}
				}
				else
				{
					std::cerr<<"remote filename missing for -F option"<<std::endl;
					return (-1);
				}
				break;
			case 'f':
				if(optarg != NULL)
				{
					std::ifstream file(optarg); 
					std::string host_name;
					if(file.is_open())
					{
						while(!file.eof())
						{
							file>>host_name>>std::ws;
							if(hookcount==0)
								topo->hostsmap.push_back(new topology::host(host_name,
																														_connector,
																														"",
																														"",
																														_login,
																														_passwd,
																														"",
																														_timeout,
																														_optuserargs));
							else
								_taktukargs=_taktukargs + " " + "-m" + host_name;
						}
						file.close();
					}
					else
					{
						std::cerr<<"local file specified not found"<<std::endl;
						return (-1);
					}
				}
				else
				{
					std::cerr<<"local filename missing for -f option"<<std::endl;
					return (-1);
				}
				break;
			case 'K':
        std::cerr<<"\n\
									\nGateau mre Bonvallet\n\n Ingrdients: \n\n3  Paquets de biscuits   THE BRUN.\
									\n70 Gr de beurre\
									\n70 Gr de sucre en poudre\
									\n2 jaunes d'oeuf\
									\n\n\nFaire du caf , verser dans un saladier mlanger avec du rhum ,\nlaisser refroidir.\n\
									\n\nVerser le sucre dans un saladier, ajouter les jaunes d'oeuf \net fouetter avec une fourchette.\n\
									\n\nFaire fondre le beurre dans une casserole en le remuant \nconstamment sans le laisser dorer .\n\n\
									Verser les jaunes d'oeuf,fouetter et bien remuer\
									\njusqu' ce que la crme devienne paisse.\
									\n\nTremper les biscuits dans le caf, les poser en ranges dans un plat.\
									\n\ntaler  une cuillre de crme avec une fourchette ,\
									\nsaupoudrer de cacao et recommencer l'opration par couche \
									\njusqu' ce que le gteau soit fini.\
									\n\nNapper avec du chocolat rp et du sucre glace ou petits bonbons.\
									\n\n\n\n\n\n"<<std::endl;
				exit(0);
        break;
      case 'c':
        if(optarg != NULL)
				{
					if(hookcount==0)
					{
						_timeout=0;
						char * tokens[]= {"connector","timeout",NULL};
						char *optionargs=strncpy((char*)malloc(strlen(optarg)),optarg,strlen(optarg));
						char * subargs;
						int suboption=0;
						int nosuboptions=1;
							
						while((suboption=suboptions(&optionargs,(char**)&tokens,&subargs))!=-1)
						{
							nosuboptions=0;
							switch(suboption)
							{
								case 0:
									if(subargs!=NULL)
										_connector=subargs;
									break;
								case 1:
									if(subargs!=NULL)
										_timeout=atoi(subargs);
										dbg.log("timeout","-c option timeout= %d\n",_timeout);
									break;
								default:
									break;
							} 
						}
						if(nosuboptions)
						{
							_connector=optarg;
						}
					}
					else
						_taktukargs=_taktukargs + " " + "-c" + optarg;
				}
        else
				{
					std::cerr<<"bad -c option look at -h or --help for display usage"<<std::endl;
					return (-1);
				}
        break;
      case 'a':
        if(optarg != NULL)
				{
					if(hookcount==0)
					{
						topo->arity = atoi(optarg);
						if(topo->arity==0)
							topo->arity++;
					}
					else
						_taktukargs=_taktukargs + " " + "-a" + optarg;
				}
        else
				{
					std::cerr<<"bad -a option"<<std::endl;
					return(-1);
				}
        break;
			case 'o':
        if(optarg != NULL)
				{
					if(hookcount==0)
					{
						_optuserargs = optarg;
					}
					else
						_taktukargs=_taktukargs + " " + "-o'" + optarg +"'";
				}
        else
				{
					_optuserargs="";
				}
        break;
			case 'd':
        if(optarg != NULL)
				{
					if(hookcount==0)
					{
						topo->steal_num = atoi(optarg);
						topo->arity=0;
					}
					else
						_taktukargs=_taktukargs + " " + "-d" + optarg;
				}
        else
				{
					if(hookcount==0)
					{
						topo->steal_num = DEFAULTSTEAL;
						topo->arity=0;
					}
					else
					{
						_taktukargs=_taktukargs + " " + "-d";
					}
				}
        break;
      case 'l':
        if(optarg != NULL)
				{
			 		if(hookcount==0)
					{
		 				topo->user.append(optarg);
						_login=optarg;
					}
	 				else
 						_taktukargs=_taktukargs + " " + "-l" + optarg;
				}
				else
				{
					std::cerr<<"bad -l option"<<std::endl;
					return(-1);
				}
        break;
      case 'p':
        if(optarg != NULL)
				{
					if(hookcount==0)
						topo->passwd.append(optarg);
					else
						_taktukargs=_taktukargs + " " + "-p" + optarg;
				}
        else
				{
					std::cerr<<"bad -p option"<<std::endl;
					return(-1);
				}
        break;
      case 'h':
        return(-1);
        break;
      case 'v':
        topo->verbose=1;
        break;
      case 'i':
        topo->cloexec=0;
        break;
			case 'r':
				topo->my_root.append(optarg);
				topo->info.rank=~acom::Network::LocalNodeNo(0);
				break;
      case '[':
        hookcount++;
        if(hookcount>1)
          _taktukargs=_taktukargs + " " + "-[";
        break;
      case ']':
        hookcount--;
        if(hookcount==0)
				{
					if(topo->hostsmap.size()>0)
					{
						topo->hostsmap[topo->hostsmap.size()-1]->taktukargs=_taktukargs;
						topo->hostsmap[topo->hostsmap.size()-1]->size=hooktreesize+1;// size of this subtree root include
					}
					else
						return -1; //syntaxe error one "-m host" must be declare before
					_taktukargs="";
					hooktreesize=0;
				}
        else
          _taktukargs=_taktukargs + " " + "-]";
        break;
      case 'P':
        topo->perf=1;
        break;
      case 'm':
        if(optarg != NULL)
				{
					if(hookcount==0)
					{
						char* tokens[]= {"host","connector","login","passwd","gate","timeout",NULL};
						char *optionargs=strncpy((char*)malloc(strlen(optarg)+1),optarg,strlen(optarg));
						optionargs[strlen(optarg)]='\0';
						char *subargs;
						int suboption=0;
						
						int nosuboptions=1;
						//_login=topo->local_user;
						//_passwd=topo->passwd;

						std::string l_host=_host;
						std::string l_connector=_connector;
						std::string l_login=_login;
						std::string l_passwd=_passwd;
						std::string l_gateaccess;
						int l_timeout=_timeout;
							
						while((suboption=suboptions(&optionargs,(char**)&tokens,&subargs))!=-1)
						{
							nosuboptions=0;
							switch(suboption)
							{
								case 0:
									if(subargs!=NULL)
										l_host=subargs;
									break;
								case 1:
									if(subargs!=NULL)
										l_connector=subargs;
										l_timeout=0;
									break;
								case 2:
									if(subargs!=NULL)
										l_login=subargs;
									break;
								case 3:
									if(subargs!=NULL)
										l_passwd=subargs;
									break;
								case 4:
									if(subargs!=NULL)
										l_gateaccess=subargs; 
									break;
								case 5:
									l_timeout=atoi(subargs); // compute node as default
									break;
								default:
									break;
							} 
						}
						if(nosuboptions)
						{
							l_host=optarg;
						}
						topo->hostsmap.push_back(new topology::host(l_host,l_connector,"","",l_login,l_passwd,l_gateaccess,l_timeout,_optuserargs));
					}
					else
					{
						_taktukargs=_taktukargs + " " + "-m" + optarg;
						hooktreesize++;
					}
				}
        else
				{
					std::cerr<<"bad -m option"<<std::endl;
					return(-1);
				}
        break;
      default:
        return(-1);
        break;
      }
	// default behavior flat tree
  if(topo->arity==0)
    topo->arity=topo->hostsmap.size();
//	if(topo->steal_num==0)
//		topo->steal_num=topo->arity;

  if(hookcount>0)// forget a -]
	{
		std::cerr<<"Syntaxe ERROR forget almost one -] at the end???"<<std::endl;
		return(-1);
	}
  //current argv == "--" or nothing else error 
  
  dbg.log("","optind=%d\n",optind);
  dbg.log("","(*argv)[optind]=%X\n",(*argv)[optind]);
  dbg.log("","(*argv)[optind-1]=%s\n",(*argv)[optind-1]);
  if (!new_arg_spec) {
    dbg.log("", "this is old argument parsing\n");
    if((*argv)[optind]==NULL || (strcmp((*argv)[optind-1],"--")==0) )
      return(0);
    else
      return(-1);
  } else {
    dbg.log("", "this is new argument parsing\n");
    if(optind == *argc)
      return(0);
    else
      return(-1);
  }
}

void topology_analyser::args_parser(int* argc, char** argv[],topology* topo)
{
  if (!new_arg_spec) {
    dbg.log("", "this is old argument parsing\n");
    //stock the global user arguments
    std::string total_args;
    prog_name=(*argv)[0];
    for(int i=optind;i<(*argc);i++)
      {
        total_args=total_args + (*argv)[i]+" ";
      }
	
	//reinit optind for user
	optind=0;
    //add global args to the hostsmap
    for(int i=0;i<topo->hostsmap.size();i++)
      {
        topo->hostsmap[i]->userargs+=total_args;
      }
    
    
    //clean the argc argv for the user of the library
    int i=1;
    int badopt=0;
    
    while((*argv)[i]!=0)
      {
        if(strcmp((*argv)[i],"--")==0)
          {
            //jump "--"
            i++;
            break;
          }
        if(strstr((*argv)[i],"-m")!=NULL||strstr((*argv)[i],"-f")!=NULL||strstr((*argv)[i],"-F")!=NULL||strstr((*argv)[i],"-o")!=NULL)
          {
            badopt=1;
          }
        // save general options like -i -v -P -a -c but jump -rargs option
        if(badopt<1 && !(strncmp((*argv)[i],"-r",2)==0))
          //if(badopt<1 && strcmp((*argv)[i],"-r")<=0)
          taktuk_args= taktuk_args+ " " + (*argv)[i];
        i++;
      }
    //build new argc & argv for the user
    if((*argv)[i]!=0)
      {
        for(int j=0;j<(*argc-i);j++)
          (*argv)[j+1]=(*argv)[j+i];
        *argc=*argc-i+1;
        (*argv)[*argc]=NULL;
      }
    else
      {
        *argc=1;
      }
    
  } else { // new arg parsing 
    dbg.log("", "this is new argument parsing\n");
    prog_name=ACOM::Acom::arg->progname();
    for(int i=0;i<topo->hostsmap.size();i++) {
      //    topo->hostsmap[i]->userargs = "-- TAKTUK " + topo->hostsmap[i]->userargs;
      ACOM::Arg::const_iterator iend = ACOM::Acom::arg->end();  
      ACOM::Arg::iterator icurr;
      for (icurr = ACOM::Acom::arg->begin(); icurr != iend; icurr++) {
        if ((icurr->first != "_all_") && (icurr->first != "TAKTUK")) {
          ACOM::Arg::ArgVect av = icurr->second;
          if (av.size() == 1)
            continue;
          topo->hostsmap[i]->userargs += " -- " + icurr->first;
          for (int j = 1; j < av.size(); j++)
            topo->hostsmap[i]->userargs += " " + av[j] ;
        }
      }
    }
    
    int badopt=0;
    
    ACOM::Arg::ArgVect av = (*ACOM::Acom::arg)["TAKTUK"];
    for (int j = 1; j < av.size(); j++) {
      if(strstr(av[j].c_str(),"-m")!=NULL||strstr(av[j].c_str(),"-f")!=NULL)
        {
          badopt=1;
        }
      // save general options like -i -v -P -a -c but jump -rargs option
      if(badopt<1 && !(strncmp(av[j].c_str(),"-r",2)==0))
        //if(badopt<1 && strcmp(av[j].c_str(),"-r")<=0)
        taktuk_args += " " + av[j];
    }
  }

}

int topology_analyser::get_user_env(topology* topo)
{
  char* user;
  //get env variable
  if(topo->user.empty())
    {
      if((user=getenv("LOGNAME"))!=NULL)
        topo->user.append(user);
    }
  if(topo->user.empty())
    {
      if((user=getenv("USER"))!=NULL)
        topo->user.append(user);
    }
  if(topo->user.empty())
    {
     if((user=getenv("USERNAME"))!=NULL)
        topo->user.append(user);
    }
  if(topo->user.empty())
    {
      return(-2);
    }
  topo->local_user=getenv("USER");
  if(topo->local_user.empty())
    topo->local_user=getenv("LOGNAME");
  if(topo->local_user.empty())
    topo->local_user="nobody";
  return 1;
}


void topology_analyser::addnodecmd(const topology::host* h, std::string& cmd)
{
	cmd = cmd +	"-o'" +h->optuserargs + "'" 
						+ " -m host=" +  h->name 
						+ ",connect=" + h->connector 
						+ ",login=" +	h->login;
	if(!h->passwd.empty())						
		cmd = cmd + ",passwd=" + h->passwd;
		
	cmd = cmd	+ " -[ " + h->taktukargs + " -] ";

}


void topology_analyser::get_childs_cmds(topology* topo)
{
  int nbtoreduce=0;
	int nodeadded=0;
	int noderest=0;
	//reduce hostsmap to arity

	if(topo->arity<topo->hostsmap.size())
	{
		nbtoreduce=topo->hostsmap.size()-topo->arity;
		nodeadded=nbtoreduce/topo->arity;
		noderest=nbtoreduce%topo->arity;
		dbg.log("","nbtoreduce %d\n",nbtoreduce);
		dbg.log("","nodeadded %d\n",nodeadded);
		dbg.log("","noderest %d\n",noderest);
		for(int i=0;i<topo->arity;i++)
		{
			for(int j=0;j<nodeadded;j++)
			{
				//add node j in the cmd of node i
				addnodecmd(topo->hostsmap[j+topo->arity+i*nodeadded],topo->hostsmap[i]->taktukargs);
				topo->hostsmap[i]->size+=topo->hostsmap[j+topo->arity+i*nodeadded]->size;
			}
			if(i==0)
			{
				//the first node herit of the rest
				for(int j=0;j<noderest;j++)
				{
					//add node j in the cmd of node i
					addnodecmd(topo->hostsmap[j+topo->arity+topo->arity*nodeadded],topo->hostsmap[i]->taktukargs);
					topo->hostsmap[i]->size+=topo->hostsmap[j+topo->arity+topo->arity*nodeadded]->size;
				}
			}
		}
		//remove old nodes hostsmap[arity .. end]
		topo->hostsmap.resize(topo->arity);
	}

	// now hostsmap containt all we must proceed not more
  for(int j=0;j<topo->hostsmap.size();j++)
	{
      if (!new_arg_spec)
		topo->hostsmap[j]->cmd = prog_name + " -r" + topo->HostName + " " + taktuk_args + " " + topo->hostsmap[j]->taktukargs;
      else
		topo->hostsmap[j]->cmd = prog_name + " -- TAKTUK" + " -r" + topo->HostName + " " + taktuk_args + " " + topo->hostsmap[j]->taktukargs;


	}
}

